/* cfs.c -- core routines for Coda inode manipulation
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include "MC.h"
#include <string.h>
#pragma intrinsic(strcpy, strlen, memcpy)

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

int coda_root(){
	int res = 0;
	struct cnode *cn = NULL;
	ViceFid VFid;

	res = venus_root(&VFid);
	if (res)
		return ERROR_DEV_NOT_EXIST;

	cn = cnode_alloc();
	coda_instantiate(root, cn, &VFid);
	return 0;
}

int openVFid (struct cnode *cn, int flags, WORD access, WORD attr, BYTE action, 
			  struct request *inreq, PDWORD phdl, PDWORD ptakenaction,
			  PWORD pattr, PWORD hidden)
{
	char path[MAX_PATH];
	DWORD takenaction;
	int res = 0;

	LOGDEBUG("--openVFid--", 1);

	*phdl = 0;
	printf("---openVFid flags: %x ---", flags);

	if (!(flags & C_O_WRITE) && VALID_CON(cn) && hidden){
		strcpy((char *)path, cn->c_container);
		res = 0;
		*hidden = 1;
	}
	else {
		res = venus_open_by_path(cn->c_fid, (char *)&path, inreq, flags);		
		LOGDEBUG("CONTAINER WAS OPENED", 100);
		print_fid(&cn->c_fid);
		if (hidden) {
			if (cn->c_container) _HeapFree(cn->c_container, 0); 
			cn->c_container = (char *) _HeapAllocate(strlen(path)+1, 0);
			strcpy(cn->c_container, path);
			cn->c_flags |= C_CONTAINER;
			*hidden = 0;
		}
	}

	if (res) goto error;

	if (!ptakenaction)
		ptakenaction = &takenaction;

//	access = (access | SHARE_DENYNONE) & ACCESS_MODE_MASK & SHARE_MODE_MASK;
	res = R0_OpenCreateFile ((BOOL) FALSE,
		(WORD) (access & ACCESS_MODE_MASK),
		attr,
		ACTION_EXISTS_OPEN/*action*/,
		(BYTE) 0,
		path,
		phdl,
		ptakenaction);

	if (res) {
		printf ("R0_OpenCreate failed in openVFid %x", res);
		goto error;
	}

	/* sometimes we need attributes of opened file */
	if (pattr) {
		res = R0_GetFileAttributes(path, pattr);
		if (res) {
			LOGDEBUG("GetFileAttributes failed", 100);
			goto error;
		}
	}
	printf("open handle: %x", *phdl);

error:
	if (res && *phdl)
		R0_CloseFile(*phdl);
	LOGDEBUG("--end of openVFid--", 1);
	return res;
}


int coda_getattr (struct cnode *cn)
{
	int res = 0;
	LOGDEBUG("--coda_getattr--",1);
#ifdef CACHE
	if (VALID_VATTR(cn)){
		LOGDEBUG("---CACHEHIT---", 10);
		return 0;
	}
#endif
	res = venus_getattr(cn);

	if (res == 0)
		cn->c_flags |= C_VATTR;
	else
		cn->c_flags &= ~C_VATTR;

	LOGDEBUG("--end of coda_getattr--", 1);
	return res;
}

int coda_setattr (struct cnode *cn)
{
	int res = 0;
	LOGDEBUG("--coda_setattr--", 1);
	res = venus_setattr(cn);

	if (res == 0)
		cn->c_flags |= C_VATTR;
	else
		cn->c_flags &= ~C_VATTR;

	LOGDEBUG("--end of coda_setattr--", 1);
	return res;
}

int coda_namei(char *upp, struct dentry **de, int flags){

	int res = 0;
	LOGDEBUG("---coda_namei---", 1);

	if (!root->d_vnode)
		res = coda_root(root);
	if (res){
		LOGDEBUG("---end of coda_namei---", 1);
		return res;
	}

	if (!upp){
		LOGDEBUG("----root----", 10);
		*de = dget(root);
		return 0;
	}

	printf("de: %x", de);
	printf("*de: %x", *de);
	res = d_lookup(my_fs, upp, NULL, de, flags); 
	printf("de: %x", de);
	printf("*de: %x", *de);
	LOGDEBUG("---end of coda_namei---", 1);

	return res;
}

