/* devio.c -- DeviceIoControl interface with Venus
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/


#include "mc.h"

#include <string.h>
#pragma intrinsic(strlen, strcpy, memset, memcpy, strcmp, memcmp)

#include "vmmextra.h"

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

int calls[10] = {0,0,0,0,0, 0,0,0,0,0};
static PCHAR Path[] = {"\\\\CODA\\CLUSTER"};
int first_call = 0;
HANDLE hHandle = 0;
struct filesystem *my_fs = NULL;

DWORD DevIo_Status (PDIOCPARAMETERS);
DWORD DevIo_Mount (PDIOCPARAMETERS);
DWORD DevIo_GetRequest (PDIOCPARAMETERS);
DWORD DevIo_ServiceRequest (PDIOCPARAMETERS);
DWORD DevIo_RPC(PDIOCPARAMETERS);
DWORD DevIo_InsertUpcall(PDIOCPARAMETERS);
DWORD DevIo_Unmount (PDIOCPARAMETERS);
DWORD DevIo_AllocateFD (PDIOCPARAMETERS);
DWORD DevIo_DeallocateFD (PDIOCPARAMETERS);
DWORD DevIo_Pioctl (PDIOCPARAMETERS);

void PM_API (HVM, DWORD, PCRS);

void __declspec(naked) CODADEV_PM_API ()
{
	_asm {
		push	ebp
		push	edi
		push	ebx
		call	PM_API
		pop		eax
		pop		eax
		pop		eax
		ret
	}
}

void PM_API (HVM hVM, DWORD hThread, PCRS pRegs)
{
	DIOCPARAMETERS params;

        calls[1] = pRegs->Client_EDX;

	params.VMHandle = (DWORD) hVM;
	params.dwIoControlCode = pRegs->Client_EAX;
	params.lpvInBuffer = Client_Ptr_Flat(CLIENT_DS, CLIENT_SI);
	params.cbInBuffer = pRegs->Client_EBX;
	params.lpvOutBuffer = Client_Ptr_Flat(CLIENT_DS, CLIENT_DI);
	params.cbOutBuffer = pRegs->Client_ECX;
	params.lpcbBytesReturned = (DWORD) &pRegs->Client_ECX;
	params.lpoOverlapped = 0;
	params.hDevice = 0;
	params.tagProcess = 0;

	pRegs->Client_EAX = OnW32DeviceIoControl (hVM, &params);
	if (pRegs->Client_EAX)
		pRegs->Client_EFlags |= CF_MASK;
	else
		pRegs->Client_EFlags &= ~CF_MASK;
}

DWORD OnW32DeviceIoControl (HVM hVM, PDIOCPARAMETERS p)
{
	
//	printf("IOCTL Nr:%x\n", p->dwIoControlCode);
	switch (p->dwIoControlCode) {
	case DIOC_GETVERSION:
		return 0;

	case 1:
		return DevIo_Status(p);

	case 2:
		return DevIo_Mount(p);

	case 3:
		return DevIo_GetRequest(p);

	case 4:
		return DevIo_ServiceRequest(p);

	case 6:
		return DevIo_RPC(p);

	case 7:
		return DevIo_InsertUpcall(p);

	case 8:
		return DevIo_Unmount(p);

	case 9:
		return DevIo_AllocateFD(p);

	case 10:
		return DevIo_DeallocateFD(p);

	case 11:
	case 304922:/* specified for nt. used here to keep cygwin apps unchanged for win95*/
		return DevIo_Pioctl(p);

	}

	return ERROR_INVALID_PARAMETER;
}


DWORD DevIo_Status (PDIOCPARAMETERS p)
{
	PDWORD pOut;
	int i, res;

	if (!p->lpvOutBuffer || p->cbOutBuffer < 10 * sizeof(DWORD))
		return ERROR_INVALID_PARAMETER;

	res = initialize_ipc();
	if (res)
		return res;

	pOut = (PDWORD) p->lpvOutBuffer;

	pOut[0] = uiNetId;
	pOut[1] = uiProviderId;
	for (i=0; i<8; i++)
		pOut[i+2] = calls[i];

	return 0;
}


DWORD Do_Mount (char cDrive, char *szPath);

DWORD DevIo_Mount (PDIOCPARAMETERS p)
{
	char *lpcInBuffer = (char *) p->lpvInBuffer;
	char cDrive;
	char *szPath;
	int cbPath;
	DWORD res;

	LOGDEBUG("---DevIo_Mount---", 1);
	if (!p->lpvInBuffer || (p->cbInBuffer < 2))
		return ERROR_INVALID_PARAMETER;

	cDrive = lpcInBuffer[0];

	cbPath = strlen (lpcInBuffer + 1);
	szPath = _HeapAllocate (cbPath + 1, 0);
	if (!szPath) 
		return ERROR_NOT_ENOUGH_MEMORY;
	strcpy (szPath, lpcInBuffer + 1);

	res = Do_Mount (cDrive, szPath);

	_HeapFree (szPath, 0);
	LOGDEBUG("---End of DevIo_Mount---", 1);
	return res;
}

DWORD Do_Mount (char cDrive, char *szPath)
{
	int res;
	struct filesystem *_f = VDCACHE_New_FS();

	LOGDEBUG("Do_Mount", 1);

	if (!_f)
		return ERROR_NOT_ENOUGH_MEMORY;

	_f->f_letter = cDrive;
	_f->f_path = Path[0];

	_f->f_type.fst_type = remote;
	_f->f_type.fst_read_super = ConnectNetResource;
	_f->f_type.fst_put_super = DisconnectResource;

	res = register_filesystem(_f);
	if (res){
		VDCACHE_Del_FS(_f);
		return res;
	}

	my_fs = _f;

	LOGDEBUG("End of Do_Mount", 1);

	return 0;
}


DWORD DevIo_Unmount(PDIOCPARAMETERS p)
{
	char *lpcInBuffer = (char *) p->lpvInBuffer;
	char *szDriveStr = "X:";
	int res = 0;
	LOGDEBUG("--Unmount--", 1);

//	if (!p->lpvInBuffer || (p->cbInBuffer < 2))
//		return ERROR_INVALID_PARAMETER;

//	szDriveStr[0] = lpcInBuffer[0];

//	res = VDCACHE_Detach_FS(szDriveStr);
	res = unregister_filesystem(my_fs);

	if (res){
		printf("error Unmount %x", res);
		return res;
	}

	VDCACHE_Del_FS(my_fs);

	return 0 ;
}


DWORD DevIo_GetRequest (PDIOCPARAMETERS p)
{
	struct request *req;
	BOOL blockflag = FALSE;

	if (!p->lpvOutBuffer)
		return ERROR_INVALID_PARAMETER;

	if (p->lpvInBuffer && (p->cbInBuffer >= sizeof(BOOL)))
		blockflag = *(BOOL *)p->lpvInBuffer;

	req = get_pending(blockflag);
	if (!req) {
		if (blockflag)
			LOGDEBUG("blocking get_pending returned NULL", 10);
		return ERROR_HANDLE_EOF;
	}

	if (p->cbOutBuffer < req->length) 
		return ERROR_INVALID_PARAMETER;
	memcpy ((PBYTE) p->lpvOutBuffer, &req->r, req->length);

	/* if it is an upcall, kill the request buffer */
	if (req->r.in.ih.opcode >= 100)
		FreeRequest (req);
	return 0;
}


DWORD DevIo_ServiceRequest (PDIOCPARAMETERS p)
{
	union outputArgs *out;
	struct request *req;

    if (!p->lpvInBuffer)
		return ERROR_INVALID_PARAMETER;

	if (!first_call){
		HVM hvm = (HVM) p->VMHandle;
		int flags = VMSTAT_BACKGROUND|VMSTAT_HIGH_PRI_BACK;

		_asm	mov     ebx, [hvm]
		VMMCall(Get_Time_Slice_Priority)
		_asm	or      eax, flags
		VMMCall(Set_Time_Slice_Priority)

/*		hHandle = Get_Cur_Thread_Handle();*/

		first_call = 1;
	}

	out = (union outputArgs *) p->lpvInBuffer;

	if (DOWNCALL(out->oh.opcode)) {
		/* handle downcalls */
		return coda_downcall(out);
	}

	req = find_request(out->oh.unique);
	if (!req) {
		LOGDEBUG("find_request returned NULL", 10);
		return ERROR_INVALID_DATA;
	}

	/* copy reply into kernel request struct */
	if (p->cbInBuffer > req->length) {
		LOGDEBUG("ERROR -- insufficient space in kernel request buffer for server reply", 10);
		return ERROR_INVALID_DATA;
	}
	memcpy (&req->r, (PBYTE) p->lpvInBuffer, p->cbInBuffer);

	signal_request(req);
	return 0;
}


DWORD DevIo_RPC (PDIOCPARAMETERS p)
{
	struct request *req;

	if (!p->lpvInBuffer || 
		(p->cbInBuffer < USER_LEVEL_REQUEST_SIZE) ||
		!p->lpvOutBuffer || 
		(p->cbOutBuffer < USER_LEVEL_REQUEST_SIZE))
		return ERROR_INVALID_PARAMETER;

	/* first, copy in the result of a previous RPC */
	req = find_request(((union outputArgs *) p->lpvInBuffer)->oh.unique);
	if (!req) {
		LOGDEBUG("find_request returned NULL", 10);
		return ERROR_INVALID_DATA;
	}

	/* copy reply into kernel request struct */
	if (p->cbInBuffer > req->length) {
		LOGDEBUG("ERROR -- insufficient space in kernel request buffer for server reply", 10);
		return ERROR_INVALID_DATA;
	}
	memcpy (&req->r, (PBYTE) p->lpvInBuffer, p->cbInBuffer);

	signal_request(req);

	/* second, get a new request */
	req = get_pending(TRUE);
	if (!req) {
		LOGDEBUG("blocking get_pending returned NULL", 10);
		return ERROR_HANDLE_EOF;
	}

	if (p->cbOutBuffer < req->length) 
		return ERROR_INVALID_PARAMETER;
	memcpy ((PBYTE) p->lpvOutBuffer, &req->r, req->length);

	/* if it is an upcall, kill the request buffer */
	if (req->r.in.ih.opcode >= 100)
		FreeRequest (req);
	return 0;
}

DWORD DevIo_InsertUpcall (PDIOCPARAMETERS p)
{
	DWORD func;
	struct request *req;

	if (!p->lpvInBuffer || 
		(p->cbInBuffer != sizeof(DWORD)))
		return ERROR_INVALID_PARAMETER;

	memcpy((PBYTE) &func, (PBYTE) p->lpvInBuffer, sizeof(DWORD));

	req = NewRequest(0);
	if (!req)
		return ERROR_NOT_ENOUGH_MEMORY;

	req->r.in.ih.opcode = func;
	return insert_upcall (req);
}

#define SOCK_DEVICE_ID			0x1235
Begin_Service_Table(SOCK, none)
#define SOCK_Service(X) Declare_Service(X, none)
SOCK_Service(SOCK_Get_Version)
SOCK_Service(_SOCK_AllocateFD)
SOCK_Service(_SOCK_DeallocateFD)
End_Service_Table(SOCK, none)

#pragma warning(disable: 4035)
DWORD SOCK_Get_Version()
{
	DWORD res;
	VxDCall(SOCK_Get_Version)
	_asm jnc noerror
	_asm mov eax, 0h
	_asm noerror:
	_asm mov [res], eax
	return res;
}

typedef struct _FDOPS {
        DWORD   (*fdop_poll) (PVOID context, PDWORD post);
        void    (*fdop_close) (PVOID context);
} FDOPS, *PFDOPS;

FDOPS mc_ops = {do_poll, do_close};

__declspec(naked) DWORD VXDINLINE SOCK_AllocateFD (HVM hVM, DWORD fd, 
												   PFDOPS fdops, 
												   PVOID context)
{
	VxDJmp(_SOCK_AllocateFD)
}

__declspec(naked) DWORD VXDINLINE SOCK_DeallocateFD (HVM hVM, DWORD fd)
{
	VxDJmp(_SOCK_DeallocateFD)
}
#pragma warning(default: 4035)


DWORD DevIo_AllocateFD (PDIOCPARAMETERS p)
{
	DWORD fd;

	if (!SOCK_Get_Version())
		return ERROR_FILE_NOT_FOUND;

	if (!p->lpvInBuffer || (p->cbInBuffer != sizeof(DWORD)))
		return ERROR_INVALID_PARAMETER;

	fd = *(PDWORD)p->lpvInBuffer;

	return SOCK_AllocateFD (Get_Cur_VM_Handle(), fd, &mc_ops, NULL);
}

DWORD DevIo_DeallocateFD (PDIOCPARAMETERS p)
{
	DWORD fd;

	if (!SOCK_Get_Version())
		return ERROR_FILE_NOT_FOUND;

	if (!p->lpvInBuffer || (p->cbInBuffer != sizeof(DWORD)))
		return ERROR_INVALID_PARAMETER;

	fd = *(PDWORD)p->lpvInBuffer;

	return SOCK_DeallocateFD (Get_Cur_VM_Handle(), fd);
}

DWORD DevIo_Pioctl (PDIOCPARAMETERS p)
{
	PCHAR szFilename;
	struct PioctlData PID;
	struct cnode *cn = NULL;
	struct dentry *de = NULL;
	int res = 0, i=0, j=0;
	unsigned long cmd;
	ViceFid fid;

	if (!p->lpvInBuffer || (p->cbInBuffer < sizeof(struct PioctlData)))
		return ERROR_INVALID_PARAMETER;

	LOGDEBUG("--pioctl--", 1);
	
	memcpy(&PID, (struct PioctlData *)p->lpvInBuffer, sizeof(struct PioctlData));
	memcpy(&cmd, (unsigned long *)p->lpvOutBuffer, sizeof(cmd));

	szFilename = _HeapAllocate(MAX_PATH + 1, 0);
	if (!szFilename)
		return ERROR_NOT_ENOUGH_MEMORY;

	memcpy(szFilename, (char *)(p->lpvInBuffer + sizeof(struct PioctlData)), 
		strlen((char *)(p->lpvInBuffer + sizeof(struct PioctlData)))+1);

	j = (strlen(szFilename));
	for (i = 0; i < j; i++) if (szFilename[i]=='\\') szFilename[i]='/';

	PID.vi.out = (void *)(p->lpvInBuffer + sizeof(struct PioctlData) + strlen(szFilename)+1);
	PID.vi.in  = (void *)(p->lpvInBuffer + sizeof(struct PioctlData) + strlen(szFilename)+1 + PID.vi.out_size);

	printf ("-----cmd %x\n", cmd);

	if (cmd == 533260){ /* this number results from com = 8972 */
		/* windows - specific getfid */
		LOGDEBUG("pioctl cmd 8972", 1);
		
		printf("de: %x", de);
		printf("&de: %x", &de);
		res = coda_namei (!(strcmp(szFilename, "/coda")) ? NULL : szFilename , &de, CLU_CASE_INSENSITIVE);
		printf("de: %x", de);
		printf("&de: %x", &de);
		printf("de->cn: %x", de->d_vnode);
		_HeapFree (szFilename, 0);
		if (res) {
			printf ("namei failed %x", res);
			return res;
		}
		printf("hallo1");
		*(PDWORD)p->lpcbBytesReturned = 0;
		printf("hallo2");
		cn = (struct cnode*)de->d_vnode;
		printf("hallo3");
		if (!cn || !IS_INCONF(cn)){
			printf("hallo4");
			dput(de);
			printf("hallo5");
			return res;
		}
		printf("hallo6");
		if (sizeof(ViceFid) <= (DWORD)p->cbOutBuffer){
			memcpy((PCHAR)p->lpvOutBuffer, &cn->c_fid, sizeof(ViceFid));
			*(PDWORD)p->lpcbBytesReturned = sizeof(ViceFid);
		}
		printf("hallo7");
		dput(de);
		printf("hallo8");
		return res;
	}

	LOGDEBUG(szFilename, 1);

	res = coda_namei (!(strcmp(szFilename, "/coda")) ? NULL : szFilename , &de, CLU_CASE_INSENSITIVE);
	printf("hallo0");
	_HeapFree (szFilename, 0);
	if (res) {
		printf ("namei failed %x", res);
		return res;
	}
	printf("hallo1");
	cn = (struct cnode*)de->d_vnode;
	printf("hallo2");
	fid = cn->c_fid;
	printf("hallo3");
	dput(de);
	printf("hallo4");
	res = venus_pioctl(fid, cmd, &PID, &(PCHAR)p->lpvOutBuffer, 
			(DWORD)p->cbOutBuffer, (PDWORD)p->lpcbBytesReturned);

	return res;
}
