/* find.c -- Windows95 IFS routines for FindFirstFile/FindNextFile
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include "mc.h"

#include <string.h>
#pragma intrinsic (memcpy, memset, memcmp, strlen, strcpy, strcat)


#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG


static struct hndlmisc hfFind = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_HNDLMISC,

	ErrorStub2,				/* FileSeek */
	ErrorStub2,			/* Close */
	ErrorStub2,				/* Commit */
	ErrorStub2,				/* LockFile,*/
	ErrorStub2,				/* FileDateTime, */

	ErrorStub2,				/* NamedPipeRequest */
	ErrorStub2,				/* NamedPipeHandleInfo, */
	ErrorStub2,			/* EnumerateHandle XXX -- should be handled*/
};



#define CodaTestMustMatch(ir_attr, attr)	((((ir_attr & (attr)<<8)	\
											^ ir_attr)		\
										& FILE_ATTRIBUTE_MUSTMATCH) == 0)

int SearchFile (pioreq pir, struct dentry *de)
{
	int res = ERROR_INVALID_FUNCTION;
	LOGDEBUG("--searchfile--", 1);
	LOGDEBUG("--end of searchfile--", 1);
	return pir->ir_error = res;
}

struct FindFile {
	struct cnode *ff_cn;
	PREADDIR	 ff_preaddir;
	DWORD		 ff_attr;
	WORD		 ff_pattern[MAX_PATH + 2];
	WORD		 ff_thisname[LFNMAXNAMELEN +1];
	WORD		 ff_found;
};

static int getnextmatch (struct FindFile *ff, _WIN32_FIND_DATA *wfd)
{
	DWORD matchflag = 0;
	DWORD hasdot = 0;
	DWORD excludeattr;
	DWORD len, attr;
	struct cnode *cn = NULL;
	struct dentry *de = NULL;
	int res = 0, i=1;

	if (ff->ff_attr & (FILE_FLAG_KEEP_CASE | FILE_FLAG_IS_LFN))
		matchflag |= UFLG_NT;
	else
		matchflag |= UFLG_DOS;
	if (ff->ff_attr & FILE_FLAG_WILDCARDS)
		matchflag |= UFLG_META;

	hasdot = (ff->ff_attr & FILE_FLAG_HAS_DOT) != 0;
	excludeattr = (~ff->ff_attr) & (FILE_ATTRIBUTE_HIDDEN | 
		FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_DIRECTORY);
//	mustattr = ((ff->ff_attr & FILE_ATTRIBUTE_MUSTMATCH) >> 8);
	/* special case (thank you, Microsoft!) -- if LABEL attribute is
	   set in the "may-have" category, make it set in "must-have" because
	   that's the interpretation that Microsoft uses */
//	if (ff->ff_attr & FILE_ATTRIBUTE_LABEL)
//		mustattr |= FILE_ATTRIBUTE_LABEL;


	if (!(matchflag & UFLG_META))
		logucode (ff->ff_pattern);

	memset ((PCHAR) wfd, 0, sizeof(_WIN32_FIND_DATA));
	
	while ((res = readdir(ff->ff_preaddir, (PCHAR) ff->ff_thisname, &len)) 
			== 0) {
		if (len == 0) {
			/* end of directory */
			res = ERROR_NO_MORE_FILES;
#ifdef CACHE
			if (ff->ff_cn->c_search == (void *) ff->ff_preaddir){
				LOGDEBUG("getnextmatch: SET C_HAVEDATA", 1);
				ff->ff_cn->c_flags |= C_HAVEDATA;
			}
#endif
			break;
		}
		((PCHAR) ff->ff_thisname)[len] = 0; 

		/* see if name matches */
		BCSToUni (wfd->cFileName, (PCHAR) ff->ff_thisname, len, BCS_OEM);
		wfd->cFileName[len] = 0;
//		UniToUpper (wfd->cFileName, wfd->cFileName, len * 2);

		logucode (wfd->cFileName);
		logucode (ff->ff_pattern);

		if (!IFSMgr_MetaMatch(ff->ff_pattern, wfd->cFileName, matchflag))
			continue;

		if (len == 0 || !memcmp("..", (PCHAR) ff->ff_thisname, len) || !memcmp(".", (PCHAR) ff->ff_thisname, len))
				continue;

		dput(de);
		res = d_lookup(my_fs, (PCHAR) ff->ff_thisname, ff->ff_cn->c_dentry, &de, CLU_CASE_INSENSITIVE);
		if (res) {
			printf ("lookup error %x", res);
			goto error;
		}

		cn = (struct cnode*) de->d_vnode;

		attr = ((cn->c_vattr.va_type == C_VDIR) ? FILE_ATTRIBUTE_DIRECTORY : 0);
		printf ("attr %x excl %x", attr, excludeattr);
//		if ((attr & excludeattr) || 
//			((attr & mustattr) != mustattr))
		if ((attr & excludeattr) || !CodaTestMustMatch(ff->ff_attr, attr))
			continue;
		
		/* got it! */
		break;
	}
	if (!res) {

		res = coda_getattr(cn);
		if (res) {
			printf ("coda_getattr error %x", res);
			goto error;
		}

		if (cn->c_vattr.va_type == C_VLNK) {
			/* look for conflict */
			char str[CODA_MAXPATHLEN];
			int res = 0;
			LOGDEBUG("-- look for conflict --", 10);
			
			res = venus_readlink(cn, str);

			if ( !res && !memcmp(str, "@", 1)){
				cn->c_flags |= C_INCONF;
			}

			if (IS_INCONF(cn)){
				memcpy(str, (PCHAR) de->d_name.name, de->d_name.len);
				memcpy(str + de->d_name.len, ".$cf", 4);
				BCSToUni (wfd->cFileName, str, de->d_name.len + 4, BCS_OEM);
				wfd->cFileName[de->d_name.len + 4] = 0;
			}

		}		

//		memset ((PCHAR) wfd, 0, sizeof(_WIN32_FIND_DATA));
//		BCSToUni (wfd->cFileName, (PCHAR) ff->ff_thisname, len, BCS_OEM);
//		wfd->cFileName[len] = 0;
		wfd->nFileSizeHigh = 0;
		wfd->nFileSizeLow = (unsigned long) cn->c_vattr.va_size;
		wfd->dwFileAttributes = attr;
		wfd->ftLastAccessTime = IFSMgr_NetToWin32Time(cn->c_vattr.va_atime.ts_sec);
		wfd->ftLastWriteTime = IFSMgr_NetToWin32Time(cn->c_vattr.va_mtime.ts_sec);
		wfd->ftCreationTime = IFSMgr_NetToWin32Time(cn->c_vattr.va_ctime.ts_sec);
//		strcpy(l, (PCHAR) ff->ff_thisname);
//		convert_table_add(&ff->ff_contable, &s, l);
//		BCSToUni(wfd->cAlternateFileName, s, strlen(s), BCS_OEM);
//		wfd->cAlternateFileName[strlen(s)] = 0;
//		_HeapFree(s,0);
	//	_HeapFree(l,0);
	}
error:
	dput(de);
	return res;
}

int FindFirstFile (pioreq pir, struct dentry *dirde)
{
	struct FindFile *ff = NULL;
	struct dentry *de = NULL;// *de = NULL;
	struct cnode *cn = NULL, *dircn = NULL;
	char *BCSname;
	_WIN32_FIND_DATA * fd = NULL;
	int res = 0, len = 0;

	CYCLES_ON();

	LOGDEBUG("--findfirstfile--", 1);

	logucode(pir->ir_upath+3);

	if (!dirde) {
		res = ERROR_INVALID_PARAMETER;
		goto error;
	}

	dircn = (struct cnode*) dirde->d_vnode;

	if (!dircn){
		if (is_root(dirde))
			res = coda_root();		
		else{
			res = ERROR_INVALID_PARAMETER;
			goto error;
		}

		if (res) {
			res = ERROR_INVALID_PARAMETER;
			goto error;
		}

		dircn = (struct cnode*) dirde->d_vnode;
	}


	ff = _HeapAllocate(sizeof(struct FindFile), 0);
	if (!ff) {
		LOGDEBUG("out of memory", 10);
		res = ERROR_NOT_ENOUGH_MEMORY;
		goto error;
	}
	ff->ff_preaddir = NULL;
	ff->ff_found = 0;

	BCSname = tofilename(to_BCS(pir->ir_upath+3, ucodelen(pir->ir_upath)-3));
	if (memcmp(BCSname, "*", 1) && memcmp(BCSname, "*.*", 3)){
		res = d_lookup(my_fs, BCSname, dirde, &de, CLU_CASE_INSENSITIVE);
		if (!res) {
			cn = (struct cnode*) de->d_vnode;

			res = coda_getattr(cn);
			if (res) {
				printf ("getattr error %x", res);
				goto error;
			}
		
			memset ((PCHAR) ((_WIN32_FIND_DATA *)pir->ir_data), 0, sizeof(_WIN32_FIND_DATA));
			BCSToUni (((_WIN32_FIND_DATA *)pir->ir_data)->cFileName, BCSname, strlen(BCSname), BCS_OEM);
			((_WIN32_FIND_DATA *)pir->ir_data)->cFileName[strlen(BCSname)] = 0;
			((_WIN32_FIND_DATA *)pir->ir_data)->nFileSizeHigh = 0;
			((_WIN32_FIND_DATA *)pir->ir_data)->nFileSizeLow = (unsigned long) cn->c_vattr.va_size;
			((_WIN32_FIND_DATA *)pir->ir_data)->dwFileAttributes = ((cn->c_vattr.va_type == C_VDIR) ? FILE_ATTRIBUTE_DIRECTORY : 0);
			((_WIN32_FIND_DATA *)pir->ir_data)->ftLastAccessTime = IFSMgr_NetToWin32Time(cn->c_vattr.va_atime.ts_sec);
			((_WIN32_FIND_DATA *)pir->ir_data)->ftLastWriteTime = IFSMgr_NetToWin32Time(cn->c_vattr.va_mtime.ts_sec);
			((_WIN32_FIND_DATA *)pir->ir_data)->ftCreationTime = IFSMgr_NetToWin32Time(cn->c_vattr.va_ctime.ts_sec);
			
			ff->ff_attr = pir->ir_attr;
			ff->ff_cn   = dircn;
			ff->ff_found = 1;
		} 
	} else {

		res = opendir(dircn, &ff->ff_preaddir);
		if (res) {
			printf ("opendir failed %x", res);
			goto error;
		}

		/* copy the last path element into the pattern field -- 
		here we exploit the fact that following the last path element is
		a zero short, which we can use to delimit our pattern */

		memcpy ((PCHAR) ff->ff_pattern, 
			(PCHAR) (&IFSLastElement(pir->ir_ppath)->pe_unichars),
			IFSLastElement(pir->ir_ppath)->pe_length);

		/* upcase the pattern */
		UniToUpper (ff->ff_pattern, ff->ff_pattern, 
				IFSLastElement(pir->ir_ppath)->pe_length);

		ff->ff_attr = pir->ir_attr;
		ff->ff_cn   = dircn;

		logucode (ff->ff_pattern);
		if (ff->ff_attr & 0x1000) ff->ff_attr &= 0xffff0fff;
		printf ("specified attr %x", ff->ff_attr);

		if (pir->ir_attr & FILE_ATTRIBUTE_LABEL){
			fd = (_WIN32_FIND_DATA *) pir->ir_data;
			LOGDEBUG("--findfirstfile volumelabel--", 10);
			fd->dwFileAttributes = FILE_ATTRIBUTE_LABEL;
			memcpy(fd->cFileName, coda_volume_name, 10);
			res = 0;		
		}else {
			res = getnextmatch(ff, (_WIN32_FIND_DATA *) pir->ir_data);
			if (res == ERROR_NO_MORE_FILES)
				res = ERROR_FILE_NOT_FOUND;
		}
	}

	if (res)
		pir->ir_fh = NULL;
	else
		pir->ir_fh = (PVOID) ff;

	SetHandleFunc (pir, FindNextFile, ErrorStub2, &hfFind);

error:

	dput(de);

	if (res) {
		if (ff && ff->ff_preaddir)
			closedir(ff->ff_preaddir);
		if (ff) _HeapFree (ff, 0);
		dput (dirde);
	}

	printf ("--end of findfirstfile: %x--", res);

	CYCLES_OFF(FINDFIRSTFILE);

	return pir->ir_error = res;
}

int FindNextFile (pioreq pir)
{
	struct FindFile *ff = (struct FindFile *) pir->ir_fh;
	int res = 0;

	CYCLES_ON();

	LOGDEBUG("--findnextfile--", 1);

	if (!ff->ff_found)
		res = getnextmatch (ff, (_WIN32_FIND_DATA *) pir->ir_data);
	else res = ERROR_NO_MORE_FILES;

	printf ("--end of findnext: %x--", res);

	CYCLES_OFF(FINDNEXTFILE);

	return pir->ir_error = res;
}


int FindFileClose (pioreq pir)
{
	struct FindFile *ff = (struct FindFile *) pir->ir_fh;
	int res =0;

	LOGDEBUG("--findclose--", 1);

	if (ff){		
		if (ff->ff_preaddir) {
			closedir(ff->ff_preaddir);
			_HeapFree(ff->ff_preaddir, 0);
		}
		dput(ff->ff_cn->c_dentry);
		_HeapFree(ff, 0);
	}

	printf ("--end of findclose: %x--", res);

	return pir->ir_error = res;
}

int EnumerateFind (pioreq pir)
{
	LOGDEBUG("--enumeratefind--", 1);
	LOGDEBUG("--end of enumeratefind--", 1);
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}
