/* mc.h -- minicache headers
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/


#define WANTVXDWRAPS
#include <basedef.h>
#include <vmm.h>
#include <ios.h>
#include <ifs.h>
#include <vwin32.h>
#include <regdef.h>
#include <vxdwraps.h>

#include <winerror.h>

#include "ifsmgr.h"
#include "iosdcls.h"

#include "ipc.h"

#include "dllist.h"

#include "cnode.h"
#include "vdcache.h"

#include "errno.h"

#pragma warning(disable: 4102)


/* init.c */

extern int inUseCount;
BOOL OnDeviceInit();
extern DWORD Debuglevel;

/* devio.c */

DWORD OnW32DeviceIoControl(HVM, PDIOCPARAMETERS);

extern int calls[10];
extern HANDLE hHandle;
extern struct filesystem *my_fs;

/* vol.c */

extern struct dentry *root;
USHORT	coda_volume_name[]; 

int ConnectNetResource (pioreq, pFilesystem fs);
int DisconnectResource (pioreq, pFilesystem fs);

int ErrorStub (pioreq, struct dentry *);
int ErrorStub2 (pioreq);

int FileAttributes (pioreq, struct dentry *);
int DeleteFile (pioreq, struct dentry *);
int FlushVolume (pioreq);
int Dir (pioreq, struct dentry *);
int GetDiskInfo (pioreq);
int QueryResourceInfo (pioreq);
int RenameFile (pioreq, struct dentry *, struct dentry *);
int UNCNamedPipeRequest (pioreq);

extern UINT uiNetId;
extern UINT uiProviderId;

/* misc.c */

int errno2res (int errno);

void logucode (PWORD);
void logmsg (char *);
void printf (char *, ...);

struct request *NewRequest(int);
void FreeRequest (struct request *);

void fmt_dword (char *, DWORD);
void print_pir (pioreq);
void print_ppath (ParsedPath *);
void print_cnode (struct cnode *cnode);
void print_fid(ViceFid *fid);
int is_root(struct dentry *);
PUSHORT to_ucode(PCHAR name);
PCHAR to_BCS(PUSHORT name, int length);
int ucodelen(PWORD pu);
char *parent_path(char *path);
char *tofilename(char *path);
char *tolower(char *tar, const unsigned char *src);
char *memicpy(char *tar, char *src, int len);
void logcachemsg (char *buffer);
void logcache_lrulist(struct dllist_head *head);
void printf_cache(char *, ...);
void logcycles (char *buffer);
void printcyclesf(char *fmt, ...);

#ifdef DEBUG
#define LOGDEBUG(str, level)\
do{\
	if (level <= Debuglevel) logmsg(str);\
}while(0)	
#else
#define LOGDEBUG(str, level)	0
#endif


/* file.c */

int OpenFile (pioreq, struct dentry *);
int ReadFile (pioreq);
int WriteFile (pioreq);
int FileSeek (pioreq);
int CommitFile (pioreq);
int CloseFile (pioreq);
int LockFile (pioreq);
int FileDateTime (pioreq);
int NamedPipeRequest (pioreq);
int NamedPipeHandleInfo (pioreq);
int EnumerateHandle (pioreq);

/* find.c */

int SearchFile (pioreq, struct dentry *);
int FindFirstFile (pioreq, struct dentry *);
int FindNextFile (pioreq);
int FindFileClose (pioreq);
int EnumerateFind (pioreq);

/* ipc.c */

int initialize_ipc();
int process_rpc (struct request *);
int insert_upcall (struct request *req);
struct request *get_pending();
struct request *find_request(long);
void signal_request(struct request *);
int set_R3_event(DWORD);
void clear_resource_requests(DWORD rh);
void heartbeat();
DWORD do_poll(PVOID, PDWORD);
void do_close(PVOID);

/* readdir.c */

#define READDIR_BUFSIZE 2048

typedef struct READDIR {
	ViceFid		VFid;
	DWORD		hdl;
	DWORD		filesize;
	DWORD		pos;
	DWORD		bufstart;
	DWORD		bufend;
	WORD		hidden;
	char		buf[READDIR_BUFSIZE];
} READDIR, *PREADDIR;

int opendir (struct cnode *cn, PREADDIR *p);
int readdir (PREADDIR p, PCHAR q, PDWORD plen);
void closedir(PREADDIR p);

/* cfs.c */

int coda_root();
int openVFid (struct cnode *cn, int flags, WORD access, WORD attr, BYTE action, 
			  struct request *inreq, PDWORD phdl, PDWORD ptakenaction,
			  PWORD pattr, PWORD hidden);
int coda_getattr (struct cnode *);
int coda_setattr (struct cnode *);
int coda_namei(char *upp, struct dentry **de, int flags);

/* upcall.c */

int venus_remove(struct cnode *, pioreq );
int venus_mkdir(pioreq, struct dentry **);
int venus_rmdir(struct cnode *, pioreq );
int venus_root(ViceFid *);
int venus_close(ViceFid, DWORD );
int venus_create(pioreq , struct dentry **);
int venus_pioctl(ViceFid, unsigned long cmd,
				 struct PioctlData *, PCHAR *, DWORD, PDWORD);
int venus_lookup(struct dentry *, struct cstr *, struct dentry **, int);
int venus_open_by_path(ViceFid, char*, struct request *, int);
int venus_getattr(struct cnode *cn);
int venus_setattr(struct cnode *cn);
int venus_rename(pioreq pir, struct dentry *, struct dentry *);
int venus_readlink(struct cnode *cn, PCHAR str);
int coda_downcall(union outputArgs *out);

/* cache.c */

/* flags */
#define C_VATTR       0x1   /* Validity of vattr in inode */
#define C_PURGE       0x8
#define C_ZAPDIR      0x10
#define C_DYING       0x4   /* from venus (which died) */
#define C_INITED      0x20
#define C_FLUSH       0x2   /* used after a flush */
#define C_INCONF	  0x40  /* in conflict */	
#define C_CONTAINER	  0x100	
#define C_HAVEDATA	  0x200	

#define VALID_CON(cp)	((cp->c_flags) & C_CONTAINER)
#define VALID_VATTR(cp)	((cp->c_flags) & C_VATTR)
#define IS_DYING(cp)	((cp->c_flags) & C_DYING)
#define IS_FLUSHED(cp)	((cp->c_flags) & C_FLUSH)
#define IS_OLD(cp)		(IS_DYING((cp)) || IS_FLUSHED((cp)))
#define IS_NEW(cp)		((cp->c_flags) & C_INITED)	
#define IS_VALID(cp)	(!IS_DYING((cp)) && !IS_FLUSHED((cp))
#define IS_INCONF(cp)	((cp->c_flags) & C_INCONF)
#define HAVEDATA(cp)	((cp->c_flags) & C_HAVEDATA)



/* cache functions */
void init_cache();
struct cnode *cnode_alloc();
void cnode_free(struct cnode *);
struct cnode *coda_lookupFid(struct ViceFid *fid);
int coda_revalidate(struct dentry *de);
void coda_unflag_cnode(struct cnode *cn, int flag);
void coda_flag_cnode(struct cnode *cn, int flag);
void coda_flag_children(struct cnode *parent, int flag);
int coda_comparefid(struct ViceFid *f1, struct ViceFid *f2);
int coda_hashfid(struct ViceFid *fid);	
int coda_delete(struct dentry *de);
int coda_iput(struct dentry *de);
void coda_instantiate(struct dentry *de, struct cnode *cn, struct ViceFid *fid);
int coda_ifree(struct dentry *de);
int coda_lookup(struct dentry **de, struct cstr *str, int flags);
void coda_replace_fid(struct cnode * cn, ViceFid *oldfid, ViceFid *newfid);

#ifndef CYCLES
#define CYCLES
#define rdtsc __asm _emit 0x0f __asm _emit 0x31

#define getstamp(tick)  _asm{\
							rdtsc\
						}\
						_asm {\
							mov DWORD PTR tick, eax\
						}
					
#define CYCLES_ON()\
do {\
	unsigned long be, en;\
	char *str;\
	getstamp(be)\

#define CYCLES_OFF(text)\
	getstamp(en);\
	str = "CYCLES: " #text ": %x";\
	printcyclesf(str, en - be);\
} while (0)
#endif


static struct dentry_operations coda_dentry_operations = {
	coda_revalidate,
	coda_delete,
	NULL,
	coda_iput,
	coda_ifree
};

static struct file_operations coda_file_operations = {
	FileSeek,
	ReadFile,
	WriteFile,
	OpenFile,
	CloseFile,
	NULL,
	NULL,
	FileDateTime,
	NULL,
	NULL,
	NULL
};


static struct vnode_operations coda_vnode_operations = {
	&coda_file_operations,
	NULL,
	coda_lookup,
	DeleteFile,
	Dir,
	RenameFile,
	NULL,
	FileAttributes,
	GetDiskInfo,
	NULL,
	QueryResourceInfo,
	NULL,
	NULL,
	NULL,
	NULL,
	FindFirstFile,
	FindNextFile,
	FindFileClose,
	NULL
};
