/* readdir.c -- routines for reading Coda directory files
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include "MC.h"

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

#include <string.h>
#pragma intrinsic (memcpy, strcat, strlen, strcpy)

int do_potemkin_readdir(PREADDIR p){
	int res = 0;

	if (p->hdl) {// 0 means not first time, -1 means first time
		res = venus_open_by_path(p->VFid, (char *)&p->buf, 0, C_O_READ);
		if (!res)
			p->hdl=0; // indicates, that venus has opened the file by path
			logmsg("got path");
			strcpy((char*)&p->buf[strlen(p->buf)-1], "*.*");
			logmsg(p->buf);
			res = R0_FindFirstFile(0, p->buf, (_WIN32_FIND_DATA*)&(p->bufstart), &p->pos); 
			//pos has find handle for FindNextFile and bufstart contains adress of Find Data
			if (res) printf("error in R0_FindFirstFile %x", res);
	}
	else{
		res = R0_FindNextFile(p->pos, (_WIN32_FIND_DATA*)&(p->bufstart));
		if (res == ERROR_NO_MORE_FILES){
			return res;
		}
	}

	if (res){
		logmsg("File Corrupt");
		return ERROR_FILE_CORRUPT;
	}

	return 0;
}
 
int opendir (struct cnode *cn, PREADDIR *p)
{
	int res =0;

	LOGDEBUG("--opendir--", 1);

	*p = _HeapAllocate (sizeof(READDIR), 0);

	if (!*p)
		return ERROR_NOT_ENOUGH_MEMORY;

	(*p)->VFid= cn->c_fid;
	(*p)->hdl = 0;
	(*p)->pos = 0;
	(*p)->bufstart = 0;
	(*p)->bufend = 0;
	(*p)->hidden = 0;

	print_fid(&cn->c_fid);
	if (!HAVEDATA(cn)){
		res = openVFid (cn, C_O_READ, 
			(WORD) ACCESS_READONLY, (WORD) 0, (BYTE) ACTION_OPENEXISTING, 
			NULL, &(*p)->hdl, NULL, NULL, NULL);
		if (res) {
			printf ("open dir failed : %x", res);
			_HeapFree (*p, 0);
			*p = NULL;
			return res;
		}

		res = R0_GetFileSize((*p)->hdl, &(*p)->filesize);
		if (res) {
			printf ("get file size failed: %x", res);
			closedir(*p);
			_HeapFree (*p, 0);
			*p = NULL;
		} else 		
			cn->c_search = (void *) *p;

	} else {
		(*p)->bufstart = (DWORD) cn;
	}

	LOGDEBUG("--end of opendir--", 1);

	return res;
}

static int getbytes (PREADDIR preaddir, DWORD n, char *p) 
{
	DWORD len, len2, res, readlen;
	printf ("getbytes[%x]: pos %x buf [%x,%x) len %x", (DWORD) preaddir,
		preaddir->pos, preaddir->bufstart, preaddir->bufend, n);
	len = n;
	if (len > preaddir->bufend - preaddir->pos)
		len = preaddir->bufend - preaddir->pos;
	memcpy (p, preaddir->buf + preaddir->pos - preaddir->bufstart, len);

	p += len;
	preaddir->pos += len;
	printf ("getbytes: p %x, pos %x ", p, preaddir->pos);

	if (preaddir->pos < preaddir->bufend){
		printf ("getbytes: len %x ", len);
		return len;
	}

	readlen = READDIR_BUFSIZE;
	if (readlen > preaddir->filesize - preaddir->pos)
		readlen = preaddir->filesize - preaddir->pos;

	if (readlen)
		res = R0_ReadFile(TRUE,
			preaddir->hdl,
			readlen,
			preaddir->pos,
			preaddir->buf,
			(PDWORD) &readlen);
	else
		res = 0;

	preaddir->bufstart = preaddir->pos;
	preaddir->bufend = preaddir->bufstart + readlen;
	printf ("getbytes: res %x readlen %x", res, readlen);
	if (res) {
		printf ("getbytes: res %x ", res);
		return len;
	}
	len2 = n - len;
	if (len2 > readlen)
		len2 = readlen;

	memcpy (p, preaddir->buf, len2);
	preaddir->pos += len2;
	printf ("getbytes: return %x ", len+len2);
	return len + len2;
}

static void skip(PREADDIR preaddir, int n)
{
	preaddir->pos += n;
}

#define GETBYTES(preaddir, n, p) ((preaddir->pos + n < preaddir->bufend) ? \
	memcpy(p, preaddir->buf + (preaddir->pos - preaddir->bufstart), n), \
		preaddir->pos += n, n : \
	getbytes(preaddir, n, p))


int readdir (PREADDIR p, PCHAR q, PDWORD plen)
{
	int len = 0, res = 0;
	struct venus_dirent deh;
	int deHdrLen = (char *) (&deh.d_name) - (char *) (&deh);

	CYCLES_ON();

start:

	if (!p->hdl){
		/* read out of cache */
		struct dentry *dirde = ((struct cnode *) p->bufstart)->c_dentry;
		struct dentry *de = NULL;
		LOGDEBUG("read dir out of cache", 10);
		de = d_getnextchild(dirde, (p->pos) ? (struct dllist_head *) p->pos : NULL);
		if (!de){
			*plen = 0;
		} else {
			p->pos = (DWORD) de->d_child.next;
			*plen = de->d_name.len;
			memcpy(q, de->d_name.name, de->d_name.len);
			q[de->d_name.len] = '\0';
			LOGDEBUG(q, 100);
		}
		return 0;
	}

	len = GETBYTES(p, deHdrLen, (char *) &deh);
	printf ("readdir: pos %x bufstart %x bufend %x", p->pos, p->bufstart, p->bufend);
	LOGDEBUG("readdir: out first GETBYTES", 10);

	if (/*deh.d_namlen == 0 && */deh.d_reclen+p->pos-len >= p->bufend && p->pos == p->filesize) {
		/* end of directory */
		printf ("readdir: reclen %x namlen %x fileno %x", deh.d_reclen, deh.d_namlen, deh.d_fileno);
		LOGDEBUG("readdir: > bufend", 10);
		*plen = 0;
		res = 0;
		goto out;
	}
	
	if (len	!= deHdrLen) {
		LOGDEBUG("error getting dirent header", 10);
		res = ERROR_FILE_CORRUPT;
		goto out;
	}

	printf ("getbytes: d_namelen %x ", deh.d_namlen);
	if (GETBYTES(p, deh.d_namlen, q) != deh.d_namlen) {
		LOGDEBUG("error getting dirent name", 10);
		res = ERROR_FILE_CORRUPT;
		goto out;
	}
	printf ("readdir: pos %x bufstart %x bufend %x", p->pos, p->bufstart, p->bufend);

	*plen = deh.d_namlen;
	q[deh.d_namlen] = '\0';

	skip (p, deh.d_reclen - deh.d_namlen - deHdrLen);

out:

	CYCLES_OFF(READDIR);

	return 0;
}

void closedir(PREADDIR p)
{
	int res = 0;

	LOGDEBUG("--closedir--", 1);
	printf("closedir: close handle: %x", (p) ? p->hdl : NULL);
	if (p){
		if (p->hdl && !p->hidden) {
			res = venus_close(p->VFid, C_O_READ); 
			LOGDEBUG("DIR WAS CLOSED", 100);
			print_fid(&p->VFid);
		}
		if (p->hdl) res = R0_CloseFile(p->hdl);
		if (res) {
			printf ("R0_CloseFile Error in closedir: %x", res);
			goto error;
		}
	}

error:

	printf("--end of closedir res: %x--", res);
}