/* upcalls.c
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/
#include "mc.h"

#include "string.h"

#pragma intrinsic(strlen, strcpy, memcmp, memset, memcpy)

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

#define INSIZE(tag) sizeof(struct coda_ ## tag ## _in)
#define OUTSIZE(tag) sizeof(struct coda_ ## tag ## _out)
#define SIZE(tag)  max(INSIZE(tag), OUTSIZE(tag))




/* put the current process credentials in the cred */
void coda_load_creds(struct coda_cred *cred)
{
        cred->cr_uid = 500;
        cred->cr_euid = 500;
        cred->cr_suid = 500;
        cred->cr_fsuid = 500;

        cred->cr_groupid = 500;
        cred->cr_egid = 500;
        cred->cr_sgid = 500;
        cred->cr_fsgid = 500;
}

#define UPARG(op, length)\
do {\
	req = NewRequest(length);\
	if (!req) {\
		res = ERROR_NOT_ENOUGH_MEMORY;\
		return res;\
	}\
    req->r.in.ih.opcode = (op);\
	coda_load_creds(&(req->r.in.ih.cred));\
} while (0)

int venus_remove(struct cnode *cn, pioreq pir){
	struct request *req = NULL;
	PCHAR p;
	int res = 0, len, i;

	for (len = 0; pir->ir_uFName[len]; len++)
		;

	UPARG(CODA_REMOVE, len+1);

	req->r.in.coda_remove.VFid = cn->c_fid;
	p = (char *) (&req->r.in.coda_remove.name + 1);
	req->r.in.coda_remove.name = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	res = process_rpc(req);
	if (res) {
		printf ("remove process_rpc failed %x", res);
		if (req) FreeRequest (req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("remove rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest (req);		
		return res;
	}
	return res;
}

int venus_mkdir(pioreq pir, struct dentry **de){
	struct request *req = NULL;
	struct cnode *cn = NULL;
	int len, i;
	int res = ERROR_INVALID_FUNCTION;
	PCHAR p;

	for (len = 0; pir->ir_uFName[len]; len++)
			;
	UPARG(CODA_MKDIR, len+1);

	cn = (struct cnode*) (*de)->d_parent->d_vnode;

	req->r.in.coda_mkdir.VFid = cn->c_fid;
	req->r.in.coda_mkdir.attr.va_mode = 0666;
	p = (char *) (&req->r.in.coda_mkdir.name + 1);
	req->r.in.coda_mkdir.name = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	res = process_rpc(req);
	if (res) {
		printf ("mkdir process_rpc failed %x", res);
		if (req) FreeRequest (req);		
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("mkdir rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest (req);		
		return res;
	}

	cn = cnode_alloc();
	coda_instantiate(*de, cn, &req->r.out.coda_mkdir.VFid);
//	print_cnode(*cn);
	cn->c_vattr = req->r.out.coda_mkdir.attr;

	LOGDEBUG("--end of venus_mkdir--", 1);

	return res;
}

int venus_rmdir(struct cnode *cn, pioreq pir){
	struct request *req = NULL;
	int len, i;
	int res = ERROR_INVALID_FUNCTION;
	PCHAR p;

	for (len = 0; pir->ir_uFName[len]; len++)
		;

	UPARG(CODA_RMDIR, len+1);

	req->r.in.coda_rmdir.VFid = cn->c_fid;
	p = (char *) (&req->r.in.coda_rmdir.name + 1);
	req->r.in.coda_rmdir.name = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	res = process_rpc(req);
	if (res) {
		printf ("rmdir process_rpc failed %x", res);
		if (req) FreeRequest (req);		
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("rmdir rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest (req);		
		return res;
	}
	return res;
}

int venus_root(ViceFid *fid){
	struct request *req = NULL;
	int res = 0;

	UPARG(CODA_ROOT, 0);

	res = process_rpc(req);
	if (res) {
		LOGDEBUG("CODA_ROOT rpc failed", 10);
		if (req) FreeRequest (req);		
		return res;
	}
	if (req->r.out.oh.result) {
		LOGDEBUG("CODA_ROOT nonzero venus result", 10);
		res = errno2res (req->r.out.oh.result);
		if (req) FreeRequest (req);		
		return res;
	}

	printf ("ROOT VFID: %x %x %x", req->r.out.coda_root.VFid.Volume,
		req->r.out.coda_root.VFid.Vnode, req->r.out.coda_root.VFid.Unique);

	fid->Volume = req->r.out.coda_root.VFid.Volume;
	fid->Vnode = req->r.out.coda_root.VFid.Vnode;
	fid->Unique = req->r.out.coda_root.VFid.Unique;

	return res;
}

int venus_close(ViceFid VFid, DWORD flags){
	struct request *req;
	int res = ERROR_ACCESS_DENIED;
	
	UPARG(CODA_CLOSE, 0);

	req->r.in.coda_close.VFid = VFid;
	req->r.in.coda_close.flags = flags;

	CYCLES_ON();
	res = process_rpc(req);
	CYCLES_OFF(CLOSE);

	if (res) 
		printf ("CODA_CLOSE process_rpc failed: %x", res);

	res = req->r.out.oh.result;
	if (res) 
		printf ("CODA_CLOSE rpc failed: %x", res);

	if (req) 
		FreeRequest(req);

	return res;
}

int venus_create(pioreq pir, struct dentry **de){
	struct request *req = NULL;
	struct cnode *cn = NULL;
	PCHAR p;
	int i, res;

	UPARG(CODA_CREATE, MAX_PATH+1);

	if (pir->ir_options & ACTION_TRUNCATE)
		req->r.in.coda_create.attr.va_size = 0;

	cn = (struct cnode*) (*de)->d_parent->d_vnode;

	req->r.in.coda_create.VFid = cn->c_fid;
	req->r.in.coda_create.attr.va_mode = 0666;
	req->r.in.coda_create.excl = (pir->ir_options & ACTION_TRUNCATE) ? 0 : 1;
	req->r.in.coda_create.mode = 0666;
	p = (char *) (&req->r.in.coda_create.name + 1);
	req->r.in.coda_create.name = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	LOGDEBUG((char *) (&req->r.in.coda_create.name + 1), 10);
	
	res = process_rpc(req);
	if (res) {
		logmsg ("create process_rpc failed");
		if (req) FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		LOGDEBUG("create failed", 10);
		res = errno2res(res);
		if (req) FreeRequest(req);
		return res;
	}

	cn = cnode_alloc();
	coda_instantiate(*de, cn, &req->r.out.coda_create.VFid);
	cn->c_vattr = req->r.out.coda_create.attr;
//	print_cnode(*cn);
	LOGDEBUG("--end of venus_create--", 1);

	return res;
}

typedef unsigned char RPC2_EncryptionKey[8];
typedef unsigned char EncryptedSecretToken[56];
typedef struct{
	long AuthHandle;
	RPC2_EncryptionKey HandShakeKey;
	long ViceId;
	long BeginTimestamp;
	long EndTimestamp;
} ClearToken;
typedef struct {
    int			    sTokenSize;
    EncryptedSecretToken    stoken;
    int			    cTokenSize;
    ClearToken		    ctoken;
} venusbuff;


int venus_pioctl(ViceFid dirfid, unsigned long cmd, struct PioctlData *pPID, 
				 PCHAR *out, DWORD out_size, PDWORD ret_count){
	
	struct request *req = NULL;
	DWORD offset, size, res = 0, iocsize;
//	venusbuff *vb;

	size = pPID->vi.in_size;
	if (out_size > size)
		size = out_size;

	UPARG(CODA_IOCTL, size);
	offset = INSIZE(ioctl);

    /* the cmd field was mutated by increasing its size field to
     * reflect the path and follow args. We need to subtract that
     * out before sending the command to Venus.  */
    req->r.in.coda_ioctl.cmd = (cmd & ~(PIOCPARM_MASK << 16));	
    iocsize = ((cmd >> 16) & PIOCPARM_MASK) - sizeof(char *) - sizeof(int);
    req->r.in.coda_ioctl.cmd |= (iocsize & PIOCPARM_MASK) <<	16;	

	req->r.in.coda_ioctl.VFid = dirfid;
	req->r.in.coda_ioctl.len = pPID->vi.in_size;
	req->r.in.coda_ioctl.data = (PCHAR)offset;

//	vb = (venusbuff *)pPID->vi.in;
//	printf ("cmd: %x , venusbuff: ctoken.EndTimestamp: %x, size: %x, offset: %x", req->r.in.coda_ioctl.cmd, vb->ctoken.EndTimestamp, size, offset);
	memcpy ((PCHAR)&(req->r.in) + offset, pPID->vi.in, pPID->vi.in_size);

	LOGDEBUG("--Pioctl: before process_rpc--", 10);
	res = process_rpc(req);
	if (res) {
		printf ("CODA_IOCTL rpc failed %x", res);
		if (req) FreeRequest(req);
		return res;
	}
	if (req->r.out.oh.result) {
		printf ("CODA_IOCTL nonzero venus result %x", req->r.out.oh.result);
		res = req->r.out.oh.result;
	}

	if (out_size != 0){
		if ((unsigned)req->r.out.coda_ioctl.len > out_size) {
			LOGDEBUG("--Pioctl: output bigger than buffer--", 10);
			if (req) FreeRequest(req);
			return ERROR_INVALID_PARAMETER;

		}
		memcpy(*out, ((PCHAR) &req->r.out) + (DWORD) req->r.out.coda_ioctl.data,
			req->r.out.coda_ioctl.len);
		*ret_count = req->r.out.coda_ioctl.len;
	}

	if (req) FreeRequest(req);
	return res;
}

int venus_readlink(struct cnode *cn, PCHAR str){
	struct request *req = NULL;
	int offset, res = 0;

	UPARG(CODA_READLINK, CODA_MAXPATHLEN);
	offset = INSIZE(readlink);

	req->r.in.coda_readlink.VFid = cn->c_fid;
	res = process_rpc(req);
	if (res) {
		LOGDEBUG("readlink process_rpc failed", 10);
		if (req) FreeRequest(req);
		return res;
	}

	res = req->r.out.oh.result;
	if (res) {
		res = errno2res (res);
		printf ("readlink error %x", res);
		if (req) FreeRequest(req);
		return res;
	}

	memcpy(str, ((PCHAR) &req->r.out) + (DWORD) req->r.out.coda_readlink.data, req->r.out.coda_readlink.count);
	printf("--venus_readlink count %x --", req->r.out.coda_readlink.count);
	str[req->r.out.coda_readlink.count] = '\0';
	LOGDEBUG(str, 100);
	if (req) FreeRequest(req);
	return res;

}

int venus_lookup(struct dentry *dirde, struct cstr *name, struct dentry **de,
				 int flags){
	struct request *req = NULL;
	int offset, res = 0, len = 0;
	struct cnode *lcn = NULL, *cn = NULL;
	char *p;

	len = name->len;
	if (len >= 4 && name->name[len-3] == '$' && name->name[len-2] == 'c' && 
		name->name[len-1] == 'f' && name->name[len-4] =='.')
		len = len - 4;

	UPARG(CODA_LOOKUP, len+1);

	cn = (struct cnode*) dirde->d_vnode;

	offset = INSIZE(lookup);
	req->r.in.coda_lookup.VFid = cn->c_fid;
	req->r.in.coda_lookup.name = offset;
	req->r.in.coda_lookup.flags = flags;
	/* send Venus a null terminated string */
	memcpy((char *)(&(req->r.in)) + offset, name->name, len);
	*((char *)(&(req->r.in)) + offset + len) = '\0';
	/* first, if the request flags allow it, we try to open existing file */
	LOGDEBUG((char *) (&(req->r.in)) + offset, 10);

	res = process_rpc(req);
	if (res) {
		LOGDEBUG("lookup process_rpc failed", 10);
		if (req) FreeRequest(req);
		return res;
	}

	res = req->r.out.oh.result;
	if (res) {
		res = errno2res (res);
		printf ("lookup error %x", res);
		if (req) FreeRequest(req);
		return res;
	}

	p = (char *) (&(req->r.in)) + offset;
	len = strlen(p);
	if (len != 0 && !(len == 1 && !memcmp(p, ".", 1)) && !(len == 2 && !memcmp(p, "..", 2))) {
		lcn = coda_lookupFid(&req->r.out.coda_lookup.VFid);
		if (!lcn || lcn == (struct cnode *) (*de)->d_vnode){
			if (!lcn) lcn = cnode_alloc();
			coda_instantiate(*de, lcn, &req->r.out.coda_lookup.VFid);
			lcn->c_vattr.va_type = req->r.out.coda_lookup.vtype;
		}
		else {
			char *newname = NULL;
			int newlen = len;
			LOGDEBUG("-- BAAANG WARNING TWO CNODES ONE FID MUST BE INCONFLICT--", 1);
			/* must be inconsistent cnode */
			d_drop(*de);
			dput(*de);
			*de = lcn->c_dentry;
			lcn->c_flags &= ~C_INCONF;
		}

	} else {
		LOGDEBUG("--- not valid name ---", 10);
		res = ERROR_FILE_NOT_FOUND;
	}

	if (req) FreeRequest(req);
	return res;
	
}

int venus_open_by_path(ViceFid VFid, char * path, struct request *inreq, int flags){
	struct request *req = inreq;
	int res;

	UPARG(CODA_OPEN_BY_PATH, MAX_PATH+1);

	printf ("VFID: %x %x %x", VFid.Volume,VFid.Vnode, VFid.Unique);

	req->r.in.coda_open_by_path.VFid = VFid;
	req->r.in.coda_open_by_path.flags = flags;

	CYCLES_ON();
	res = process_rpc(req);
	CYCLES_OFF(OPENBYPATH);
	if (res) {
		printf ("CODA_OPEN_BY_PATH process_rpc failed");
		if (!inreq)	FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("CODA_OPEN_BY_PATH rpc failed");
		res = errno2res (res);
		if (!inreq)	FreeRequest(req);
		return res;
	}

	/* get path */
	strcpy (path, ((char *) &req->r.out) + req->r.out.coda_open_by_path.path);
	LOGDEBUG(path, 10);

	if (!inreq) FreeRequest(req);
	return res;
}

int venus_getattr(struct cnode *cn){
	struct request *req = NewRequest(0);
	int res;

	UPARG(CODA_GETATTR, 0);

	req->r.in.coda_getattr.VFid.Volume = cn->c_fid.Volume;
	req->r.in.coda_getattr.VFid.Vnode = cn->c_fid.Vnode;
	req->r.in.coda_getattr.VFid.Unique = cn->c_fid.Unique;
	print_cnode(cn);

	res = process_rpc(req);
	if (res) {
		printf ("getattr process_rpc failed %x", res);
		if (req) FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("getattr rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest(req);
		return res;
	}

	cn->c_vattr = req->r.out.coda_getattr.attr;

error:
	if (req) FreeRequest(req);
	return res;
}

int venus_setattr(struct cnode *cn){
	struct request *req = NULL;
	int res = 0;
	struct coda_vattr vattr;

	vattr = cn->c_vattr;
	vattr.va_nlink     = -1;
	vattr.va_blocksize = -1;
	vattr.va_rdev      = -1;
	vattr.va_flags     = -1;
	vattr.va_fileid    = -1;
	vattr.va_bytes     = -1;
	vattr.va_size      = -1;

	UPARG(CODA_SETATTR, 0);

	req->r.in.coda_setattr.VFid = cn->c_fid;
	req->r.in.coda_setattr.attr = vattr;

	res = process_rpc(req);
	if (res) {
		printf ("setattr process_rpc failed %x", res);
		if (req) FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("setattr rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest(req);
		return res;
	}
	if (req) FreeRequest(req);
	return res;
}

int venus_rename(pioreq pir, struct dentry *de1, struct dentry *de2){

	struct request *req = NULL;
	int res = 0, i, len;
	PathElement* old_name = NULL;
	PCHAR p;
	struct cnode *cn1 = NULL, *cn2 = NULL;

	old_name = IFSLastElement( pir->ir_ppath );

	for (len = 0; pir->ir_uFName[len]; len++)
		;

	UPARG(CODA_RENAME, len+1);

	cn1 = (struct cnode *) de1->d_parent->d_vnode;
	cn2 = (struct cnode *) de2->d_parent->d_vnode;

	req->r.in.coda_rename.sourceFid = cn1->c_fid;
	req->r.in.coda_rename.destFid = cn2->c_fid;

	p = (char *) (&req->r.in.coda_rename.destname + 1);
	req->r.in.coda_rename.destname = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	p += i+1;
	req->r.in.coda_rename.srcname = (p - (char *) &req->r.in);
	for (i = 0; old_name->pe_unichars[i]; i++) {
		p[i] = (CHAR) old_name->pe_unichars[i];
		if (p[i] >= 'A' && p[i] <= 'Z')
			p[i] += 'a' - 'A';
	}
	p[i] = '\0';

	res = process_rpc(req);
	if (res) {
		printf ("rename process_rpc failed %x", res);
		if (req) FreeRequest (req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("rename rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest (req);
		return res;
	}

	if (req) FreeRequest (req);
	return res;
}

int coda_downcall(union outputArgs *out){
	struct cnode *cn = NULL;
	struct dentry *de = NULL;
	ViceFid *fid; 
	switch(out->oh.opcode){
	case CODA_FLUSH:
		LOGDEBUG("---CODA_FLUSH---", 1);
		d_shrink_cache();
		break;
	case CODA_PURGEFID:
		fid = &out->coda_purgefid.CodaFid;
		LOGDEBUG("---CODA_PURGEFID---", 1);

		cn = coda_lookupFid(fid);

		if (cn) {
			de = cn->c_dentry;
			coda_flag_children(cn, C_PURGE);
			coda_unflag_cnode(cn, C_VATTR | C_CONTAINER | C_HAVEDATA);
			if (de->d_parent) coda_unflag_cnode((struct cnode *)de->d_parent->d_vnode, C_HAVEDATA);
			d_shrink_cache_parent(de);
			dget(de);
			d_drop(de);
			dput(de);
		}else{
			LOGDEBUG("purgefid: no cnode", 1);
			print_fid(&out->coda_purgefid.CodaFid);
		}

		break;
	case CODA_ZAPDIR:
		fid = &out->coda_zapdir.CodaFid;
		LOGDEBUG("---CODA_ZAPDIR---", 1);

		cn = coda_lookupFid(fid);

		if (cn) {
			coda_flag_children(cn, C_PURGE);
	        coda_unflag_cnode(cn, C_VATTR | C_CONTAINER | C_HAVEDATA);
		}else{
			LOGDEBUG("zapdir: no cnode", 1);
			print_fid(&out->coda_zapdir.CodaFid);
		}

		break;
	case CODA_ZAPFILE:
		fid = &out->coda_zapfile.CodaFid;
		LOGDEBUG("---CODA_ZAPFILE---", 1);

		cn = coda_lookupFid(fid);

		if (cn) {
			coda_unflag_cnode(cn, C_VATTR | C_CONTAINER);
		}else{
			LOGDEBUG("zapfile: no cnode", 1);
			print_fid(&out->coda_zapfile.CodaFid);
		}

		break;
	case CODA_PURGEUSER:
		LOGDEBUG("---CODA_PURGEUSER---", 1);
		break;
	case CODA_REPLACE:
		{
			ViceFid *oldfid = &out->coda_replace.OldFid;
			ViceFid *newfid = &out->coda_replace.NewFid;
			struct cnode *cn = NULL;
			LOGDEBUG("---CODA_REPLACE---",1);
			print_fid(newfid);
			print_fid(oldfid);
			cn = coda_lookupFid(oldfid);
			if ( cn ) { 
				LOGDEBUG("replacefid", 10);
				coda_replace_fid(cn, oldfid, newfid);
			} else 
				LOGDEBUG("replace fid: no cnode", 10);
		  
			break;
		}

		break;
	default:
		LOGDEBUG("---UNKNOWN DOWNCALL---", 1);
	}
	return 0;
}
