/* vol.c -- Windows95 IFS routines for volume manipulation
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/


#include "mc.h"

#include "string.h"
#pragma intrinsic(strlen, memset, memcpy, memcmp)

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

#ifdef O
static struct volfunc vf = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_VOLFUNC,

	DeleteFile,				/* DeleteFile */
	Dir,					/* Dir */
	FileAttributes,			/* FileAttributes */
	FlushVolume,			/* FlushVolume */
	GetDiskInfo,			/* GetDiskInfo */

	OpenFile,				/* OpenFile */
	RenameFile,				/* RenameFile */
	SearchFile,				/* SearchFile */
	QueryResourceInfo,		/* QueryResourceInfo */
	DisconnectResource,		/* DisconnectResource */

	UNCNamedPipeRequest,	/* UNCNamedPipeReq */
	ErrorStub,				/* Ioctl16Drive */
	ErrorStub,				/* GetDiskParms */
	FindFirstFile,			/* FindFirstFile */
	ErrorStub				/* DirectDiskIO */
};

#endif



UINT uiNetId = 0x001b0000;
UINT uiProviderId = 0;

struct dentry *root = NULL;
USHORT	coda_volume_name[] = { 'C', 'o', 'd', 'a', 0};

int ConnectNetResource (pioreq pir, pFilesystem fs)
{
	int res = 0; 

	LOGDEBUG("--ConnectNetResource--", 1);

	if ((pir->ir_flags != RESTYPE_WILD) &&
		(pir->ir_flags != RESTYPE_DISK))
		return pir->ir_error = ERROR_NOT_SUPPORTED;

	/* get root VFid */
//	res = coda_root(de);

	root = d_alloc_root(fs, NULL);
	root->d_ops = &coda_dentry_operations;
	root->d_vnode_ops = &coda_vnode_operations;

	if (res)
		return pir->ir_error = res;

	pir->ir_flags = RESTYPE_DISK;
	pir->ir_pathSkip = 2;
//	pir->ir_vfunc = &vf;

	inUseCount++;

	LOGDEBUG("--end of ConnectNetResource--", 1);
	return pir->ir_error = res;
}


int DisconnectResource (pioreq pir, pFilesystem fs)
{
	LOGDEBUG("--disconnect--", 1);
	if (fs != my_fs) return pir->ir_error = ERROR_INVALID_PARAMETER;

	clear_resource_requests ((DWORD) pir->ir_rh);
	inUseCount--;
	dput(root);
	root = NULL;
	LOGDEBUG("--end of disconnect--", 1);
	return pir->ir_error = 0;
}


int ErrorStub (pioreq pir, struct dentry *de)
{
	LOGDEBUG("--errorstub--", 10);
	LOGDEBUG("--end of errorstub--", 10);
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}

int ErrorStub2 (pioreq pir)
{
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}

int FileAttributes (pioreq pir, struct dentry *de)
{
	struct cnode *cn = NULL;
	int res = 0;

	printf ("--fileattrib-- (subfunc %x)", pir->ir_flags);
	logucode (pir->ir_upath);

	if (!de)
		goto error;

	cn = (struct cnode *)de->d_vnode;
	
	if (!cn) {
		res = ERROR_FILE_NOT_FOUND;
		goto error;
	}

	if (pir->ir_flags == GET_ATTRIBUTES) {
		pir->ir_attr = (cn->c_vattr.va_type == C_VDIR) ?
					FILE_ATTRIBUTE_DIRECTORY : 0;
		printf("--fileattrib file type %x", pir->ir_attr);
		res = 0;
		goto error;
	}

	if (pir->ir_flags == GET_ATTRIB_FIRST_CLUST) {
		res = ERROR_INVALID_FUNCTION;
		goto error;
	}

	/* otherwise, it is one of the date/time functions:
	   get current attributes */
	res = coda_getattr(cn);
	if (res)
		goto error;

	switch (pir->ir_flags) {
	case GET_ATTRIB_COMP_FILESIZE:
		LOGDEBUG("--fileattrib: GET_ATTRIB_COMP_FILESIZE--", 1);
		pir->ir_size = (unsigned long)cn->c_vattr.va_size;
		break;

	case SET_ATTRIB_MODIFY_DATETIME:
		LOGDEBUG("--fileattrib: SET_ATTRIB_MODIFY_DATETIME--", 1);
		cn->c_vattr.va_mtime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		cn->c_vattr.va_mtime.ts_nsec = 0;
		break;

	case GET_ATTRIB_MODIFY_DATETIME:
		LOGDEBUG("--fileattrib: GET_ATTRIB_MODIFY_DATETIME--", 1);
		pir->ir_dostime = IFSMgr_NetToDosTime(cn->c_vattr.va_mtime.ts_sec);
		break;

	case SET_ATTRIB_LAST_ACCESS_DATETIME:
		LOGDEBUG("--fileattrib: SET_ATTRIB_LAST_ACCESS_DATETIME--", 1);
		cn->c_vattr.va_atime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		cn->c_vattr.va_atime.ts_nsec = 0;
		break;

	case GET_ATTRIB_LAST_ACCESS_DATETIME:
		LOGDEBUG("--fileattrib: GET_ATTRIB_LAST_ACCESS_DATETIME--", 1);
		pir->ir_dostime = IFSMgr_NetToDosTime(cn->c_vattr.va_atime.ts_sec);
		break;

	case SET_ATTRIB_CREATION_DATETIME:
		LOGDEBUG("--fileattrib: SET_ATTRIB_CREATION_DATETIME--", 1);
		cn->c_vattr.va_ctime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		cn->c_vattr.va_ctime.ts_nsec = pir->ir_options * 10000000;
		break;

	case GET_ATTRIB_CREATION_DATETIME:
		LOGDEBUG("--fileattrib: GET_ATTRIB_CREATION_DATETIME--", 1);
		pir->ir_dostime = IFSMgr_NetToDosTime(cn->c_vattr.va_ctime.ts_sec);
		pir->ir_options = cn->c_vattr.va_ctime.ts_nsec / 10000000;
		break;
	}

	if (pir->ir_flags == SET_ATTRIB_MODIFY_DATETIME ||
		pir->ir_flags == SET_ATTRIB_LAST_ACCESS_DATETIME ||
		pir->ir_flags == SET_ATTRIB_CREATION_DATETIME)
		res = coda_setattr(cn);

error:
	printf ("--end of fileattrib (%x)--", res);
	dput(de);
	return pir->ir_error = res;
}


int DeleteFile (pioreq pir, struct dentry *de)
{
	struct cnode *cn = NULL;
	struct dentry *parent = NULL;
	int res = 0;

	LOGDEBUG("--deletefile--", 1);
	/* according to docs, returning INVALID_FUNCTION on a wildcard
		request will cause the IFSMgr to retry with a sequence of
		non-wildcard requests */
	if (pir->ir_attr & FILE_FLAG_WILDCARDS)
		return pir->ir_error = ERROR_INVALID_FUNCTION;


	if (!de)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (!memcmp(de->d_name.name + de->d_name.len - 4, ".$cf", 4)){
		res = ERROR_ACCESS_DENIED;
		goto error;
	}

	parent = dget(de->d_parent);
	res = venus_remove((struct cnode*)parent->d_vnode, pir);
	dput(parent);

	d_drop(de);

error:
	dput (de);

	printf ("delete returning %x", res);
	return pir->ir_error = res;

}


int FlushVolume (pioreq pir)
{
	LOGDEBUG("--flushvolume--", 10);
	LOGDEBUG("--end of flushvolume--", 10);
	return pir->ir_error = 0;
}


int Dir (pioreq pir, struct dentry *de)
{
	int res = ERROR_INVALID_FUNCTION;
	struct dentry *parent = NULL;
	struct cnode *cn = NULL;
	int i = 1;

	printf ("--dir-- (subfunc %x)", pir->ir_flags);

	if (pir->ir_flags == QUERYLONG_DIR ) {		
		LOGDEBUG("--querylong_dir--", 1);
		pir->ir_ppath2 = pir->ir_ppath;
		logucode((PWORD)pir->ir_ppath2);
		goto error;
	}

	if (pir->ir_flags == QUERY83_DIR){
		LOGDEBUG("--query83_dir--", 1);
		pir->ir_ppath2 = pir->ir_ppath;
		logucode((PWORD)pir->ir_ppath2);
		goto error;
	}

	if (pir->ir_flags == CHECK_DIR) {		
		LOGDEBUG("--check dir--", 1);

		if (pir->ir_uFName[0] == 0){
			res = 0;
			goto error;
		}

        if (IFSIsRoot(pir->ir_ppath)) {
            res = 0;
			goto error;
		}

		logucode(pir->ir_upath);

		if (!de) {
			res = ERROR_PATH_NOT_FOUND;
			goto error;
		}

		cn = (struct cnode*) de->d_vnode;

		if (!cn){
			res = ERROR_PATH_NOT_FOUND;
			goto error;
		}
		if (cn->c_vattr.va_type != C_VDIR)
			res = ERROR_PATH_NOT_FOUND; 
		goto error;
	}

	if (pir->ir_flags == CREATE_DIR) {
		LOGDEBUG("--create dir--", 1);

		if (!de) {
			res = ERROR_INVALID_PARAMETER;
			goto error;
		}

		cn = (struct cnode*) de->d_vnode;

		if (!memcmp(de->d_name.name + de->d_name.len - 4, ".$cf", 4)){
			res = ERROR_ACCESS_DENIED;
			d_drop(de);
			goto error;
		}

		if (!cn)
			res = venus_mkdir(pir, &de);
		else res = ERROR_FILE_EXISTS;
	}

	if (pir->ir_flags == DELETE_DIR) {
		LOGDEBUG("--delete dir--", 1);

		if (pir->ir_attr & FILE_FLAG_WILDCARDS)
			return ERROR_INVALID_FUNCTION;

		if (!de) {
			res = ERROR_INVALID_PARAMETER;
			goto error;
		}

		cn = (struct cnode*) de->d_vnode;

		if (!memcmp(de->d_name.name + de->d_name.len - 4, ".$cf", 4)){
			res = ERROR_ACCESS_DENIED;
			goto error;
		}
		
		parent = dget(de->d_parent);
		res = venus_rmdir((struct cnode*) parent->d_vnode, pir);
		dput(parent);

		d_drop(de);
	}

error:
	dput(de);

	printf ("--end of dir: %x--", res);
	return pir->ir_error = res;
}



int GetDiskInfo (pioreq pir)
{
	LOGDEBUG("--getdiskinfo--", 1);

	pir->ir_length = 4096;			/* bytespersector */
	pir->ir_size = 65000;			/* totalallocunits */
	pir->ir_sectors = 8;			/* sectorsperallocunit */
	pir->ir_numfree = 40000;		/* numfreeallocunits */

	LOGDEBUG("--end of getdiskinfo--", 1);
	return pir->ir_error = 0;
}

//static USHORT resourcename[] = { 32, 16,
// 12,'\\', 'C', 'O', 'D', 'A',
//16, 'C', 'L', 'U', 'S', 'T', 'E', 'R',
//	0 };

static USHORT resourcename[] = { 'C', 'O', 'D', 'A', 'C', 'L', 'U', 'S', 'T', 'E', 'R', 0 };

int QueryResourceInfo (pioreq pir)
{
	DWORD len;
	LOGDEBUG("--queryresourceinfo--", 1);

	switch (pir->ir_options) {
	case 1:
		pir->ir_options = RESSTAT_OK;
		pir->ir_flags = RESTYPE_DISK;
		/* fallthrough */
	case 0:
		/* URGH! hooking Microsoft's FSDs reveals what we're supposed to do
		   here -- pity the documentation didn't */
		memcpy ((PCHAR) pir->ir_ppath, (PCHAR) resourcename, sizeof(resourcename));
		break;

	case 2:
	/* note: be careful to use pir->ir_length before overwriting it */
		printf ("level 2: data %x len %x", pir->ir_data, pir->ir_length);
		if (pir->ir_length) {
			len = 16;
			if (len > pir->ir_length - 1)
				len = pir->ir_length - 1;
			memcpy((unsigned char *) pir->ir_data, "Coda FS", 
				len);
			((PCHAR) pir->ir_data) [len] = '\0';
		}
		pir->ir_length = (CODA_MAXPATHLEN << 16) + CODA_MAXNAMLEN; /* old : 0x010400ff; */
		pir->ir_options = FS_CASE_IS_PRESERVED | FS_VOL_SUPPORTS_LONG_NAMES;
		pir->ir_pos = 4096;
		break;

	default:
		LOGDEBUG("QUERYRESOURCEINFO: unknown level", 10);
	}

	LOGDEBUG("--end of queryresourceinfo--", 1);
	return pir->ir_error = 0;
}

int RenameFile (pioreq pir, struct dentry *de1, struct dentry *de2)
{
	struct cnode *cn1  = NULL, *cn2 = NULL;
	int res = ERROR_INVALID_PARAMETER;

	LOGDEBUG("--renamefile--", 1);

	if (pir->ir_attr & FILE_FLAG_WILDCARDS || pir->ir_attr2 & FILE_FLAG_WILDCARDS)
		return ERROR_INVALID_FUNCTION;

	if (!de1 || !de2)
		goto error;

	cn1 = (struct cnode*) de1->d_vnode;


	cn2 = (struct cnode*) de2->d_vnode;

	if (!cn1)
		goto error;

	if (cn2 || de2->d_neg == 0){
		res = ERROR_FILE_EXISTS;
		goto error;
	}

	if (!memcmp(de1->d_name.name + de1->d_name.len - 4, ".$cf", 4) ||
		!memcmp(de2->d_name.name + de2->d_name.len - 4, ".$cf", 4) ){
			res = ERROR_ACCESS_DENIED;
			d_drop(de2);
			goto error;
	}
	
	res = venus_rename(pir, de1, de2);

	if (!res)
		d_switchhash(de1, de2);
			
error:
	dput (de1);
	dput (de2);

	printf ("--end of rename returning %x--", res);
	return pir->ir_error = res;
}

int UNCNamedPipeRequest (pioreq pir)
{
	printf ("--- failing UNCNamedPipeRequest %x", pir->ir_flags);
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}