/*  
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/

#include "sock.h"

PSOCKET
NewSocket (void)
{
	PSOCKET p;
	p = (PSOCKET) _HeapAllocate(sizeof(SOCKET), HEAPZEROINIT);
	if (p)
		p->s_signature = SOCKSIG;
	return p;
}

int
GetSocket (HVM hVM, PCRS pRegs, PSOCKET *pp)
{
	PFDTABLE pfd = (PFDTABLE) *((PDWORD)((DWORD) hVM + cbmyVMarea));
	DWORD fd;
	fd = pRegs->Client_EDI;
	if (fd > 31)
		return -SOCKERR_INVALID_SOCKET;
	if (!pfd || !(pfd->fd_inuse & (1 << fd)))
		return -SOCKERR_INVALID_SOCKET;
	*pp = (PSOCKET) pfd->fd_fd[fd].fd_context;
	if ((*pp)->s_signature != SOCKSIG)
		return -SOCKERR_INVALID_SOCKET;
	return 0;
}

void
FreeSocket (PSOCKET pSock)
{
	TDI_REQUEST r;
	PSOCKET p, prev;

	for (p = pSock->s_next; p; ) {
		prev = p;
		p = p->s_next;
		if (prev->s_conhdl) {
			r.Handle.ConnectionContext = prev->s_conhdl;
			r.RequestNotifyObject = (PVOID) NullCallback;
			r.RequestContext = NULL;
			pDT->TdiCloseConnectionEntry (&r);
		}
		if (prev->s_addrobj)
			ReleaseAddress(prev->s_addrobj);
		if (prev->s_recvbuf)
			FreeCBuf(prev->s_recvbuf);
		_HeapFree(prev, 0);
	}

	if (pSock->s_conhdl) {
		r.Handle.ConnectionContext = pSock->s_conhdl;
		r.RequestNotifyObject = (PVOID) NullCallback;
		r.RequestContext = NULL;
		pDT->TdiCloseConnectionEntry (&r);
	}

	if (pSock->s_recvbuf)
		FreeCBuf(pSock->s_recvbuf);

	if (pSock->s_addrobj)
		ReleaseAddress(pSock->s_addrobj);

	_HeapFree(pSock, 0);
}

PADDRESSOBJECT
NewAddressObject (void)
{
	PADDRESSOBJECT obj;

	obj = (PADDRESSOBJECT) _HeapAllocate(sizeof(ADDRESSOBJECT), 0);
	if (!obj)
		return NULL;
	obj->addr_count = 1;
	obj->addr_hdl = 0;
	return obj;
}

void
AddRefAddress (PADDRESSOBJECT obj)
{
	obj->addr_count++;
}

TDI_STATUS
ReleaseAddress (PADDRESSOBJECT obj)
{
	TDI_REQUEST r;
	obj->addr_count--;
	if (obj->addr_count)
		return TDI_SUCCESS;

	if (!obj->addr_hdl)
		return TDI_SUCCESS;

	r.Handle.AddressHandle = obj->addr_hdl;
	r.RequestNotifyObject = (PVOID) NullCallback;
	r.RequestContext = NULL;
	_HeapFree(obj, 0);
	return pDT->TdiCloseAddressEntry(&r);
}

int
GetAllocatedFD(HVM hVM)
{
	PFDTABLE *p = (PFDTABLE *)((DWORD) hVM + cbmyVMarea);
	PFDTABLE pfd = *p;

	if (!pfd)
		return -1;

	return pfd->fd_inuse;
}

int
AllocateFD (HVM hVM, DWORD fd, PFDOPS fdops, PVOID context)
{
	PFDTABLE *p = (PFDTABLE *)((DWORD) hVM + cbmyVMarea);
	PFDTABLE pfd = *p;

	if (!pfd) {
		pfd = (PFDTABLE) _HeapAllocate (sizeof(FDTABLE), HEAPZEROINIT);
		if (!pfd)
			return -SOCKERR_NO_MEMORY;
		*p = pfd;
	}

	if (pfd->fd_inuse & (1 << fd))
		return -SOCKERR_FD_INUSE;

	pfd->fd_inuse |= (1 << fd);
	pfd->fd_fd[fd].fd_ops = fdops;
	pfd->fd_fd[fd].fd_context = context;
	return 0;
}

int
DeallocateFD (HVM hVM, DWORD fd)
{
	PFDTABLE pfd = (PFDTABLE) *((PDWORD)((DWORD) hVM + cbmyVMarea));

	if (fd > 31)
		return -SOCKERR_INVALID_SOCKET;

	if (!pfd || !(pfd->fd_inuse & (1 << fd)))
		return -SOCKERR_INVALID_SOCKET;

	pfd->fd_inuse &= ~(1 << fd);
	return 0;
}

/* these are wrappers for the VxD service table which we export */

int SOCK_AllocateFD (HVM hVM, DWORD fd, PFDOPS fdops, PVOID context)
{
	return AllocateFD (hVM, fd, fdops, context);
}

int SOCK_DeallocateFD (HVM hVM, DWORD fd)
{
	return DeallocateFD (hVM, fd);
}


PCALLBACKMEMO
NewCallback (void)
{
	PCALLBACKMEMO p = _HeapAllocate(sizeof(CALLBACKMEMO), 0);
	if (p) {
		p->cbm_sem = Create_Semaphore(0);
		if (!p->cbm_sem) {
			_HeapFree(p, 0);
			return NULL;
		}
	}
	return p;
}

void
FreeCallback (PCALLBACKMEMO p)
{
	if (p) {
		if (p->cbm_sem)
			Destroy_Semaphore(p->cbm_sem);
		_HeapFree(p, 0);
	}
}

TDI_STATUS
WaitCallback (PCALLBACKMEMO p)
{
	Wait_Semaphore (p->cbm_sem, BLOCK_SVC_INTS);
	return p->cbm_status;
}

void
SignalCallback (void *Context, TDI_STATUS final, uint ByteCount)
{
	PCALLBACKMEMO p = (PCALLBACKMEMO) Context;
	if (p) {
		p->cbm_status = final;
		if (p->cbm_sem)
			Signal_Semaphore(p->cbm_sem);
	}
}

PTRANSPORT_ADDRESS
MakeTransportAddress (DWORD dwAddr, WORD wPort)
{
	PTRANSPORT_ADDRESS pTA;
	PTA_ADDRESS pTAA;
	PTDI_ADDRESS_IP pTAIP;

	pTA = (PTRANSPORT_ADDRESS) _HeapAllocate (
			4 + (2 + 2 + TDI_ADDRESS_LENGTH_IP), HEAPZEROINIT);
	// packing in structures???

	if (!pTA) 
		return NULL;

	pTA->TAAddressCount = 1;
	pTAA = (PTA_ADDRESS) (&pTA->Address[0]);
	pTAA->AddressLength = TDI_ADDRESS_LENGTH_IP;
	pTAA->AddressType = TDI_ADDRESS_TYPE_IP;
	pTAIP = (PTDI_ADDRESS_IP) (&pTAA->Address[0]);
/*
	pTAIP->sin_port = (short) pRegs->Client_EDX;
	pTAIP->in_addr = pRegs->Client_EBX;
*/
	pTAIP->sin_port = wPort;
	pTAIP->in_addr = dwAddr;
	return pTA;
}

void
FreeTransportAddress (PTRANSPORT_ADDRESS p)
{
	_HeapFree (p, 0);
}
