
/* sock.h -- definitions for socket VxD   
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/

#ifdef SOCKVXD

#define WANTVXDWRAPS
#include <basedef.h>
#include <vmm.h>
#include <vwin32.h>
#include <shell.h>
#include <vxdwraps.h>
#include "vmmextra.h"

#define VXD
#include <ndis.h>
#undef LCODE
#include <tdi.h>
#include <tdistat.h>

#include <string.h>
#pragma intrinsic(strlen, strcpy, memset, memcpy)

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;
typedef void CTEReqCmpltRtnFn (PVOID, TDI_STATUS, uint);
typedef CTEReqCmpltRtnFn *CTEReqCmpltRtn;
#include <tdivxd.h>

#include "cbuf.h"

#ifndef _LinPageUnLock
#define _LinPageUnLock _LinPageUnlock
#endif

typedef struct _FDOPS {
	DWORD	(*fdop_poll) (PVOID context, PDWORD post);
	void	(*fdop_close) (PVOID context);
} FDOPS, *PFDOPS;

extern FDOPS sockops;

#define POLL_READ		0x00000001
#define POLL_WRITE		0x00000002
#define	POLL_EXCEPTION	0x00000004

typedef struct _FDTABLE {
	DWORD	fd_inuse;
	DWORD	fd_post;
	struct {
		PFDOPS	fd_ops;
		PVOID	fd_context;
	} fd_fd[32];
	/* used for communication between sys_select & timeout event --
		note that sock.asm code has unholy knowledge
	    of the layout of this struct;
		also, EndPMApp uses this to cancel the timeout if
		we nuke a DPMI app that's paused in select
	*/
	struct timeout_event_info {
		PDWORD		post;
		DWORD		timeeventhdl;
	} fd_tei;   	
} FDTABLE, *PFDTABLE;

typedef struct _ADDRESSOBJECT {
	DWORD	addr_count;
	HANDLE	addr_hdl;
} ADDRESSOBJECT, *PADDRESSOBJECT;

typedef struct SendCallbackMemo  SENDCALLBACKMEMO, *PSENDCALLBACKMEMO, 
	**PPSENDCALLBACKMEMO;
typedef struct _SOCKET SOCKET, *PSOCKET, **PPSOCKET;

struct SendCallbackMemo {
	PSENDCALLBACKMEMO scbm_next;
	PSOCKET			scbm_socket;
	PCHAR			scbm_buffer;
	DWORD			scbm_len;
	PNDIS_BUFFER	scbm_ndis;
};

struct _SOCKET {
	DWORD			s_signature;		/* signature */
	DWORD			s_flags;
	PADDRESSOBJECT	s_addrobj;			/* address object */
	HANDLE			s_conhdl;			/* handle for connection object */
	DWORD			s_recvpackets;		/* number of packets in buffer (UDP only) */
	PCBUF			s_recvbuf;			/* receive circular buffer */
	PSENDCALLBACKMEMO s_sendcbm;		/* linked list of send callback requests */
	TDI_STATUS		s_sendstatus;		/* result of last send */
	DWORD			s_sendbytes;		/* how many in queue */
	DWORD			s_sendmax;			/* maximum number of bytes to buffer */
	DWORD			s_sendlowwater;		/* minimum number of free bytes in buffer to be writable */
	PSOCKET			s_next;				/* listen queue linked list */
	PSOCKET			s_parent;			/* link to listening socket for listen queue sockets */
	PDWORD			s_post;				/* select post value */
	DWORD			s_remoteip;
	WORD			s_remoteport;
	PTDI_CONNECTION_INFORMATION	s_connectinfo;	/* used in listen/accept */
	PTDI_CONNECTION_INFORMATION s_acceptable;	/* ditto */
};

typedef struct _CALLBACKMEMO {
	VMM_SEMAPHORE	cbm_sem;
	TDI_STATUS		cbm_status;
} CALLBACKMEMO, *PCALLBACKMEMO;

#endif

#define SOCKSIG				0xbeefade0

#define SOCKFL_TCP			0x00000001
#define SOCKFL_UDP			0x00000002

#define SOCKFL_BOUND		0x00000100	/* bind() has been done */
#define SOCKFL_NONBLOCK		0x00000200	/* setsockopt(NONBLOCK) called */
#define SOCKFL_CONNECTED	0x00000400  /* connect() done */
#define SOCKFL_LISTENUSED	0x00000800  /* this socket already allocated by TDI_CONNECT_EVENT handler */
#define SOCKFL_LISTENING	0x00001000	/* this socket has a s_next chain of listening connections */
#define SOCKFL_PEERCLOSED	0x00002000	/* peer closed this socket */
#define SOCKFL_PEERABORTED	0x00004000  /* peer aborted this socket */

#ifdef SOCKVXD

extern int ConnectEvents;
extern int ConnectCallbacks;

int sys_socket(HVM, PCRS);
int sys_getsetmode(HVM, PCRS, PSOCKET);
int do_bind(HVM, PCRS, PSOCKET, PTRANSPORT_ADDRESS);
int sys_bind(HVM, PCRS, PSOCKET);
int sys_connect(HVM, PCRS, PSOCKET);
int sys_listen(HVM, PCRS, PSOCKET);
int sys_accept(HVM, PCRS, PSOCKET);
int sys_getname(HVM, PCRS, PSOCKET);
int sys_getpeer(HVM, PCRS, PSOCKET);
int sys_send(HVM, PCRS, PSOCKET);
int sys_sendto(HVM, PCRS, PSOCKET);
int sys_recvfrom(HVM, PCRS, PSOCKET);
int sys_receive(HVM, PCRS, PSOCKET);
int sys_close(HVM, PCRS, PSOCKET);
int sys_select(HVM, PCRS);
int do_listen (PSOCKET);
int GetAllocatedFD(HVM);
void ListenCallback (PVOID, TDI_STATUS, uint);
TDI_STATUS RcvDatagram(PVOID, uint, PTRANSPORT_ADDRESS, uint, PVOID, 
					   uint, uint, uint, uint *, uchar *, EventRcvBuffer **);
TDI_STATUS Receive(PVOID, PVOID, ulong, uint, uint, uint *, uchar *, 
				   EventRcvBuffer *);
HANDLE makeconnection (HANDLE, PSOCKET);
void ConnectCallBack (PVOID, TDI_STATUS, uint);
TDI_STATUS ConnectEvent (PVOID, uint, PTRANSPORT_ADDRESS, uint, PVOID, 
						 uint, PVOID, PVOID *, ConnectEventInfo *);
TDI_STATUS DisconnectEvent(PVOID, PVOID, uint, PVOID, uint, PVOID, ulong);
void getipport (PTRANSPORT_ADDRESS, PWORD, PDWORD); 

void do_close(PVOID);
DWORD do_poll(PVOID, PDWORD);
void wait(PSOCKET, DWORD);

CTEReqCmpltRtnFn NullCallback;
extern int cbmyVMarea;

TDIDispatchTable *pDT;

PSOCKET NewSocket(void);
int GetSocket(HVM, PCRS, PSOCKET *);
void FreeSocket(PSOCKET);
PADDRESSOBJECT NewAddressObject(void);
void AddRefAddress(PADDRESSOBJECT);
TDI_STATUS ReleaseAddress(PADDRESSOBJECT);
int AllocateFD(HVM, DWORD, PFDOPS, PVOID);
int DeallocateFD(HVM, DWORD);
PTRANSPORT_ADDRESS MakeTransportAddress(DWORD dwAddr, WORD wPort);
void FreeTransportAddress(PTRANSPORT_ADDRESS);
#define IP_TRANSPORT_ADDRESS_LENGTH (4 + (2 + 2 + TDI_ADDRESS_LENGTH_IP))

PCALLBACKMEMO NewCallback(void);
TDI_STATUS WaitCallback(PCALLBACKMEMO);
void FreeCallback(PCALLBACKMEMO);
CTEReqCmpltRtnFn SignalCallback;
#endif

#define SOCKERR_INVALID_PARAM		0x00000100
#define SOCKERR_NO_MEMORY			0x00000200
#define SOCKERR_INVALID_SOCKET		0x00000300
#define SOCKERR_ALREADY_BOUND		0x00000400
#define SOCKERR_NOT_BOUND			0x00000500
#define SOCKERR_ACCESS				0x00000600
#define SOCKERR_INTERNAL			0x00000700
#define SOCKERR_FD_INUSE			0x00000800
#define SOCKERR_INFINITE_WAIT		0x00000900
#define SOCKERR_NOT_CONNECTED		0x00000a00
#define SOCKERR_WOULD_BLOCK			0x00000b00
#define SOCKERR_NOT_LISTENING		0x00000c00
