/*

  sockc.c  -- C code for sock VxD

  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.

*/

#include "sock.h"
#include <pbt.h>

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

int cbmyVMarea = 0;

int OnSystemBroadcast (DWORD uMsg, DWORD wParam, DWORD lParam, DWORD ref);

static char copyright[] = "Copyright 1997 Michael Callahan\n";

#pragma warning(disable: 4035)
#define VTDI_DEVICE_ID			0x0488
Begin_Service_Table(VTDI, none)
#define VTDI_Service(X) Declare_Service(X, none)
VTDI_Service (VTDI_Get_Version)
VTDI_Service (VTDI_Start_Timer)
VTDI_Service (VTDI_Stop_Timer)
VTDI_Service (VTDI_Schedule_Event)
VTDI_Service (VTDI_Cancel_Event)
VTDI_Service (VTDI_Block)
VTDI_Service (VTDI_Signal)
VTDI_Service (VTDI_Register)
VTDI_Service (VTDI_Get_Info)
VTDI_Service (VTDI_Unload)
VTDI_Service (VTDI_Initialize)
VTDI_Service (VTDI_Register_UnloadProc)
VTDI_Service (VTDI_Register_LoadProc)
End_Service_Table(VTDI, none)

DWORD VTDI_Get_Version()
{
	VMMCall(VTDI_Get_Version)
	_asm jnc noerror
	_asm mov eax, 4321h
	_asm noerror:
}

DWORD __declspec(naked) VTDI_Get_Info(char *psz)
{
	VMMJmp(VTDI_Get_Info)
}

TDIDispatchTable *pDT = 0;


#pragma VxD_INIT_DATA_SEG
#pragma VxD_INIT_CODE_SEG


#pragma warning(disable:4035)
DWORD __declspec(naked) _Allocate_Device_CB_Area(ULONG size, ULONG flags)
{
	VMMJmp(_Allocate_Device_CB_Area)
}
#pragma warning(default:4035)


extern int BSSstart, BSSend;

static char TCPname[] = "MSTCP";
static SYSBHOOK_HANDLE sh;

BOOL OnDeviceInit ()
{
	/* clear bss segment */
	memset (&BSSstart, 0, ((char *)&BSSend) - ((char *)&BSSstart));

	/* allocate VMM area for us to use to remember which memory
	   block we allocated for the client */
	cbmyVMarea = _Allocate_Device_CB_Area (sizeof (PFDTABLE), 0);
	if (!cbmyVMarea)
		return FALSE;

	if (!VTDI_Get_Version())
		return FALSE;

	if (!(pDT = (TDIDispatchTable *) VTDI_Get_Info(TCPname)))
		return FALSE;

//	sh = _SHELL_HookSystemBroadcast(OnSystemBroadcast, 0, 0);
	
	return  TRUE;
}

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG


#pragma warning(disable:4035)
DWORD __declspec(naked) _Deallocate_Device_CB_Area(ULONG offset, ULONG flags)
{
	VMMJmp(_Deallocate_Device_CB_Area)
}
#pragma warning(default: 4035)

BOOL OnSystemBroadcast (DWORD uMsg, DWORD wParam, DWORD lParam, DWORD ref)
{
	return TRUE;
}

BOOL OnDeviceExit ()
{
	/* deallocate VMM area */
	if (cbmyVMarea)
		_Deallocate_Device_CB_Area (cbmyVMarea, 0);

//	_SHELL_UnhookSystemBroadcast(sh);

	return TRUE;
}

BOOL EndPMApp2 (HVM hVM)
{
	/* A protected-mode application is ending.  Free its memory
	   block if it allocated one. */
	PFDTABLE *p = (PFDTABLE *)((DWORD) hVM + cbmyVMarea);
	PFDTABLE pfd = *p;
	int i;

	if (!pfd)
		return TRUE;

	Cancel_Time_Out(pfd->fd_tei.timeeventhdl);

	if (pfd->fd_inuse) 
		for (i = 0; i < 32; i++) {
			if ((pfd->fd_inuse & (1 << i)) &&
				pfd->fd_fd[i].fd_ops &&
				pfd->fd_fd[i].fd_ops->fdop_close)
				(*pfd->fd_fd[i].fd_ops->fdop_close) (pfd->fd_fd[i].fd_context);
		}

	*p = NULL;
	_HeapFree(pfd, 0);
	return TRUE;
}




void PM_API (HVM, DWORD, PCRS);

void __declspec(naked) SOCK_PM_API ()
{
	_asm {
		push	ebp
		push	edi
		push	ebx
		call	PM_API
		pop		eax
		pop		eax
		pop		eax
		ret
	}
}

#pragma warning(default: 4035)


void 
NullCallback (PVOID context, TDI_STATUS finalstatus, uint ByteCount)
{
}


void PM_API (HVM hVM, DWORD hThread, PCRS pRegs)
{
	int err=0;
	PSOCKET pSock;

	switch (pRegs->Client_EAX) {
	case 0:
		pRegs->Client_EAX = 1;
		if (!err)
			pRegs->Client_EFlags &= ~CF_MASK;
		else
			pRegs->Client_EFlags |= CF_MASK;
		return;

	case 1:
		err = sys_socket(hVM, pRegs);
		pRegs->Client_EAX = err;
		if (!err)
			pRegs->Client_EFlags &= ~CF_MASK;
		else
			pRegs->Client_EFlags |= CF_MASK;
		return;

	case 6:
		Out_Debug_String("SOCK.VXD: sys_select ");
		err = sys_select(hVM, pRegs);
		pRegs->Client_EAX = err;
		if (!err)
			pRegs->Client_EFlags &= ~CF_MASK;
		else
			pRegs->Client_EFlags |= CF_MASK;
		return;

	case 16:
		err = GetAllocatedFD(hVM);
		pRegs->Client_EAX = err;
		if (!err)
			pRegs->Client_EFlags &= ~CF_MASK;
		else
			pRegs->Client_EFlags |= CF_MASK;
		return;

	case 100:
		pRegs->Client_EAX = VTDI_Get_Version();
		return;

	case 101:
		pRegs->Client_EAX = VTDI_Get_Info(TCPname);	
		return;
	}

	if (pRegs->Client_EAX > 16) {
		pRegs->Client_EAX = 0x1234;
		pRegs->Client_EFlags |= CF_MASK;
		return;
	}

	err = GetSocket(hVM, pRegs, &pSock);
	if (!err) {
		switch (pRegs->Client_EAX) {
		case 2:
			Out_Debug_String("SOCK.VXD: sys_close ");
			err = sys_close(hVM, pRegs, pSock);
			break;

		case 3:
			Out_Debug_String("SOCK.VXD: sys_bind ");
			err = sys_bind(hVM, pRegs, pSock);
			break;

		case 4:
			Out_Debug_String("SOCK.VXD: sys_sendto ");
			err = sys_sendto(hVM, pRegs, pSock);
			break;

		case 5:
			Out_Debug_String("SOCK.VXD: sys_recvfrom ");
			err = sys_recvfrom(hVM, pRegs, pSock);
			break;
	
		case 7:
			Out_Debug_String("SOCK.VXD: sys_connect ");
			err = sys_connect(hVM, pRegs, pSock);
			break;

		case 8:
			Out_Debug_String("SOCK.VXD: sys_send ");
			err = sys_send(hVM, pRegs, pSock);
			break;

		case 9:
			Out_Debug_String("SOCK.VXD: sys_receive ");
			err = sys_receive(hVM, pRegs, pSock);
			break;

		case 10:
			Out_Debug_String("SOCK.VXD: sys_listen ");
			err = sys_listen(hVM, pRegs, pSock);
			break;

		case 11:
			Out_Debug_String("SOCK.VXD: 11 ");
			err = ConnectEvents;
			pRegs->Client_EBX = ConnectCallbacks;
			break;

		case 12:
			Out_Debug_String("SOCK.VXD: sys_accept ");
			err = sys_accept(hVM, pRegs, pSock);
			break;

		case 13:
			Out_Debug_String("SOCK.VXD: sys_getname ");
			err = sys_getname(hVM, pRegs, pSock);
			break;

		case 14:
			Out_Debug_String("SOCK.VXD: sys_getpeer ");
			err = sys_getpeer(hVM, pRegs, pSock);
			break;

		case 15:
			Out_Debug_String("SOCK.VXD: sys_getsetmode ");
			err = sys_getsetmode(hVM, pRegs, pSock);
			break;
		
		}
	}

	pRegs->Client_EAX = err;
//end:
	if (!err)
		pRegs->Client_EFlags &= ~CF_MASK;
	else
		pRegs->Client_EFlags |= CF_MASK;
}
