#include <string.h>
#pragma intrinsic(strcmp, strcpy, strlen, memcpy, memcmp)

#include "cache.h"
#include "dcache.h"

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

dllist_chain	hash_cache[HASHSIZE]; /* hash table of cnodes index parent+name */
dllist_chain	lru_list;

int				lru_list_count;

#define LRU_MAX 1024

int ddd_invalidate(struct dentry *de);
struct dentry *ddd_get(struct dentry *de);
void ddd_delete(struct dentry *de);
int ddd_comparename(struct cstr *name1, struct cstr *name2, int flags);
int ddd_hashhandle(void *v);
struct dentry *ddd_cachelookup(struct dentry *parent, struct cstr *name, int flags);
void ddd_switchhash(struct dentry *de1, struct dentry *de2);
void ddd_clean_lru(int degree);
unsigned int name_hash(struct cstr *name);


void d_init(){
	int i=0;

	list_init(&lru_list);
	lru_list_count = 0;
	for ( i = 0 ; i <= HASHSIZE-1 ; i++ ) { 
		list_init(&(hash_cache[i]));
	}

}


void d_free_all(){
	int i=0, count=0;

	d_shrink_cache();
	i = Debuglevel;
	Debuglevel = 1000;
	logcache_lrulist(&lru_list);
	Debuglevel = i;
}



void dput (struct dentry *de){
	struct dentry *parent = NULL;
	int count;

	if (!de) goto out;

	LOGDEBUG("--dput--", 100);
	print_dentry(de);
repeat:
	count = de->d_refcount - 1;
	if (count != 0){
		if (de->d_refcount >= 0) de->d_refcount = count;
		goto out;
	}

	if (de->d_ops && de->d_ops->d_delete)
		de->d_ops->d_delete(de);

	/* linux does this, because the cnode specific stuff in
	   done in any file system which could block. After returning
	   the cnode could be active again, therefore check again.
	   Here it can not harm to do the check again. If the cache will be
	   a generic cache one day, this is right already. */

	count = de->d_refcount - 1;
	if (count != 0){
		if (de->d_refcount >= 0) de->d_refcount = count;
		goto out;
	}

	if (!list_empty(&de->d_lru)){
		list_del(&de->d_lru);
		lru_list_count--;
	}

	if (list_empty(&de->d_hash)){
		LOGDEBUG("--dput get rid of entry--", 100);
		print_dentry(de);
		if (de->d_ops && de->d_ops->d_vput)
			de->d_ops->d_vput(de);
		list_del(&de->d_child); 
		parent = de->d_parent;
		ddd_delete(de);
		if (!parent) goto out;
		de = parent; 
		goto repeat;  /* tail recursive */
	}
	list_add(&de->d_lru, &lru_list);
	lru_list_count++;
	ddd_clean_lru(2);

	de->d_refcount = count;

out:
	LOGDEBUG("--end of dput--", 100);

}

struct dentry *dget (struct dentry *de){

	LOGDEBUG("--dget--", 100);
	if (de){
		print_dentry(de);
		LOGDEBUG("--end of dget--", 100);
		return ddd_get(de); /* inc count */
	}
	LOGDEBUG("--end of dget--", 100);
	return de;	
}


/* public cache functions */

struct dllist_head *d_hash(struct dentry *parent, unsigned int hash){
	hash += (unsigned int) parent;
	return hash_cache + (hash % HASHSIZE);
}


int d_lookup(struct filesystem *fs, char *path, struct dentry *startde, struct dentry **de, int flags){
	int res = 0;
	struct dentry *dir = NULL;
	struct dentry *parent = NULL;
	char *end = NULL, *copy = NULL;
	struct cstr this;

	*de = NULL;
	LOGDEBUG("---d_lookup---", 1);

	if (res)
		goto out;

	if (!startde) dir = dget(fs->f_root);else dir = dget(startde);
	copy = path;
	while (1) {

		parent = dget(dir);
		dput(dir);

		for (end = copy, this.len=0;*end && *end!='\0' && *end != '\\' && *end != '/';this.len++, end++)
			;

		this.name = copy;
		this.hash = name_hash(&this);

		dir = ddd_cachelookup(parent, &this, flags);
		if (!dir){
			dir = d_alloc(fs, &this, parent);

			res = dir->d_vnode_ops->vo_lookup(&dir, &this, flags);

			if ( res != ERROR_FILE_NOT_FOUND && IS_NEGATIVE(dir) ){
				d_drop(dir);
				dput(dir);
			}
			
		} else {
			LOGDEBUG("---CACHEHIT---", 10);

			if (IS_NEGATIVE(dir))
				res = ERROR_FILE_NOT_FOUND;

		}

		dput(parent);

		if (res || !*end) break;

		copy = end + 1;
	}

	*de = dir;
out:

	LOGDEBUG("---end d_lookup---", 1);

	if (*de && IS_NEGATIVE(*de))
		printf("negativ: %x", (*de)->d_neg);

	return res;

}

void d_rehash(struct dentry *de){
	if (!de) return;

	list_del(&de->d_hash);
	list_init(&de->d_hash);
	de->d_name.hash = name_hash(&de->d_name);
	list_add(&de->d_hash, d_hash(de->d_parent, de->d_name.hash));	
}

void d_switchhash(struct dentry *de1, struct dentry *de2){
	if (de1 && de2)
		ddd_switchhash(de1, de2);
}

void d_drop(struct dentry *de){
	if (!de) return;
	LOGDEBUG("--d_drop--", 1);
	list_del(&de->d_hash);
	list_init(&de->d_hash);
	LOGDEBUG("--end of d_drop--", 1);
}


void d_instantiate(struct dentry *de, DWORD ptr){
	if (list_empty(&de->d_hash)) 
		list_add(&de->d_hash, d_hash(de->d_parent, de->d_name.hash));
	de->d_vnode = ptr;
	de->d_neg = 0;
}

struct dentry *d_alloc(struct filesystem *fs, struct cstr *name, struct dentry *parent){
	struct dentry *de;

	LOGDEBUG("--d_alloc--", 1);
	de = (struct dentry *) _HeapAllocate (sizeof(struct dentry), HEAPZEROINIT);
	if (!de){ 
		LOGDEBUG("d_alloc: out of memory", 10);
		return NULL;
	}

	list_init(&de->d_lru);
	list_init(&de->d_subdir);
	list_init(&de->d_child);
	list_init(&de->d_hash);

	de->d_name.name = _HeapAllocate(name->len, 0);

	memcpy(de->d_name.name, name->name, name->len);
	de->d_name.len = name->len;
	de->d_name.hash = name->hash;

	de->d_ops = NULL;
	de->d_vnode_ops = NULL;

	if (parent){
		list_add(&de->d_child, &parent->d_subdir);
		de->d_parent = dget(parent);
		de->d_ops = parent->d_ops;
		de->d_vnode_ops = parent->d_vnode_ops;
	} else {
		de->d_parent = NULL;
	}

	de->d_flags = 0;
	de->d_refcount = 1;
	de->d_neg = 1;
	de->d_fs = fs;
	list_add(&de->d_hash, d_hash(de->d_parent, de->d_name.hash));
	LOGDEBUG("--end of d_alloc--", 1);
	return de;
}

struct dentry *d_alloc_root(pFilesystem fs, void *vnode){
	struct cstr str = {("root", 4, 0)};
	fs->f_root = d_alloc(fs, &str , NULL);
	
	if (!vnode && fs->f_root)
		fs->f_root->d_vnode = (DWORD) vnode;

	return fs->f_root;
}


struct dentry *d_getnextchild(struct dentry *dirde, struct dllist_head *le){	
	struct dentry *de;

	if (!le)
		le = dirde->d_subdir.next;
//	else
//		le = de->d_child.next;
		
	while (&dirde->d_subdir != le) {
		de = list_entry(le, struct dentry, d_child);
		print_dentry(de);
		if (!IS_NEGATIVE(de)) 
			return de;
		le = le->next;
	}

	return NULL;
}


struct dentry *ddd_cachelookup(struct dentry *parent, struct cstr *name, int flags){
	unsigned int hash = name->hash;
	struct dllist_head *le = d_hash(parent, hash);
	struct dllist_head *lh = NULL;
	struct dentry *de = NULL;

	LOGDEBUG("--ddd_cachelookup--", 1);

	if (list_empty(le)) {
		LOGDEBUG("--ddd_cachelookup list empty--", 10);
		goto out;
	}

	lh = le;
	le = lh->next;
	while(le != lh){
		de = list_entry(le, struct dentry, d_hash);
		if (!de){
			break;
		}
		le = le->next;
		dget(de);
		if (ddd_comparename(&de->d_name, name, flags)){
			if (de && de->d_ops && de->d_ops->d_revalidate){
				if (!IS_NEGATIVE(de) && !de->d_ops->d_revalidate(de) && !ddd_invalidate(de)){
					dput(de);
					break;
				}
			}
			goto out;
		}
		dput(de);
	}
	de = NULL;
out:
	LOGDEBUG("--end of ddd_cachelookup--", 1);
	return de;

}

int ddd_invalidate(struct dentry *de){

	if (!list_empty(&de->d_subdir))
		d_shrink_cache_parent(de);
	
	if (de->d_refcount > 1)
		return 1;

	d_drop(de);

	return 0;
}

struct dentry *ddd_get(struct dentry *de){
	if (de)		
		de->d_refcount++;
	return de;
}


void ddd_clean_lru(int degree){
	struct dllist_head *le;
	struct dentry *de;
	int i = 0, d = 0;
	LOGDEBUG("---ddd_clean_lru---", 1);

	if (lru_list_count <= LRU_MAX || degree<=0) return;
	i = LRU_MAX;
	i = i / degree;

	le = &lru_list;
	le = le->prev;
	while (i!=0){
		de = list_entry(le, struct dentry, d_lru);
		le = le->prev;
		i--;
		if (de->d_refcount) continue;
		list_del(&de->d_lru);
		list_init(&de->d_lru);
		lru_list_count--;
		d++;
		d_prune_one_dentry(de);
	}
	printf("ddd_clean_lru %x cleaned",d);
	
	LOGDEBUG("---end of ddd_clean_lru---", 1);
}

void ddd_switchhash(struct dentry *de1, struct dentry *de2){
	struct dentry *de = NULL;
	char *str=NULL;
	LOGDEBUG("---ddd_rehash---", 1);

	if (!de1 || !de2){
		LOGDEBUG("---ddd_rehash de1 or de2 NULL---", 10);
		return;
	}

	_asm pushf
	_asm cli

	str = de1->d_name.name;
	de1->d_name.name = de2->d_name.name;
	de2->d_name.name = str;
	de1->d_name.len = de2->d_name.len;
	de1->d_name.hash = name_hash(&de1->d_name);
	dput(de1->d_parent);
	de1->d_parent = dget(de2->d_parent);
	

	list_del(&de1->d_hash);
	list_init(&de1->d_hash);
	d_drop(de2);

	list_add(&de1->d_hash, d_hash(de1->d_parent, de1->d_name.hash));	

	_asm popf			

	LOGDEBUG("---end of ddd_rehash---", 1);
	return;
}

void ddd_delete(struct dentry *de){
	LOGDEBUG("--ddd_delete--", 1);
	if (de){
		if (de->d_name.name) _HeapFree(de->d_name.name, 0);
		if (de->d_ops && de->d_ops->d_vfree)
			de->d_ops->d_vfree(de);
		_HeapFree(de, 0);
	}
	else
		LOGDEBUG("ddd_delete: NULL pointer", 10);
	LOGDEBUG("-- end of ddd_delete--", 1);	
}



int ddd_comparename(struct cstr *name1, struct cstr *name2, int flags){
	unsigned int i = 0;
	PCHAR p, q;
	char pc, qc;

	if (name1->len != name2->len) return 0;

	if (flags == CASE_INSENSITIVE){
		LOGDEBUG("---ddd_comparenames - case insensitive compare", 10);

		p = name1->name;
		q = name2->name;
		for (i = 0; i < name1->len; i++) {
			pc = *p++; qc = *q++;

			if (pc >= 'A' && pc <= 'Z') pc += 'a' - 'A';
			if (qc >= 'A' && qc <= 'Z') qc += 'a' - 'A';
			if (pc != qc) return 0;
		}
		return 1;
		
	}
	else{
		return !memcmp(name1->name, name2->name, name1->len);
	}
}


unsigned int VXDINLINE partial_hash(unsigned int c, unsigned int prevhash){
	unsigned int hash;	

	if (c >= 'A' && c <= 'Z')
		c += 'a' - 'A';	

	__asm
	{
	 mov eax, prevhash
	 rol eax, 4
	 xor edx, edx
	 mov edx, c
	 xor eax, edx
	 mov DWORD PTR hash, eax
	}

	return hash;
}


unsigned int name_hash(struct cstr *name){
    unsigned int hash = INIT_HASH;
	int len = name->len;
	char *c = name->name;
    while (len--)
            hash = partial_hash(*c++, hash);

    return hash;
}


void d_prune_one_dentry(struct dentry *de){
	struct dentry *parent;

	list_del(&de->d_hash);
	list_init(&de->d_hash);
	list_del(&de->d_child);
	list_init(&de->d_child);

	if (de->d_ops && de->d_ops->d_vput)
		de->d_ops->d_vput(de);
	
	parent = de->d_parent;
	ddd_delete(de);
	dput(parent);
}

void d_shrink_cache(){
	struct dllist_head *lh, *le;
	struct dentry *de;

	LOGDEBUG("d_shrink_cache", 1);

repeat:
	lh = &lru_list;
	le = lh->next;
	while(le != lh){
		de = list_entry(le, struct dentry, d_lru);
		le = le->next;
		if (de->d_refcount)
			continue;
		list_del(&de->d_lru);
		list_init(&de->d_lru);
		d_prune_one_dentry(de);
		goto repeat;
	}
}
 

int d_select_parent(struct dentry* parent){
	struct dentry *this_parent = parent;
	struct dllist_head *next;
	int found = 0;

repeat:
	next = this_parent->d_subdir.next;
resume:
	while (next != &this_parent->d_subdir) {
		struct dllist_head *tmp = next;
		struct dentry *de = list_entry(tmp, struct dentry, d_child);
		next = tmp->next;
		if (!de->d_refcount) {
			list_del(&de->d_lru);
			list_add(&de->d_lru, lru_list.prev);
			found++;
		}
		if (!list_empty(&de->d_subdir)) {
			this_parent = de;
			goto repeat;
		}
	}
	if (this_parent != parent) {
		next = this_parent->d_child.next; 
		this_parent = this_parent->d_parent;
		goto resume;
	}
	return found;
}

void d_prune_cache(int count){
	for (;;) {
		struct dentry *de;
		struct dllist_head *tmp = lru_list.prev;

		if (tmp == &lru_list)
			break;
		list_del(tmp);
		list_init(tmp);
		de = list_entry(tmp, struct dentry, d_lru);
		if (!de->d_refcount) {
			d_prune_one_dentry(de);
			if (!--count)
				break;
		}
	}
}

void d_shrink_cache_parent(struct dentry *parent){
	int found;

	while ((found = d_select_parent(parent)) != 0)
		d_prune_cache(found);
}


