/* file.c -- Windows 95 IFS routines for manipulating open files
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include <string.h>
#pragma intrinsic (strcpy,strlen, memcpy)

#include "cache.h"
#include <vmm.h>
#include "vmmextra.h"

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static struct hndlmisc hf_tbl = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_HNDLMISC,

	FileSeek,
	CloseFile,
	CommitFile,
	LockFile,				/* LockFile,*/
	FileDateTime,			/* FileDateTime, */
	NamedPipeRequest,		/* NamedPipeRequest */
	NamedPipeHandleInfo,	/* NamedPipeHandleInfo, */
	EnumerateHandle			/* EnumerateHandle */
};

int OpenFile (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	struct dentry *de = NULL;
	char *p = NULL;
	int res = 0;
	LOGDEBUG("OpenFile", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	p = to_BCS(pir->ir_upath+3, ucodelen(pir->ir_upath)-3);
	res = namei(fs, p, &de, CASE_INSENSITIVE);	
	_HeapFree(p,0);	

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_open) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_open(pir, de);
	else res = ERROR_INVALID_FUNCTION;

	if (!res){
		hfunc_t phf = pir->ir_hfunc;

		if (fs->f_root->d_vnode_ops->default_file_ops->fo_read) phf->hf_read = fs->f_root->d_vnode_ops->default_file_ops->fo_read;
		else return pir->ir_error = ERROR_INVALID_PARAMETER;

		if (fs->f_root->d_vnode_ops->default_file_ops->fo_write) phf->hf_write = fs->f_root->d_vnode_ops->default_file_ops->fo_write;
		else return pir->ir_error = ERROR_INVALID_PARAMETER;

		phf->hf_misc = &hf_tbl;		
	}

	return pir->ir_error = res;
}

int FileSeek (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("FileSeek", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_seek) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_seek(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int CloseFile (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("CloseFile", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_close) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_close(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int CommitFile (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("CommitFile", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_commit) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_commit(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int LockFile (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("LockFile", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_lock) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_lock(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int FileDateTime (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("FileDateTime", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_datetime) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_datetime(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int NamedPipeRequest (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("NamedPipeRequest", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_namedpiperequest) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_namedpiperequest(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int NamedPipeHandleInfo (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("NamedPipeHandleInfo", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_namedpipehandleinfo) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_namedpipehandleinfo(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int EnumerateHandle (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("EnumerateHandle", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops->default_file_ops && fs->f_root->d_vnode_ops->default_file_ops->fo_enumeratehandle) 
		res = fs->f_root->d_vnode_ops->default_file_ops->fo_enumeratehandle(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

