/* find.c -- Windows95 IFS routines for FindFirstFile/FindNextFile
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include "cache.h"

#include <string.h>
#pragma intrinsic (memcpy, memset, strcmp, strlen, strcpy)


#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static struct hndlmisc hfFind = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_HNDLMISC,

	ErrorStub,				/* FileSeek */
	FindFileClose,			/* Close */
	ErrorStub,				/* Commit */
	ErrorStub,				/* LockFile,*/
	ErrorStub,				/* FileDateTime, */

	ErrorStub,				/* NamedPipeRequest */
	ErrorStub,				/* NamedPipeHandleInfo, */
	EnumerateFind,			/* EnumerateHandle XXX -- should be handled*/
};

int SearchFile (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	struct dentry *de = NULL;
	char *p = NULL;
	int res = 0;

	LOGDEBUG("SearchFile", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (pir->ir_flags == SEARCH_FIRST){
		p = to_BCS(pir->ir_upath+3, ucodelen(pir->ir_upath)-3);
		res = namei(fs, p, &de, CASE_INSENSITIVE);	
		_HeapFree(p,0);
	}

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_search) 
		res = fs->f_root->d_vnode_ops->vo_search(pir, de);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}


int FindFirstFile (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	struct dentry *de = NULL;
	char *p = NULL;
	int res = 0;

	LOGDEBUG("FindFirstFile", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	
	p = to_BCS(pir->ir_upath+3, ucodelen(pir->ir_upath)-3);
	res = namei(fs, parent_path(p), &de, CASE_INSENSITIVE);	
	_HeapFree(p,0);

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_findopen) 
		res = fs->f_root->d_vnode_ops->vo_findopen(pir, de);
	else res = ERROR_INVALID_FUNCTION;

	if (!res){
		hfunc_t phf = pir->ir_hfunc;

		if (fs->f_root->d_vnode_ops->vo_findnext) phf->hf_read = fs->f_root->d_vnode_ops->vo_findnext;
		else return pir->ir_error = ERROR_INVALID_PARAMETER;

		phf->hf_misc = &hfFind;	
	}

	return pir->ir_error = res;
}


int FindFileClose (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("FindFileClose", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_findclose) 
		res = fs->f_root->d_vnode_ops->vo_findclose(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int EnumerateFind (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("EnumerateFind", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_enumfind) 
		res = fs->f_root->d_vnode_ops->vo_enumfind(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int ErrorStub (pioreq pir)
{
	LOGDEBUG("ErrorStub", 1);
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}