
#include <string.h>
#pragma intrinsic (memset, strcpy, strlen, strcmp)

#include "cache.h"


#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static char copyright[] = "Copyright 1999 Marc Schnieder\n";

dllist_chain filesystems;

#pragma warning(disable: 4035)

int connect_fs(pFilesystem fs){
	netuse_info *pninfo;
	use_info_2 *puinfo;
	int res = 0;

	pninfo = (netuse_info *) _HeapAllocate(sizeof(netuse_info), 0);
	if (!pninfo){
		return ERROR_NOT_ENOUGH_MEMORY;
	}

	puinfo = (use_info_2 *) _HeapAllocate(sizeof(use_info_2), 0);
	if (!puinfo) {
		_HeapFree(pninfo, 0);
		return ERROR_NOT_ENOUGH_MEMORY;
	}

	pninfo->nu_data = (PVOID) puinfo;
	pninfo->nu_info = 0;
	pninfo->nu_flags = FSD_NETAPI_USELFN;
	pninfo->nu_rsvd = 0xff;

	memset(puinfo, 0, sizeof(*puinfo));
	strcpy( puinfo->ui2_local, "X:" );
	puinfo->ui2_local[0] = fs->f_letter;
	puinfo->ui2_remote = fs->f_path;
	puinfo->ui2_password = NULL;

#define USE_DISKDEV 0
	puinfo->ui2_asg_type = USE_DISKDEV;
	puinfo->ui2_res_type = IFSFH_RES_UNC;

	LOGDEBUG(&fs->f_letter, 1);

	_asm pushf
    _asm cli
	list_init(&fs->f_list);
	list_add(&fs->f_list, &filesystems);
	_asm popf

	res = IFSMgr_UseAdd(NULL, uiProviderId, pninfo);

	_HeapFree (puinfo, 0);
	_HeapFree (pninfo, 0);

	if (res){
		_asm pushf
	    _asm cli
		list_del(&fs->f_list);
		_asm popf
	}

	return res;
}

int hook_fs(pFilesystem fs){
	int res = 0;
	struct dllist_head *le = &hooks;

	_asm pushf
	_asm cli

	if (!orig){
		fs->f_type.fst_prev_hook = call_orig;
		orig = init;
	}
	else {
		pFilesystem _fs = list_entry(le->next, struct filesystem, f_list);
		fs->f_type.fst_prev_hook = _fs->f_type.fst_hook_func;
	}

	list_add(&fs->f_list, &hooks);

	if (fs->f_letter >= 'a' && fs->f_letter <= 'z')
		fs->f_letter -= 'a' - 'A';

	_asm popf

	return res;
}

DWORD register_filesystem(pFilesystem _f)
{
	DWORD res = 0;
	struct filesystem  *fs = _f;

	LOGDEBUG ("---register_filesystem---", 1);

	LOGDEBUG(fs->f_path, 1);

	switch (fs->f_type.fst_type){

	case local:	
		return ERROR_INVALID_PARAMETER;

	case remote:
		if (!fs->f_type.fst_read_super || !fs->f_type.fst_put_super) 
			return ERROR_INVALID_PARAMETER;		

		res = connect_fs(fs);
		break;

	case hook:
		if (!fs->f_type.fst_hook_func)
			return ERROR_INVALID_PARAMETER;		

		res = hook_fs(fs);
		break;

	default:
		return ERROR_INVALID_PARAMETER;
	}

	LOGDEBUG ("---end of register_filesystem---", 1);
	return res ;
}

int disconnect_fs(pFilesystem fs){
	int res = 0;
	netuse_info *pninfo = NULL;
	char p[3];

	pninfo = (netuse_info *) _HeapAllocate(sizeof(netuse_info), 0);
	if (!pninfo)
		return ERROR_NOT_ENOUGH_MEMORY;

	LOGDEBUG(&fs->f_letter, 1);
	p[0] = fs->f_letter;
	p[1] = ':';
	p[2] = '\0';
	pninfo->nu_data = (PVOID)p;
	pninfo->nu_info = 3;
	pninfo->nu_flags = FSD_NETAPI_USELFN;
	pninfo->nu_rsvd = 0xff;

	res = IFSMgr_UseDel(NULL, uiProviderId, pninfo);
	if (res){
		printf("error disconnect %x", res);
//		return res;
	}

	_asm pushf
	_asm cli
	list_del(&fs->f_list);
	list_init(&fs->f_list);
	_asm popf

	return res;
}

int unhook_fs(pFilesystem fs){
	int res = 0;

	_asm pushf
	_asm cli
	list_del(&fs->f_list);
	list_init(&fs->f_list);
	_asm popf

	if (orig && hooks.next == &hooks)
		orig = NULL;

	return res;
}

DWORD unregister_filesystem(pFilesystem _fs)
{
	int res = 0;
	struct filesystem *fs = _fs;

    LOGDEBUG("unregister_filesystem", 1);

	switch (fs->f_type.fst_type){

	case local:	
		return ERROR_INVALID_PARAMETER;

	case remote:
		if (!checkFS(fs))
			return ERROR_INVALID_PARAMETER;

		res = disconnect_fs(fs);
		break;

	case hook:
		if (!checkHS(fs))
			return ERROR_INVALID_PARAMETER;

		res = unhook_fs(fs);
		break;

	default:
		return ERROR_INVALID_PARAMETER;
	}
	
    LOGDEBUG("end of unregister_filesystem", 1);
	return 0;
}

pFilesystem VDCACHE_New_FS()
{
	return (struct filesystem *) _HeapAllocate(sizeof(struct filesystem), HEAPZEROINIT);
}

void VDCACHE_Del_FS(pFilesystem f)
{
	_HeapFree(f, 0);
}


#pragma warning(default: 4035)

/* symbols defined in CACHEvxd.asm */
extern DRP VDCACHE_DRP;
extern DWORD BSSstart, BSSend;
DWORD Debuglevel;

void OnAsyncEvent(PAEP);

BOOL OnDeviceInit ()
{
	HKEY	hkey;
	char	regpath[256];
	DWORD	dl;
	DWORD	cbdata = sizeof(Debuglevel);

	/* clear bss segment */
	memset (&BSSstart, 0, ((char *)&BSSend) - ((char *)&BSSstart));

	/* check that IOS is loaded */
	if (!IOS_Get_Version())
		return FALSE;

		/* register with IOS

	   We have to supply an async event routine.  If we 
	   registered as a FSD (filesystem driver), then this 
	   would be unnecessary.  However, according to Oney, 
	   Microsoft's FSDs register as TSD's (type-specific
	   drivers).  These _do_ receive async events from 
	   the IOS.

	   Go figure.
    */
    VDCACHE_DRP.DRP_aer = (PVOID) OnAsyncEvent;
	VDCACHE_DRP.DRP_feature_code |= DRP_FC_4_SECS;
	IOS_Register(&VDCACHE_DRP);
	if ((VDCACHE_DRP.DRP_reg_result == DRP_ABORT) ||
		(VDCACHE_DRP.DRP_reg_result == DRP_INVALID_LAYER))
		return FALSE;

	/* register with IFSMgr */
	uiProviderId = IFSMgr_RegisterNet (ConnectNetResource, 
		IFSMGRVERSION, uiNetId);

	if (uiProviderId == -1)
		return FALSE;

	dl = 100;

	strcpy(regpath, "System\\CurrentControlSet\\Services\\VxD\\VDCACHE");
	_RegOpenKey(HKEY_LOCAL_MACHINE, regpath, &hkey);
	_RegQueryValueEx(hkey, "Debuglevel", NULL, NULL, (PBYTE) &dl, &cbdata);
	_RegCloseKey(hkey);

	Debuglevel = dl;

	list_init(&filesystems);
	list_init(&hooks);

	d_init();

	init = IFSMgr_InstallFileSystemApiHook(VdcacheHookProc);

	return TRUE;
}


void OnAsyncEvent (PAEP pAEP)
{
	if (pAEP->AEP_func == AEP_4_SECS) {
	}
	pAEP->AEP_result = AEP_SUCCESS;
}

int inUseCount = 0;


BOOL OnDeviceExit ()
{
	LOGDEBUG("OnDeviceExit", 1);
	IFSMgr_RemoveFileSystemApiHook(VdcacheHookProc);
	LOGDEBUG("OnDeviceExit OK", 1);
	return TRUE;
}

struct filesystem *findFS(char *letter)
{
	struct dllist_head *lh, *le;
	struct filesystem *fs;
	_asm pushf
	_asm cli
	le = &filesystems;
	lh = le;
	for ( le = lh->next ; le != lh ; le = le->next ) {
		fs = list_entry(le, struct filesystem, f_list);
		LOGDEBUG(fs->f_path, 1);LOGDEBUG(letter, 1);
		if (fs->f_letter == *letter  || !strcmp(fs->f_path, letter)) {
			LOGDEBUG("found", 1);
			_asm popf
			return fs;
		}
	}
	LOGDEBUG("not found", 1);
	_asm popf
	return NULL;
}

int checkFS(struct filesystem *_fs)
{
	struct dllist_head *lh, *le;
	struct filesystem *fs;
	_asm pushf
	_asm cli
	le = &filesystems;
	lh = le;
	for ( le = lh->next ; le != lh ; le = le->next ) {
		fs = list_entry(le, struct filesystem, f_list);
		if (fs == _fs) {
			LOGDEBUG("found checkFS", 1);
			_asm popf
			return 1;
		}
	}
	LOGDEBUG("not found checkFS", 1);
	_asm popf
	return 0;
}

int checkHS(struct filesystem *_fs)
{
	struct dllist_head *lh, *le;
	struct filesystem *fs;
	_asm pushf
	_asm cli
	le = &hooks;
	lh = le;
	for ( le = lh->next ; le != lh ; le = le->next ) {
		fs = list_entry(le, struct filesystem, f_list);
		if (fs == _fs) {
			LOGDEBUG("found checkHS", 1);
			_asm popf
			return 1;
		}
	}
	LOGDEBUG("not found checkHS", 1);
	_asm popf
	return 0;
}