/* misc.c -- miscellaneous routines
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include <string.h>
#pragma intrinsic (strlen, strcpy, memcpy)

#include "CACHE.h"
#include "vwin32wrap.h"
#include "vmmextra.h"

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG

char ucodebuf[400] = "foo";

void logucode (PWORD pu)
{
#ifdef DEBUG
	PCHAR pc = ucodebuf;

	for ( ; *pu; pu++, pc++)
		if (*pu < 256)
			*pc = (char) *pu;
		else
			*pc = '?';
	*pc = '\0';
	LOGDEBUG(ucodebuf, 10);
#endif
}

int ucodelen(PWORD pu){
	int i=0;
	for (i=0; *pu; pu++, i++);
	return i;	
}

void logmsg (char *buffer)
{
#ifdef DEBUG
	DWORD handle, action, size, nWritten;
	int err;

	err = R0_OpenCreateFile (FALSE, ACCESS_WRITEONLY | SHARE_DENYWRITE,
		0, (BYTE) ACTION_OPENALWAYS, (BYTE) R0_NO_CACHE, "C:\\VDCACHE.LOG", &handle, &action);
	if (err)
		return;
	if (R0_GetFileSize (handle, &size))
		goto error;
	R0_WriteFile (FALSE, handle, strlen((const char *)buffer), size, (PBYTE) buffer, &nWritten);
	if (R0_GetFileSize (handle, &size))
		goto error;
	R0_WriteFile (FALSE, handle, 2, size, (const unsigned char *)"\r\n", &nWritten);
error:
	R0_CloseFile (handle);
#endif
}

void logcycles (char *buffer)
{
#ifdef DEBUG
	DWORD handle, action, size, nWritten;
	int err;
	if (Debuglevel < 10) return;

	err = R0_OpenCreateFile (FALSE, ACCESS_WRITEONLY | SHARE_DENYWRITE,
		0, (BYTE) ACTION_OPENALWAYS, (BYTE) R0_NO_CACHE, "C:\\vd_cycles.log", &handle, &action);
	if (err)
		return;
	if (R0_GetFileSize (handle, &size))
		goto error;
	R0_WriteFile (FALSE, handle, strlen((const char *)buffer), size, (PBYTE) buffer, &nWritten);
	if (R0_GetFileSize (handle, &size))
		goto error;
	R0_WriteFile (FALSE, handle, 2, size, (const unsigned char *)"\r\n", &nWritten);
error:
	R0_CloseFile (handle);
#endif
}

void fmt_dword (char *p, DWORD dw)
{
	int i;
	unsigned int j;

	for (i = 7; i >= 0; i--) {
		j = dw & 0xf;
		p[i] = "0123456789abcdef"[j];
		dw = dw >> 4;
	}
}

void print_pir (pioreq p)
{
	DWORD *pdw = (DWORD *) p;
	static char *m[] = {
		"xxxxxxxx  attr             xxxxxxxx  flags/user/sfn",
		"xxxxxxxx  pid              xxxxxxxx  ppath",
		"xxxxxxxx  aux1             xxxxxxxx  data",
		"xxxxxxxx  options/error    xxxxxxxx  rh",
		"xxxxxxxx  fh               xxxxxxxx  pos",
		"xxxxxxxx  aux2             xxxxxxxx  aux3",
		"xxxxxxxx  pev              xxxxxxxx  fsdata[0]"};
	int i;

	for (i = 0; i < 7; i++) {
		fmt_dword(m[i], *pdw++);
		fmt_dword(m[i] + 27, *pdw++);
		logmsg (m[i]);
	}
}

void printf(char *fmt, ...)
{
#ifdef DEBUG
	static char buf[200] = "   ";
	static PVMMMUTEX printf_mtx = NULL;
	int *args, i;

	if (!printf_mtx) {
		printf_mtx = _CreateMutex(0, 0);
		if (!printf_mtx)
			LOGDEBUG("INIT printf--cannot get MUTEX", 10);
	}

	if (printf_mtx)
		_EnterMutex (printf_mtx, BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);

	args = ((int *) &fmt) + 1;

	for (i = 0; *fmt && i < 180; fmt++, i++) {
		if ((*fmt != '%') || (*(fmt+1) != 'x')) 
			buf[i] = *fmt;
		else {
			fmt_dword (&buf[i], *args++);
			i += 7;
			fmt++;
		}
	}
	buf[i] = 0;
	LOGDEBUG(buf, 1);
	if (printf_mtx)
		_LeaveMutex (printf_mtx);	
#endif
}

void print_dentry(struct dentry *de){
	char name[MAX_PATH];

	if (!de) return;

	memcpy(name, de->d_name.name, de->d_name.len);
	name[de->d_name.len] = '\0';
	LOGDEBUG("print dentry", 1);
	LOGDEBUG(name, 1);
	printf("dentry %x: flags %x, neg %x, refcount %x", de, de->d_flags, de->d_neg, de->d_refcount);
	
}
void printcyclesf(char *fmt, ...)
{
#ifdef DEBUG
	static char buf[200] = "   ";
	static PVMMMUTEX printf_mtx = NULL;
	int *args, i;

	if (Debuglevel < 10) return;

	if (!printf_mtx) {
		printf_mtx = _CreateMutex(0, 0);
		if (!printf_mtx)
			LOGDEBUG("INIT printf--cannot get MUTEX", 10);
	}

	if (printf_mtx)
		_EnterMutex (printf_mtx, BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);

	args = ((int *) &fmt) + 1;

	for (i = 0; *fmt && i < 180; fmt++, i++) {
		if ((*fmt != '%') || (*(fmt+1) != 'x')) 
			buf[i] = *fmt;
		else {
			fmt_dword (&buf[i], *args++);
			i += 7;
			fmt++;
		}
	}
	buf[i] = 0;
	logcycles (buf);
	if (printf_mtx)
		_LeaveMutex (printf_mtx);	
#endif
}
void print_ppath(ParsedPath* pp){
	PathElement *pe;	
	pe = pp->pp_elements;	
//	while (pe != IFSLastElement(pp)){
//		logmsg ("test");
		logucode (pe->pe_unichars);
//		pe = IFSNextElement(pe);
//	}
}


PUSHORT to_ucode(PCHAR name){
	PUSHORT  Uniname = _HeapAllocate(strlen(name)*2, 0);
	if (Uniname)
		BCSToUni(Uniname, name, strlen(name), BCS_OEM);
	return Uniname;
}

PCHAR to_BCS(PUSHORT name, int length){
	char *BCSname;
	if (length == 0) return NULL;
	BCSname = _HeapAllocate(length+1, 0);
	if (length > MAX_PATH) return NULL;
	if (BCSname) {
		UniToBCS(BCSname, name, 2*length, length, BCS_OEM);
		BCSname[length] = '\0';
	}
	return BCSname;
}

char *parent_path(char *path){
	int i = strlen(path);
	PCHAR p = path+i-1;
	
	if (i==0) return NULL;

	for ( ; p != path; p--){
		if (*p == '\\' || *p == '/'){
			*p = '\0';
			return path;
		}
	}
	return NULL;
}

char *tofilename(char *path){
	int i = strlen(path);
	PCHAR p = path+i-1;
	for ( ; p != path; p--){
		if (*p == '\\' || *p == '/'){
			return p+1;
		}
	}	
	return path;
}

char *tolower(char *tar, const unsigned char *src){
	int i=0;

	for (i = 0; src[i]; i++) {
			tar[i] = src[i];
			if (tar[i] >= 'A' && tar[i] <= 'Z')
				tar[i] += 'a' - 'A';
	}
	tar[i]='\0';
	return tar;
}

char *memicpy(char *tar, char *src, int len){
	/* copy and lower len bytes */
	int i=0;

	for (i = 0; i < len; i++) {
			tar[i] = src[i];
			if (tar[i] >= 'A' && tar[i] <= 'Z')
				tar[i] += 'a' - 'A';
	}
	return tar;
}

void logcachemsg (char *buffer)
{
#ifdef CACHE
	DWORD handle, action, size, nWritten;
	int err;

	if (Debuglevel < 1000) return;

	err = R0_OpenCreateFile (FALSE, ACCESS_WRITEONLY | SHARE_DENYWRITE,
		0, (BYTE) ACTION_OPENALWAYS, (BYTE) R0_NO_CACHE, "C:\\VD_CACHE.LOG", &handle, &action);
	if (err)
		return;
	if (R0_GetFileSize (handle, &size))
		goto error;
	R0_WriteFile (FALSE, handle, strlen((const char *)buffer), size, (PBYTE) buffer, &nWritten);
	if (R0_GetFileSize (handle, &size))
		goto error;
	R0_WriteFile (FALSE, handle, 2, size, (const unsigned char *)"\r\n", &nWritten);
error:
	R0_CloseFile (handle);

#endif
}

void logcache_lrulist(struct dllist_head *head)
{
#ifdef DEBUG
	struct dllist_head *lh, *le;
	struct dentry *de;

	if (Debuglevel < 1000) return;

	logcachemsg("---begin log lru list---");
	if (!list_empty(head)){
		_asm pushf
		_asm cli
		le = head;
		lh = le;
		for ( le = lh->next ; le != lh ; le = le->next ) {
			de = list_entry(le, struct dentry, d_lru);
//			if (de)
//				printf_cache("Cnode: %x, fid: %x %x %x, ocount: %x", de->d_cnode, de->d_cnode->c_fid.Volume,
//					de->d_cnode->c_fid.Vnode, de->d_cnode->c_fid.Unique, de->d_cnode->c_ocount);		
		}
		_asm popf
	}
	logcachemsg("---end log lru list---");
#endif
}

void printf_cache(char *fmt, ...)
{
#ifdef DEBUG
	static char buf[200] = "   ";
	static PVMMMUTEX printf_mtx = NULL;
	int *args, i;

	if (Debuglevel < 1000) return;

	if (!printf_mtx) {
		printf_mtx = _CreateMutex(0, 0);
		if (!printf_mtx)
			logcachemsg ("INIT printf--cannot get MUTEX");
	}

	if (printf_mtx)
		_EnterMutex (printf_mtx, BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);

	args = ((int *) &fmt) + 1;

	for (i = 0; *fmt && i < 180; fmt++, i++) {
		if ((*fmt != '%') || (*(fmt+1) != 'x')) 
			buf[i] = *fmt;
		else {
			fmt_dword (&buf[i], *args++);
			i += 7;
			fmt++;
		}
	}
	buf[i] = 0;
	logcachemsg (buf);
	if (printf_mtx)
		_LeaveMutex (printf_mtx);	
#endif
}