#ifndef _VDCACHE_H_
#define _VDCACHE_H_

#ifndef dllist_chain
struct dllist_head {	
	struct dllist_head *next;
	struct dllist_head *prev;
};
#define dllist_chain struct dllist_head
#endif

enum fs_type {local, remote, hook};
typedef enum fs_type fs_type;

typedef struct filesystem *pFilesystem;

typedef int _cdecl SuperFunc(pioreq pir, pFilesystem fs);
typedef SuperFunc *pSuperFunc;

typedef int _cdecl HookFunc(pIFSFunc pfn, int fn, int Drive, int ResType, 
							int CodePage, pioreq pir,
							struct dentries *des);
typedef HookFunc *pHookFunc;

struct filesystemtype{
	fs_type			fst_type;
	union {
		pSuperFunc	fst_read_super;
		pHookFunc	fst_hook_func;
	};
	union {
		pSuperFunc		fst_put_super;
		pHookFunc	fst_prev_hook;
	};

};
typedef struct filesystemtype Filesystemtype;

struct cstr {
	char *name;
	unsigned int len;
	unsigned int hash;
};

struct filesystem {
	DWORD			f_hdl;
	char			f_letter;
	char			*f_path;
	struct dentry	*f_root;
	Filesystemtype	f_type;
	dllist_chain	f_list;
};

struct file_operations{
	int (*fo_seek) (pioreq);
	int (*fo_read) (pioreq);
	int (*fo_write) (pioreq);
	int (*fo_open) (pioreq, struct dentry *);
	int (*fo_close) (pioreq);
	int (*fo_commit) (pioreq);
	int (*fo_lock) (pioreq);
	int (*fo_datetime) (pioreq);
	int (*fo_namedpiperequest) (pioreq);
	int (*fo_namedpipehandleinfo) (pioreq);
	int (*fo_enumeratehandle) (pioreq);
};

struct vnode_operations{
	struct file_operations *default_file_ops;
	int (*vo_create) (pioreq, struct dentry *);
	int (*vo_lookup) (struct dentry **, struct cstr *, int);
	int (*vo_delete) (pioreq, struct dentry *);
	int (*vo_dir) (pioreq, struct dentry *);
	int (*vo_rename) (pioreq, struct dentry *, struct dentry *);
	int (*vo_flush) (pioreq);
	int (*vo_fileattrib) (pioreq, struct dentry *);
	int (*vo_getdiskinfo) (pioreq);
	int (*vo_search) (pioreq, struct dentry *);
	int (*vo_query) (pioreq);
	int (*vo_uncpipereq) (pioreq);
	int (*vo_ioctl16drive) (pioreq);
	int (*vo_getdiskparams) (pioreq);
	int (*vo_dasdio) (pioreq);
	int (*vo_findopen) (pioreq, struct dentry *);
	int (*vo_findnext) (pioreq);
	int (*vo_findclose) (pioreq);
	int (*vo_enumfind) (pioreq);
};

struct dentry_operations {
	int  (*d_revalidate) (struct dentry *);
	void (*d_delete) (struct dentry *);
	int  (*d_hash) (struct dentry *, struct cstr *);
	void (*d_vput) (struct dentry *);
	void (*d_vfree) (struct dentry *);
};

struct dentry {
	USHORT						d_flags;
	struct filesystem			*d_fs;
	struct dentry_operations	*d_ops;
	struct vnode_operations		*d_vnode_ops;
	int							d_neg;									
	struct dentry				*d_parent;
	DWORD						d_vnode;
	struct cstr					d_name;
	USHORT						d_refcount;
	dllist_chain				d_subdir;		/* our children */
	dllist_chain				d_child;		/* parent's child list*/
	dllist_chain				d_hash;
	dllist_chain				d_lru;			/* lru cache */
};

typedef struct dentry *pDentry;

struct dentries{
	struct dentry *de1;
	struct dentry *de2;
};


int VXDINLINE call_prev_hook(pFilesystem _f, pIFSFunc pfn, int fn, int Drive, int ResType, 
							int CodePage, pioreq pir, struct dentries *des){
	return (*_f->f_type.fst_prev_hook)(pfn, fn, Drive, ResType, CodePage, pir, des);
}



#ifndef VDCACHE
#define VDCACHE_DEVICE_ID			0x1232
Begin_Service_Table(VDCACHE, none)
#define VDCACHE_Service(X) Declare_Service(X, none)
VDCACHE_Service(VDCACHE_Get_Version)
VDCACHE_Service(_register_filesystem)
VDCACHE_Service(_unregister_filesystem)
VDCACHE_Service(_VDCACHE_New_FS)
VDCACHE_Service(_VDCACHE_Del_FS)
VDCACHE_Service(_dget)
VDCACHE_Service(_dput)
VDCACHE_Service(_d_drop)
VDCACHE_Service(_d_lookup)
VDCACHE_Service(_d_switchhash)
VDCACHE_Service(_d_getnextchild)
VDCACHE_Service(_d_shrink_cache)
VDCACHE_Service(_d_prune_one_dentry)
VDCACHE_Service(_d_prune_cache)
VDCACHE_Service(_d_shrink_cache_parent)
VDCACHE_Service(_d_instantiate)
VDCACHE_Service(_d_rehash)
VDCACHE_Service(_d_alloc_root)
End_Service_Table(VDCACHE, none)

#pragma warning(disable: 4035)

static DWORD VDCACHE_Get_Version()
{
	DWORD res;
	VxDCall(VDCACHE_Get_Version)
	_asm jnc noerror
	_asm mov eax, 0h
	_asm noerror:
	_asm mov [res], eax
	return res;
}

DWORD __declspec(naked) VXDINLINE register_filesystem(pFilesystem _f)
{
	VxDJmp(_register_filesystem)	
}

DWORD __declspec(naked) VXDINLINE unregister_filesystem(pFilesystem _f)
{
	VxDJmp(_unregister_filesystem)
}

pFilesystem __declspec(naked) VXDINLINE VDCACHE_New_FS(void)
{
	VxDJmp(_VDCACHE_New_FS)
}
void __declspec(naked) VXDINLINE VDCACHE_Del_FS(pFilesystem f)
{
	VxDJmp(_VDCACHE_Del_FS)
}

pDentry __declspec(naked) VXDINLINE dget(struct dentry *de)
{
	VxDJmp(_dget)
}

DWORD __declspec(naked) VXDINLINE dput(struct dentry *de)
{
	VxDJmp(_dput)
}

DWORD __declspec(naked) VXDINLINE d_drop(struct dentry *de)
{
	VxDJmp(_d_drop)
}

DWORD __declspec(naked) VXDINLINE d_lookup(struct filesystem *fs, char *path, struct dentry *startde, struct dentry **de, int flags)
{
	VxDJmp(_d_lookup)
}

DWORD __declspec(naked) VXDINLINE d_switchhash(struct dentry *de1 , struct dentry *de2)
{
	VxDJmp(_d_switchhash)
}

pDentry __declspec(naked) VXDINLINE d_getnextchild(struct dentry *de, struct dllist_head *le)
{
	VxDJmp(_d_getnextchild)
}

DWORD __declspec(naked) VXDINLINE d_shrink_cache(void)
{
	VxDJmp(_d_shrink_cache)
}

DWORD __declspec(naked) VXDINLINE d_prune_one_dentry(struct dentry *de)
{
	VxDJmp(_d_prune_one_dentry)
}

DWORD __declspec(naked) VXDINLINE d_prune_cache(int count)
{
	VxDJmp(_d_prune_cache)
}

DWORD __declspec(naked) VXDINLINE d_shrink_cache_parent(struct dentry *de)
{
	VxDJmp(_d_shrink_cache_parent)
}

DWORD __declspec(naked) VXDINLINE d_instantiate(struct dentry *de, DWORD ptr)
{
	VxDJmp(_d_instantiate)
}

DWORD __declspec(naked) VXDINLINE d_rehash(struct dentry *de)
{
	VxDJmp(_d_rehash)
}

pDentry __declspec(naked) VXDINLINE d_alloc_root(pFilesystem fs, void *vnode)
{
	VxDJmp(_d_alloc_root)
}


#endif




#endif