/* vol.c -- Windows95 IFS routines for volume manipulation
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include "cache.h"

#include "string.h"
#pragma intrinsic(strlen, strcpy, memset, memcpy, strcmp)

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static struct volfunc vf = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_VOLFUNC,

	DeleteFile,				/* DeleteFile */
	Dir,					/* Dir */
	FileAttributes,			/* FileAttributes */
	FlushVolume,			/* FlushVolume */
	GetDiskInfo,			/* GetDiskInfo */

	OpenFile,				/* OpenFile */
	RenameFile,				/* RenameFile */
	SearchFile,				/* SearchFile */
	QueryResourceInfo,		/* QueryResourceInfo */
	DisconnectResource,		/* DisconnectResource */

	UNCNamedPipeRequest,	/* UNCNamedPipeReq */
	Ioctl16Drive,			/* Ioctl16Drive */
	GetDiskParams,			/* GetDiskParms */
	FindFirstFile,			/* FindFirstFile */
	DirectDiskIO			/* DirectDiskIO */
};

UINT uiNetId = 0x001b0000;
UINT uiProviderId = 0;

int ConnectNetResource (pioreq pir)
{
	struct filesystem *fs;
	char *bcs;
	LOGDEBUG("ConnectNetResource", 1);

	bcs = to_BCS(pir->ir_upath, ucodelen(pir->ir_upath));
	fs = findFS(bcs);
	_HeapFree(bcs, 0);
	if (!fs) {
		pir->ir_error = ERROR_NOT_CONNECTED;
	}else if (fs->f_type.fst_read_super) {

		pir->ir_error = fs->f_type.fst_read_super(pir, fs);

	}else {
		pir->ir_error = ERROR_NOT_SUPPORTED;
	}

	pir->ir_vfunc = &vf;
	if (pir->ir_rh)
		fs->f_hdl = (DWORD) pir->ir_rh;
	else fs->f_hdl = NULL;
	pir->ir_rh = fs;
	LOGDEBUG("end of ConnectNetResource", 1);
	return pir->ir_error;
}


int DisconnectResource (pioreq pir)
{
	struct filesystem *fs;
	char *bcs;
	LOGDEBUG("DisconnectResource", 1);

	fs = (struct filesystem *)pir->ir_rh;
	if (!checkFS(fs)){
		bcs = to_BCS(pir->ir_upath, ucodelen(pir->ir_upath));
		fs = findFS(bcs);
		_HeapFree(bcs, 0);
		if (!fs) {
			return pir->ir_error = ERROR_CONNECTION_INVALID;
		}
	}

	if (fs && fs->f_type.fst_put_super)
		fs->f_type.fst_put_super(pir, fs);
	else pir->ir_error = ERROR_INVALID_FUNCTION;

	if (!pir->ir_error){
		_asm pushf
		_asm cli
		list_del(&fs->f_list);
		_asm popf
		_HeapFree(fs->f_path, 0);
//		_HeapFree(fs, 0);
	}

	return pir->ir_error;
}

int DeleteFile (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	char *p = NULL;
	struct dentry *de = NULL;
	int res = 0;

	LOGDEBUG("DeleteFile", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	p = to_BCS(pir->ir_upath+3, ucodelen(pir->ir_upath)-3);
	res = namei(fs, p, &de, CASE_INSENSITIVE);
	_HeapFree(p,0);

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_delete) 
		res = fs->f_root->d_vnode_ops->vo_delete(pir, de);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}


int FileAttributes (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	char *p = NULL;
	struct dentry *de = NULL;
	int res = 0;

	LOGDEBUG("FileAttributes", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	p = to_BCS(pir->ir_upath+3, ucodelen(pir->ir_upath)-3);
	res = namei(fs, p, &de, CASE_INSENSITIVE);
	_HeapFree(p,0);

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_fileattrib) 
		res = fs->f_root->d_vnode_ops->vo_fileattrib(pir, de);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}


int FlushVolume (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("FlushVolume", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_flush) 
		res = fs->f_root->d_vnode_ops->vo_flush(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}


int Dir (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	struct dentry *de = NULL;
	char *p = NULL;
	int res = 0;

	LOGDEBUG("Dir", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (pir->ir_flags == CHECK_DIR) {	
		if (*(char *)pir->ir_upath == '\\') 
			p = to_BCS(pir->ir_upath+1, ucodelen(pir->ir_upath)-1);
		else 
			p = to_BCS(pir->ir_upath, ucodelen(pir->ir_upath));
	} else 
		p = to_BCS(pir->ir_upath+3, ucodelen(pir->ir_upath)-3);

	res = namei(fs, p, &de, CASE_INSENSITIVE);
	_HeapFree(p,0);

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_dir) 
		res = fs->f_root->d_vnode_ops->vo_dir(pir, de);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int GetDiskInfo (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("GetDiskInfo", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_getdiskinfo) 
		res = fs->f_root->d_vnode_ops->vo_getdiskinfo(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int QueryResourceInfo (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("QueryResourceInfo", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_query) 
		res = fs->f_root->d_vnode_ops->vo_query(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int RenameFile (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	struct dentry *de1 = NULL, *de2 = NULL;
	int res1 = 0, res2 = 0, len = 0;
	char srcpath[MAX_PATH];
	char *targetpath=NULL;

	LOGDEBUG("RenameFile", 1);

	res1 = CHECKFS(fs);

	if (res1)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	len = UniToBCSPath(srcpath, pir->ir_ppath->pp_elements, sizeof(srcpath)-1, BCS_OEM);
	srcpath[len]='\0';
	res1 = namei(fs, srcpath+1, &de1, CASE_INSENSITIVE);

	targetpath = to_BCS(pir->ir_upath+3, ucodelen(pir->ir_upath)-3);
	res2 = namei(fs, targetpath, &de2, CASE_INSENSITIVE);
	_HeapFree(targetpath,0);

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_rename) 
		res1 = fs->f_root->d_vnode_ops->vo_rename(pir, de1, de2);
	else res1 = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res1;
}

int UNCNamedPipeRequest (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("UNCNamedPipeRequest", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_uncpipereq) 
		res = fs->f_root->d_vnode_ops->vo_uncpipereq(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int Ioctl16Drive (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("Ioctl16Drive", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_ioctl16drive) 
		res = fs->f_root->d_vnode_ops->vo_ioctl16drive(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int GetDiskParams (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("GetDiskParams", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_getdiskparams) 
		res = fs->f_root->d_vnode_ops->vo_getdiskparams(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}

int DirectDiskIO (pioreq pir)
{
	pFilesystem fs = (pFilesystem) pir->ir_rh;
	int res = 0;
	LOGDEBUG("DirectDiskIO", 1);

	res = CHECKFS(fs);

	if (res)
		return pir->ir_error = ERROR_INVALID_PARAMETER;

	if (fs->f_root->d_vnode_ops && fs->f_root->d_vnode_ops->vo_dasdio) 
		res = fs->f_root->d_vnode_ops->vo_dasdio(pir);
	else res = ERROR_INVALID_FUNCTION;

	return pir->ir_error = res;
}
