//
// OSR Sample Network Provider DLL
//
// Developed by OSR Open Systems Resources, Inc.
//
// (C) Copyright 1995, 1996, 1997 OSR Open Systems Resources, Inc.
// All Rights Reserved.
// 
// This is an ordinary user-mode DLL, with defined entry points for MPR to call.
//
// *****************************************************************************
// Please note that this example is STILL PRELIMINARY.  Its development as
// an example is still not complete.  In no way should this development example
// be mistaken for end-user quality software -- It is an EXAMPLE ONLY
// *****************************************************************************
//
#include <windows.h>
#include <stdio.h>
#include <winsock.h>
#include <winreg.h>
#include <winioctl.h>
#include <malloc.h>
#include <string.h>

#include "osrnp.h"                // OSR and standard NP definitions

#ifdef USE_OUTPUTDEBUGSTRING
#define TRACE OutputDebugString
#define TRACE1(x,y) {char xxmsg[200]; sprintf(xxmsg, x, y); OutputDebugString(xxmsg);}
#else

FILE *debugfp = NULL;

void TRACE(LPSTR sz)
{
	if (!debugfp)
		debugfp = fopen("c:\\mprdll.log", "w+");

	if (debugfp) {
		fprintf(debugfp, "%s\n", sz);
		fflush(debugfp);
	}
}

#define TRACE1(x,y) {char xxmsg[200]; sprintf(xxmsg, x, y); TRACE(xxmsg);}

#endif

//
// DLL-Wide "static" data
// Recall, friends, that this is a DLL -- So the static data isn't REALLY
// Static, right?  When the DLL gets unloaded and then re-loaded, this
// data is all re-initialized.  No harm done in our example, but for
// a REAL PRODUCT, you probably want a nicer way to save your state
// across calls.
// 
//
static HANDLE FSDFileHandle = 0;
static HostName[OSRNFS_HOST_NAME_SIZE] = {0};
static char ServerNames[OSRNFS_SERVERS_SUPPORTED*OSRNFS_HOST_NAME_SIZE];
static DWORD ServerCount = 0;
static PUCHAR  OSRNPName = "OSR NFS Demo";
static ULONG NetworkInitialized = 0x00;

//
// Table of servers, derived from registry
// Note that the format of this structure is shared with the FSD
//
static OSRNFS_SERVER ServerTable[OSRNFS_SERVERS_SUPPORTED];

//
// Forward delclarations
//
int InitNetworkInfo();
PUCHAR CharToUnicode(PUCHAR a, PUCHAR b);
int GetSharesFromServer(PWCHAR Server, PWCHAR ShareList, ULONG ShareListSize, PULONG BytesWritten);
int GetConnections(PVOID ConnectionList, ULONG BufferSize, PULONG BytesWritten);
int ConnectDrive(PWCHAR Local, PWCHAR Remote);

//
// Windows NT Network Providers
//
// A Windows NT Network Provider is a DLL that interfaces with the Multi-Provider Router (MPR) DLL,
// that is in turn called, for example, by File Manager.  MPR is implemented to support multiple
// individual networks on NT.  Examples of these multiple networks might be the Microsoft Windows
// Network (SRV/RDR), Netware (CSNW), or some other networking scheme.
//
// This example implements V5.1 of the Network Provider (NP) specification.  Versions named V3.x
// were Windows 3.x specific and supported only one NP. In V4.0 of the spec, calls that are implemented
// by NPs start with NPxxx;  Calls implemented by Win32 (and are called by user applications) are
// prefixed by WNetXxxx.  The MPR is the only entity that calls NPxxx functions.  V5.x was created
// specifically to support Explorer.
//
// Starting with the V5.1 NP specification, Microsoft has started to make NP specifications
// publically avaiable as part of the Windows NT DDK.  Look in \DDK\DOCS\SPEC\NETWORK for the
// presently available documentation. A HINT: Read both the MPR spec and the NTNP spec togeter..
// One spec tends to "fill in the gaps" that exist in the other.
//
// Since developing a network provider isn't quite as well documented as it could be, we try
// to provide some guidance with this sample code.  Note that this code is an EXAMPLE, and
// should not be mistaken for a shipping end-user quality product.
//
// Our experience in developing network providers has shown that Explorer pretty sensitive to
// which functions are implemented by the NP, how they're implemented, and what's returned by
// them.  Supporting SOME functions in your NP and not others will tend to cause Explorer to
// engage in VERY different behaviors.
//
//
// Finding an NP and How It Gets Loaded In NT V4.x
//
// An NP is made known to the MPR by making an entry in the Registry under the key:
//
//                \HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\NetworkProvider\Order
//
// Under this key, the value "ProviderOrder" is a REG_SZ which comprises a comma separated list
// of NP service names.  Hence:
//
//                ProviderOrder:REG_SZ:NWCWorkstation,LanmanWorkstation
//
// The example above has two NPs configured: CSNW and LANMAN. The indicated values in the list
// (e.g. "LanmanWorkstation") are the names of services that can be found under the typical
// \HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services key.  Thus, the "LanmanWorkstation"
// service will be found in
//
//                \HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\LanmanWorkstation
//
// MPR finds the actual NP DLL image by looking at the service entry, under the "networkprovider"
// key.  Under this key, there are two required REG_SZ/REG_EXPAND_SZ values: (1) "ProviderPath"
// which contains the path and name of the actual NP dll image, and (2) "Name" which contains
// the name of the NP.  Thus, continuing with LANMAN, if you look under the key
//
//                ...\Services\LanmanWorkstation
//
//        There will be a key named
//
//                \networkprovider
//
//        under which there will be the values
//        
//                Name:REG_SZ:Microsoft Windows Network
//                ProviderPath:REG_EXPAND_SZ:%SystemRoot%\System32\ntlanman.dll
//
// 
// We have attempted to provide specific comments about the usage and capabilities of each of
// the NP functions where they appear, below.
//

//
// NPGetCaps
//
// This is the only function REQUIRED for a network provider.
// It accepts an index specifying the capability being queried, and
// returns a bitmask indicating the services in the set being queried
// that are supported.  A value of zero indicates no service in the
// indicated set are supported.
//
DWORD APIENTRY NPGetCaps(IN DWORD nIndex)
{
    int Status;

    TRACE("Enter: NPGetCaps\n");

    if(!NetworkInitialized)
        {

       TRACE("     Network not yet initialized.  Calling InitNetworkInfo\n");

       Status = InitNetworkInfo();

        if (!NetworkInitialized)
            {
            TRACE("     FAILED to initialize network\n");
            }
        }


	if ((nIndex != WNNC_SPEC_VERSION) &&
		(nIndex != WNNC_NET_TYPE) &&
		(nIndex != WNNC_DRIVER_VERSION) &&
		(nIndex != WNNC_USER) &&
		(nIndex != WNNC_START)) {
		TRACE1("GetCaps 0x%0x -- returning 0\n", nIndex);
		return 0;
	} else {
		TRACE1("GetCaps 0x%0x -- supported\n", nIndex);
	}

    switch (nIndex)
    {
    case WNNC_SPEC_VERSION: 
            {
            //
            // Supported version of the spec
            //
            TRACE1("GetCaps SPEC_VERSION 0x%0x\n",WNNC_SPEC_VERSION51);

            return(WNNC_SPEC_VERSION51);
            }

    case WNNC_NET_TYPE: 
            {
            //
            // Network Type
            //

            TRACE("GetCaps NET_TYPE\n");

            //
            // According to the documentation, you're supposed to get a network ID from
            // Microsoft.  While I'm sure that's true, we simply use an unused ID, higher
            // than the highest one currently assigned.  N.B.  THE ID MUST BE UNIQUE among
            // NPs on the system.
            //
            return(0x001b0000);

            }

    case WNNC_DRIVER_VERSION: 
            {
            //
            // NP Revision
            //
            TRACE("GetCaps DRIVER_VERSION\n");

            return(OSR_DRIVER_VERSION);
            }

    case WNNC_USER: 
            {
            //
            // User Info Capability Supported?
            // Returning WNNC_USR_GETUSER means that we implement NPGetUser;  Otherwise, return 0x00
            //

            TRACE("GetCaps USER\n");

            return(WNNC_USR_GETUSER);
            }

    case WNNC_CONNECTION: 
            {
            //
            // Connection Capabilities
            //
            // Return bits indicating which functions we support:
            //
            //        WNNC_CON_GETCONNECTIONS (gets connection info)
            //        WNNC_CON_ADDCONNECTION  (connects disks)
            //        WNNC_CON_CANCELCONNECTION
            //        WNNC_CON_GETCONNECTIONS
            //        WNNC_CON_ADDCONNECTION3 (connects devices, more general than ADDCONNECTION)
            //        wNNC_CON_GETPERFORMANCE
            //        WNNC_CON_DEFER
            //

            TRACE("GetCaps: CONNECTION\n");

            return( 
                    WNNC_CON_GETCONNECTIONS |
                    WNNC_CON_ADDCONNECTION  |
                    WNNC_CON_ADDCONNECTION3 |
                    WNNC_CON_GETPERFORMANCE |
                    WNNC_CON_CANCELCONNECTION );

            }

    case WNNC_DIALOG: 
            {

            //
            // Dialog Capabilities
            // Return bits set indicating which dialog capabilities are supported by this NP
            // 
            //      WNNC_DLG_DEVICEMODE              obsolete
            //      WNNC_DLG_PROPERTYDIALOG          NPPropertyDialog and NPGetPropertyText implemented
            //      WNNC_DLG_SEARCHDIALOG            NPSearchDialog implemented
            //      WNNC_DLG_FORMATNETWORKNAME       NPFormatNetworkName implemented
            //      WNNC_DLG_PERMISSIONEDITOR        NPFMXGetPermCaps, NPFMXEditPerm, NPFMXGetPermHelp implemented
            //      WNNC_DLG_GETRESOURCEPARENT       NPGetResourceParent implemented
            //      WNNC_DLG_GETRESOURCEINFORMATION  NPGetResourceInformation implemented
            //
            TRACE("GetCaps DIALOG\n");

            return(WNNC_DLG_FORMATNETWORKNAME |
//                   WNNC_DLG_PROPERTYDIALOG    |       // Only for File Manager
                   WNNC_DLG_GETRESOURCEPARENT |
                   WNNC_DLG_GETRESOURCEINFORMATION);

            }

    case WNNC_ADMIN: 
            {
            //
            // Administrative Capabilities
            // These capabilities allow an NP to manipulate and display particular network directories.
            //
            //      WNNC_ADM_GETDIRECTORYTYPE        NPGetDirectoryType implemented
            //      WNNC_ADM_DIRECTORYNOTIFY         NPDirectoryNotify implemented
            //
            TRACE("GetCaps ADMIN\n");

            return(WNNC_ADM_GETDIRECTORYTYPE | WNNC_ADM_DIRECTORYNOTIFY);
            }

    case WNNC_ENUMERATION: 
            {
            //
            // Network Entity Enumeration Support
            //
            //      WNNC_ENUM_ENUM_GLOBAL         Enumeration of global entities are supported
            //      WNNC_ENUM_LOCAL               Enumeration of local entities is supported
            //      WNNC_ENUM_CONTEXT             Ennumeration of user's default context
            //
            // Supporting ANY enumeration function implies support for NPOpenEnum, NPCloseEnum, and
            // NPEnumResource.
            //
            TRACE("GetCaps ENUMERATION\n");

            return( WNNC_ENUM_GLOBAL | 
                    WNNC_ENUM_LOCAL);

            }

    case WNNC_START: 
            {
            //
            // Current Start State -- Delay Time
            //
            // Return milliseconds to startup.  Special values:
            //      0x00        I'm dead, leave me alone
            //      0x01        I'm ready now;  Don't wait for me.
            //      0xFFFFFFFF  I'll be ready later, I think.  Use some default
            //

            TRACE("GetCaps START\n");


            if(NetworkInitialized)
                {

                //
                // "I'm ready NOW"
                //
                return(0x01);

                }
            else
                {

                //
                // "Later, dude"
                //

                return((ULONG)0XFFFFFFFFF);
                }
            }

    default:
        //
        // We don't support any other options.
        //

        TRACE("GetCaps default\n");

        return(0x00);

    }

}

//
// NPGetUser
//
// This function is used to determine the default user name of the user
// logged onto the system (as indicated by a NULL lpName or empty string
// for lpName), or the username under which the user has a device/network
// connection (indicated by a device nam or network name in lpName).
//
// lpName        Name of local device or network name, for which info is
//                                to be returned.  If NULL (or the empty string), requests
//                                info on user currently logged in.  If a network name is
//                                passed, and the user has more than one name, NP can
//                                choose which to return.
//
//        lpUserName        Buffer to receive name
//
//        lpBufferSize        Size, in bytes, of buffer pointed to by lpUserName.
// 
// Returns:
//                WN_SUCCESS                if the call is successful
//                WN_NOT_CONNECTED          lpName isn't a connected network name or redirected device
//                WN_NO_NETWORK             if the network is down
//                WN_MORE_DATA              if the buffer is too small.  In this case, lpBufferSize
//                                                                has the required size returned in it.
//
//
DWORD APIENTRY NPGetUser(LPWSTR lpName, LPWSTR lpUserName, LPDWORD lpBufferSize)
    {
	LPWSTR szUser = L"smarc";
	DWORD cbUser = sizeof(L"smarc") + 2;

    TRACE("Entering: NPGetUser...\n");

	if (*lpBufferSize < cbUser) {
		*lpBufferSize = cbUser;
		return (WN_MORE_DATA);
	}

	wcscpy(lpUserName, szUser);
    return(WN_SUCCESS);
    }

//
// NPGetConnection
//
// This function is used to get info about a particular connection.
//
//        lpLocalName                Pointer to the name (local) for which connection info is to be returned
//
//        lpRemoteName        Buffer for returned info (remote name used to make the connection)
//
//        lpBufferSize        Size of buffer.  Usual rules apply, i.e. if buffer is not big enough
//                                        then the required size is returned.
//
//
DWORD APIENTRY NPGetConnection(LPWSTR lpLocalName, LPWSTR lpRemoteName, LPDWORD lpnBufferSize)
    {
    int Return;
    WCHAR Data[4000];
    POW_FSCTL_GETCONNECTION_INFO GetConnectInfo = (POW_FSCTL_GETCONNECTION_INFO) &Data[0];
    POW_FSCTL_CONNECT_INFO ConnectInfo = &GetConnectInfo->Connections[0];
    ULONG BytesWritten;
    ULONG SharesReturned;
    PWCHAR Buffer, NPName;
    ULONG ResourceSize, i;
    WCHAR DriveName[2], LocalDrive[2];

    DriveName[1] = 0;

    LocalDrive[0] = (WCHAR)lpLocalName[0];
    LocalDrive[1] = 0;

    TRACE("Entering: NPGetConnection...\n");
    TRACE1("        LocalName %ls\n", lpLocalName);
    TRACE1("        LocalDrive %ls\n", LocalDrive);

    if(!NetworkInitialized)
        {
        TRACE("NPGetConnection: Network not initialized!\n");
        return(WN_NO_NETWORK);
        }

    //
    // We use a very simple interface between the NP and our FSD.
    // Here, we ask the FSD to tell us all it's connections.  Then
    // we'll look to see if the passed-in local name is one of these,
    // and if it is, we'll pass it back the remote name
    //
    TRACE("         Ennumerating connected resources\n");

    Return = GetConnections(Data, sizeof(Data), &BytesWritten);

    if(Return < 0)
        {
        TRACE("Network error on GetConnections\n");
        SetLastError(WN_NO_NETWORK);
        return(WN_NO_NETWORK);
        }

    //
    // Extract the # of shares returned
    //
    SharesReturned = GetConnectInfo->NumberOfEntries;

    TRACE1("     There are %d. connected shares\n", SharesReturned);

    //
    // If we don't have any connections, then the specified local
    // name is certainly not one of our connections.
    //
    if(!SharesReturned)
        {
        return(WN_NOT_CONNECTED);
        }


    //
    // Make a pointer into the caller's buffer to where we'll be copying
    // the returned share names
    //
    Buffer = (PWCHAR)lpRemoteName;

    //
    // Walk through each of the shares returned
    //
    for(i=0; i < SharesReturned; i++)
        {
        
        DriveName[0] = ConnectInfo->ConnectionIdentifier;

        TRACE1("     Connected Drive = %ls\n", (PWSTR)DriveName);
    
        //
        // Compare the connection name (the redirected drive letter) with
        // the local name -- non-case sensitive
        //
        if(!wcsicmp((PWSTR)DriveName, (PWSTR)LocalDrive) )
            {
            break;
            }

        ConnectInfo = (POW_FSCTL_CONNECT_INFO)
						((PWSTR)&(ConnectInfo->ConnectionName[0])
						+ wcslen((PWSTR)&(ConnectInfo->ConnectionName[0]))
						+1);
        }


    if(i >= SharesReturned)
        {
        TRACE("     DID NOT FIND localname among our connected drives\n");
        return(WN_NOT_CONNECTED);
        }

    *lpRemoteName = L'\\';
    wcscpy((PWSTR)((PWCHAR)lpRemoteName+1), (PWSTR)&ConnectInfo->ConnectionName[0]);

    TRACE1("        Remote connection found = %ls\n", lpRemoteName);

    return(WN_SUCCESS);
    }

//
// NPAddConnection
//
// This function is used to redirect a local device name to a network
// resource.  This is the "ordinary" connect function, that is called when a
// user double-clicks  For connections
// that are "different" the NPAddConnection3 function is called.
//
//        lpNetResource->lpRemoteName                Network resource to which to connect
//        lpNetResource->lpLocalName                Name of local device to redirect
//        lpNetResource->dwType                        ResourceType_PRINT, ResourceType_DISK, or
//                                                                        even ResourceType_ANY
//
//        lpPassword                                Password to use for connect, or NULL for default.
//        lpUserName                                Username for connect, or NULL for default
//
DWORD APIENTRY NPAddConnection(LPNETRESOURCEW lpNetResource, LPWSTR lpPassword, LPWSTR lpUserName)
    {
    int Status = 0;

    TRACE("Entering: NPAddConnection...\n");
    TRACE1("        Local = %ls\n", lpNetResource->lpLocalName);
    TRACE1("        Remote = %ls\n", lpNetResource->lpRemoteName);
    TRACE1("        User = %ls\n", lpUserName);
    TRACE1("        Password = %ls\n", lpPassword);

    //
    // *** Make the actual connection here ***
    //
    // Of course, other NPs will provide whatever info they need on the connect.  Since
    // our simple demonstration FSD doesn't pay any attention to user names or passwords,
    // we just pass the local and remote names for the connection.
    //
    if(lpNetResource->lpLocalName != NULL)
        {
        Status = ConnectDrive(lpNetResource->lpLocalName, lpNetResource->lpRemoteName);
        }

    if(Status < 0)
        {
        TRACE1("Connect failed!  Status = 0x%0x\n", Status);
        return(WN_NO_NETWORK);
        }

    return(WN_SUCCESS);
    }


//
// NPAddConnection3
//
// This function is used to redirect a local device name to a network
// resource, where an interactive, prompted, or temporary connection are
// to be made.
//
//        hwndOwner                                If CONNECT_INTERACTIVE is set in dwFlags,
//                                                                this is the window owner's handle
//
//        lpNetResource->lpRemoteName                Network resource to which to connect
//        lpNetResource->lpLocalName                Name of local device to redirect
//        lpNetResource->dwType                        ResourceType_PRINT, ResourceType_DISK, or
//                                                                        even ResourceType_ANY
//
//        lpPassword                                Password to use for connect, or NULL for default.
//        lpUserName                                Username for connect, or NULL for default
//        dwFlags                                        CONNECT_INTERACTIVE indicates a dialog with the
//                                                                user is possible for authentication purposes;
//                                                        CONNECT_PROMPT indicates that username and password
//                                                                should be prompted for (do not use default)
//                                                        CONNECT_TEMPORARY indicates a temporary connection
//                                                                is being established for browsing.
//
DWORD APIENTRY NPAddConnection3(HWND hwndOwner, LPNETRESOURCEW lpNetResource,
                                    LPWSTR lpPassword, LPWSTR lpUserName, DWORD dwFlags)
    {
    TRACE("Entering: NPAddConnection3...\n");

    TRACE1("        Local = %ls\n", lpNetResource->lpLocalName);
    TRACE1("        Remote = %ls\n", lpNetResource->lpRemoteName);

    if(dwFlags & CONNECT_INTERACTIVE)
        TRACE("    dwFlags = CONNECT_INTERACTIVE\n");

    if(dwFlags & CONNECT_PROMPT)
        TRACE("    dwFlags = CONNECT_PROMPT\n");

    if(dwFlags & CONNECT_TEMPORARY)
        TRACE("    dwFlags = CONNECT_TEMPORARY\n");

    if(dwFlags & CONNECT_DEFERRED)
        TRACE("    dwFlags = CONNECT_DEFERRED\n");


    //
    // We just turn-around and call NpAddConnection...
    //
    return(NPAddConnection(lpNetResource, lpPassword, lpUserName));
    }

//
// NPCancelConnection
//
// Disconnects a network connection.  If a DEVICE is disconnected, this
// reverses the AddConnection activity (undoing any redirection).  If a remote
// network resource is named, this is a "deviceless disconnection"
//
// lpName        The name of the redirected local device, or the remote network resource
//
// fForce        If TRUE, forces a disconnect, EVEN IF files or jobs are in progress
//                                on the remote device.  If FALSE, the call should fail if there
//                                are files or jobs open.
//
DWORD APIENTRY NPCancelConnection(LPWSTR lpName, BOOL fForce)
    {
    OW_FSCTL_DISCONNECT_INFO DisconnectInfo;
    int Status;
    ULONG BytesWritten;

    TRACE("Entering: NPCancelConnection...\n");

    TRACE1("        Local = %ls\n", lpName);
    TRACE1("        Force = %d.\n", fForce);

    DisconnectInfo.ConnectionIdentifier = lpName[0];
    DisconnectInfo.Force = fForce;
    DisconnectInfo.ConnectionName[0] = '\0';

    //
    // Open the FSD so we can send it the request
    //
    FSDFileHandle = CreateFile(
        "\\\\.\\OSRNFS",                     // Device name = OSRNFS
        GENERIC_READ | GENERIC_WRITE,        // Access
        0,                                   // share mode
        NULL,                                // lpsa
        OPEN_EXISTING,                       // create
        0,
        NULL);

    //
    // Check for an error on create
    //
    if (FSDFileHandle == INVALID_HANDLE_VALUE)
        {
        TRACE("OSRNP: Open driver failed\n");
        return((DWORD) ~0);
        }

    Status = DeviceIoControl(FSDFileHandle,
                             (ULONG)OW_FSCTL_DISCONNECT,
                             &DisconnectInfo,
                             2,
                             NULL,
                             0,
                             &BytesWritten,
                             0);

    TRACE1("Back from IOCTL;  Status = 0x%0x\n", Status);

    CloseHandle(FSDFileHandle);


    if(!Status)
        {

        //
        // The usual reason for this to fail is that files are opened on the
        // volume.
        //
        
        TRACE1("DISCONNECT failed!\n", Status);

        return (fForce ? WN_NO_NETWORK : WN_OPEN_FILES);

        }

    return(WN_SUCCESS);

    }


//
// NPOpenEnum
//
// This function declares the intention of MPR to enumerate (list) network resources
// of one or more given types.  MPR calls this function to "open" the enumeration, and
// subsequently calls NpEnumResource to actually retrieve the information about the
// resources that it has opened here.  This function returns a handle to MPR that MPR
// will use in a later call to NPEnumResource.
//
// Our experience with this function is that calls to NPOpenEnum(...) are performed
// immediately prior to calls to NPEnumResource(...) using that handle.  Subsequently
// NPCloseEnum(...) is called.  This makes things pretty easy, but there's no REASON
// per se, that it HAS to work this way.
//
//
//        dwScope               RESOURCE_CONNECTED                all connected resources
//                              RESOURCE_GLOBALNET                all network resources
//
//        dwType                RESOURCETYPE_DISK                 all disk-type resources
//                              RESOURCETYPE_PRINT                all print-type resources
//                                        (Bitmask.  0 requests ALL types -- If NP doesn't
//                                        know disk from print resources, all can always be
//                                        returned)
//
//        dwUsage               RESOURCEUSAGE_CONNECTABLE         all connectable resources
//                              RESOURCEUSAGE_CONTAINER           all container resources
//                                        (Bitmask.  0 requests ALL.)
//
//        lpNetResource        The container to enumerate, or NULL (to indicate the network
//                                        root).  If dwScope == Resource_CONNECTED this
//                                        MUST be NULL.
//
//        lphEnum                Returned handle to use for later call to NpEnumResource, if
//                                        the call returns WN_SUCCESS.
//
static ULONG NextToEnum = 0xFFFFFFFF;

DWORD APIENTRY NPOpenEnum(DWORD dwScope, DWORD dwType, DWORD dwUsage, LPNETRESOURCEW lpNetResource, LPHANDLE lphEnum)
    {

    TRACE("Entering: NPOpenEnum...\n");
    if(dwScope == RESOURCE_CONNECTED)
        TRACE("        Scope = RESOURCE_CONNECTED\n");
    if(dwScope == RESOURCE_GLOBALNET)
        TRACE("        Scope = RESOURCE_GLOBALNET\n");
#ifdef RESOURCE_CONTEXT
    if(dwScope == RESOURCE_CONTEXT)
        TRACE("        Scope = RESOURCE_CONTEXT\n");
#endif // RESOURCE_CONTEXT
    if(dwType == 0x00)
        TRACE("        Type = ALL\n");
    if(dwType & RESOURCETYPE_DISK)
        TRACE("        Type = RESOURCETYPE_DISK\n");
    if(dwType & RESOURCETYPE_PRINT)
        TRACE("        Type = RESOURCETYPE_PRINT\n");
    if(dwUsage == 0x00)
        TRACE("        Usage = ALL\n");
    if(dwUsage & RESOURCEUSAGE_CONNECTABLE)
        TRACE("        Usage = RESOURCEUSAGE_CONNECTABLE\n");
    if(dwUsage & RESOURCEUSAGE_CONTAINER)
        TRACE("        Usage = RESOURCEUSAGE_CONTAINER\n");
    TRACE1("        Resource Ptr = 0x%0x\n", lpNetResource);

    if(lpNetResource)
        {
        if(lpNetResource->lpRemoteName)
            TRACE1("    lpNetResource->lpRemoteName = %ls\n",lpNetResource->lpRemoteName); 

        if(lpNetResource->lpLocalName)
            TRACE1("    lpNetResource->lpLocalName = %ls\n",lpNetResource->lpLocalName); 

        }
    //
    // If the network hasn't been initialized yet, we're not
    // enumerating ANYthing.
    //
    if(!NetworkInitialized)
        {
        TRACE("     (network not initialized... no enums!\n");
        SetLastError(WN_NO_NETWORK); 
        return(WN_NO_NETWORK);
        }

    //
    // Are we being asked to ennumerate CONNECTED resources?
    //
    if (dwScope & RESOURCE_CONNECTED)
        {
        *lphEnum = (HANDLE)OSRNP_ENUM_CONNECTED;
        return(WN_SUCCESS);
        }

    //
    // We have no print resources
    //
    if(dwType & RESOURCETYPE_PRINT)
        {
        *lphEnum = (HANDLE)OSRNP_ENUM_NONE;
        return(WN_SUCCESS);
        }        

    //
    // Looking for CONNECTABLE resources?
    //
    if( (lpNetResource && ((dwUsage & RESOURCEUSAGE_CONNECTABLE))) || (!dwUsage && lpNetResource) )
        {
        WCHAR ServerName[OSRNFS_HOST_NAME_SIZE];
        ULONG i;

        //
        // Look through the server table by name for the resource
        // being requested.
        //
        for(i=0; i < ServerCount; i++)
            {
            (void)CharToUnicode(&ServerTable[i].Name[0], (PUCHAR)ServerName);

            if( !wcsicmp(ServerName, (PWCHAR)lpNetResource->lpRemoteName))
                break;
            }

        if(i >= ServerCount)
            {
            return(WN_NOT_CONTAINER);
            }

        //
        // The server index becomes the enum handle
        //
        *lphEnum = (HANDLE)i;

        return(WN_SUCCESS);

        }        

    //
    // Looking to enumerate servers?
    //
    if((!dwUsage && !lpNetResource && (dwScope == RESOURCE_GLOBALNET) ) || (dwUsage & RESOURCEUSAGE_CONTAINER) )
        {
        *lphEnum = (HANDLE)OSRNP_ENUM_SERVERS;

        //
        // Start enumerating from server index zero
        //
        NextToEnum = 0;

        return(WN_SUCCESS);
        }

    //
    // Looking for CONNECTABLE resources?
    //
    if( (dwUsage & RESOURCEUSAGE_CONNECTABLE) || !dwUsage)
        {
        WCHAR ServerName[OSRNFS_HOST_NAME_SIZE];
        ULONG i;

        //
        // There are no connectable resources that are also global
        //
        if(!lpNetResource)
            {
            *lphEnum = (HANDLE)OSRNP_ENUM_NONE;
            return(WN_SUCCESS);
            }

        //
        // Look through the server table by name for the resource
        // being requested.
        //
        for(i=0; i < ServerCount; i++)
            {
            (void)CharToUnicode(&ServerTable[i].Name[0], (PUCHAR)ServerName);

            if( !wcsicmp(ServerName, (PWCHAR)lpNetResource->lpRemoteName))
                break;
            }

        if(i >= ServerCount)
            {
            return(WN_NOT_CONTAINER);
            }

        //
        // The server index becomes the enum handle
        //
        *lphEnum = (HANDLE)i;

        return(WN_SUCCESS);

        }        

    *lphEnum = (HANDLE)OSRNP_ENUM_NONE;

    return(WN_SUCCESS);
    }


//
// NPCloseEnum
//
// This function closes a previously opened enumeration.   Not much to say here.
//
//
DWORD APIENTRY NPCloseEnum(HANDLE hEnum)
    {
    NextToEnum = 0xFFFFFFFF;
    TRACE("Entering: NPCloseEnum...\n");
    return(WN_SUCCESS);
    }
    
//
// NPEnumResource
//
// This function is used to enumerate a given network resource (or group of network resources)
// given a handle previously obtained from NpOpenEnum.  When called, the NP fills the lpBuffer
// with as many entries that will fit, that meet the criteria from the NPOpenEnum call, up to the
// number of entries requested by lpcCount. *lpcCount == -1 means "as many as will fit".  The data
// returned is returned in NETRESOURCE structures, as follows:
//
// NETRESOURCE .dwScope        RESOURCE_GLOBALNET or RESOURCE_CONNECTED
//             .dwType                RESOURCETYPE_DISK or RESOURCETYPE_PRINT or 0x00 (for neither)
//
//             .dwDisplayType One of the following:
//                                      RESOURCEDISPLAYTYPE_GENERIC                Provider doesn't care, or doesn't know
//                                      RESOURCEDISPLAYTYPE_DOMAIN                Group of servers
//                                      RESOURCEDISPLAYTYPE_SERVER                A server
//                                      RESOURCEDISPLAYTYPE_SHARE                A single share point
//
//             .dwUsage        RESOURCEUSAGE_CONNECTABLE or RESOURCEUSAGE_CONTAINER, if .dwScope==_GLOBALNET
//
//             .lpLocalName if .dwScope == RESOURCE_CONNECTED then this has the name of a redirected device.  It
//                                              could also be zero, if the connection does not correspond to a local device.
//
//             .lpRemoteName The remote network name.  If dwUsage==_CONNECTABLE then whatever is
//                                              returned here gets passed to NPAddConnection(...) if a connection is
//                                              to be made.  If .dwScope==_CONNECTED this is the network name to which the
//                                              local device .lpLocalName is connected.
//
//             .lpComment        The comment associated with the network resource
//
//             .lpProvider        The network provider name
//
// If WNNC_DLG_FORMATNETWORKNAME was specified in the NPGetCaps(...) call, MPR will call the
// NPFormatNetworkName(...) function with any .lpRemoteName data returned from this function, so that
// this data can be appropriately formatted for display.
//
// Note that the text strings to be returned are, in NT, all WCHAR strings.  These strings are supposed
// to reside in the returned buffer, so we store them AFTER the returned NETRESOURCEW blocks.
//
// On return, the value pointer to by lpcCount is updated to reflect the number of items actually
// returned.
//
// This is kind of a difficult function, since (a) the number of entries returned is be constrained
// by the smaller of the buffer size and the count, AND (b) we don't have the opportunity to pass
// back any context value that is subsequently returned to us, so we know where we left off.  So, we're
// forced to store some sort of context between calls, which is pretty ugly.  Of course, if you assume
// that there will always only be one call to OpenEnum, followed by one or more calls to EnumResource,
// followed by a call to CloseEnum, this isn't a problem.  While it seems that File Man / MPR work this
// way, there's apparently nothing that REQUIRES this.  Sigh!
//
DWORD APIENTRY NPEnumResource(HANDLE hEnum, LPDWORD lpcCount, LPVOID lpBuffer, LPDWORD lpBufferSize)
    {
    ULONG EnumRequested = (ULONG)hEnum;
    NETRESOURCEW * Resource;

    TRACE("Entering: NPEnumResource...\n");
    TRACE1("        hEnum = 0x%0x\n", hEnum);
    TRACE1("        Count = %d.\n", *lpcCount);
    TRACE1("        Size = %d.\n", *lpBufferSize);
    TRACE1("        NextToEnum = %d.\n", NextToEnum);

    EnumRequested = (ULONG)hEnum;


    //
    // Case based on the handle, which indicates the enumeration
    // that was requested
    //
    switch(EnumRequested)
    {

    //
    // The dummy handle for "nothing to enum"
    //
    case OSRNP_ENUM_NONE:
        *lpcCount = 0;
        SetLastError(WN_NO_MORE_ENTRIES);
        return(WN_NO_MORE_ENTRIES);

    //
    // Enumerate connections
    //        
    case OSRNP_ENUM_CONNECTED:
            {
            int Return;
            WCHAR Data[4000];
            POW_FSCTL_GETCONNECTION_INFO GetConnectInfo = (POW_FSCTL_GETCONNECTION_INFO) &Data[0];
            POW_FSCTL_CONNECT_INFO ConnectInfo = &GetConnectInfo->Connections[0];
            ULONG BytesWritten;
            ULONG SharesReturned;
            PWCHAR Buffer, NPName;
            ULONG ResourceSize, i;

            //
            // We return all the connected resources that we have in one go.  SO, if we've
            // been here, tell 'em no more left.
            //
            if(NextToEnum == EnumRequested)
                {
                *lpcCount = 0;
                SetLastError(WN_NO_MORE_ENTRIES);
                return(WN_NO_MORE_ENTRIES);
                }


            TRACE("Ennumerating connected resources\n");

            Return = GetConnections(Data, sizeof(Data), &BytesWritten);

            if(Return < 0)
                {
                TRACE("Network error on GetConnections\n");
                SetLastError(WN_NO_NETWORK);
                return(WN_NO_NETWORK);
                }

            //
            // Extract the # of shares returned
            //

            SharesReturned = GetConnectInfo->NumberOfEntries;

            TRACE1("  There are %d. connected shares to enum\n", SharesReturned);

            if(!SharesReturned)
                {
                NextToEnum = EnumRequested;
                *lpcCount = 0;
                SetLastError(WN_NO_MORE_ENTRIES);
                return(WN_NO_MORE_ENTRIES);
                }

            //
            // How much room will the share resource blocks take?
            //
            ResourceSize = SharesReturned * sizeof(NETRESOURCEW);

            //
            // Make a pointer into the caller's buffer to where we'll be copying
            // the returned share names
            //
            Buffer = (PWCHAR)((PUCHAR)lpBuffer + ResourceSize);

            //
            // Stick the NP name into the returned buffer first
            //
            NPName = Buffer;

            Buffer = (PWCHAR)CharToUnicode(OSRNPName, (PUCHAR)Buffer);

            //
            // If all our shares won't fit, tell 'em how much space we need
            //
            if( (*lpcCount != -1 && *lpcCount < SharesReturned) ||
                (ResourceSize+BytesWritten+(SharesReturned*sizeof(WCHAR)) > *lpBufferSize) )
                {
                TRACE1("TOO SMALL!  Need %d. bytes", ResourceSize+BytesWritten+
                    (SharesReturned*sizeof(WCHAR)));
                *lpBufferSize = ResourceSize+BytesWritten;
                SetLastError(WN_MORE_DATA);
                return(WN_MORE_DATA);
                }

            Resource = (NETRESOURCEW *)lpBuffer;

            for(i=0; i < SharesReturned; i++, Resource++)
                {
                Resource->dwScope = RESOURCE_CONNECTED;
                Resource->dwType = RESOURCETYPE_DISK;
                Resource->dwDisplayType = RESOURCEDISPLAYTYPE_SHARE;

                //
                // Point to the local name, store it in the data area and null terminate it.
                //
                Resource->lpLocalName = Buffer;
                *Buffer++ = ConnectInfo->ConnectionIdentifier;
                *Buffer++ = L':';
                *Buffer++ = 0x00;

                TRACE1("                LocalName = %ls                ", Resource->lpLocalName);

                //
                // Likewise, process the name of the connection name, which is in UNC format
                //
                Resource->lpRemoteName = Buffer;
                *Buffer++ = L'\\';          // PeterGV: Start with UNC name?
                wcscpy(Buffer, &ConnectInfo->ConnectionName[0]);
                Buffer += wcslen(Buffer);
                *Buffer++ = 0x00;
                ConnectInfo = (POW_FSCTL_CONNECT_INFO)
								((PWSTR)&(ConnectInfo->ConnectionName[0])
								+ wcslen((PWSTR)&(ConnectInfo->ConnectionName[0]))
								+1);


                TRACE1("        RemoteName = %ls        \n", Resource->lpRemoteName);
                Resource->lpComment = NULL;
                Resource->lpProvider = NPName;
                }

            *lpcCount = SharesReturned;

            NextToEnum = EnumRequested;

            return(WN_SUCCESS);
            }

    //
    // Enumerate servers
    //                
    case OSRNP_ENUM_SERVERS:
            {
            ULONG ResourceSize;
            PUCHAR Buffer, NPName;
            ULONG i, ServersToDo;
            ULONG StartingServer = NextToEnum;

            //
            // Are we done from a previous request?
            //
            if (StartingServer >= ServerCount)
                {
                *lpcCount = 0;
                SetLastError(WN_NO_MORE_ENTRIES);
                return(WN_NO_MORE_ENTRIES);
                }

            //
            // How many servers were we asked to enum?
            //
            if(*lpcCount == -1)
                {
                //
                // -1 means "as many as will fit", so plan to give him
                // all that remain.
                //
                ServersToDo = ServerCount-StartingServer;

                }
            else
                {
                //
                // Give him as many as he requested, up to the number we have remaining
                //
                ServersToDo = ( (ServerCount-StartingServer) > *lpcCount ? *lpcCount : (ServerCount-StartingServer) );
                }

            //
            // How much space will the resources take?
            //
            ResourceSize = (ServersToDo * sizeof(NETRESOURCEW));

            //
            // Is the info about the requested number of servers too bit to
            // fit into the supplied buffer?
            //
            if ( (ResourceSize + (ServersToDo*OSRNFS_HOST_NAME_SIZE) + strlen(OSRNPName))
                > *lpBufferSize )
                {

                TRACE("All servers WILL NOT FIT\n");

                //
                // ALL the servers we know about won't fit.  Therefore
                // only return ONE at a time.
                //
                ServersToDo = 1;
                ResourceSize = sizeof(NETRESOURCEW);

                //
                // Just to be sure... S'pose ONE entry WON'T even fit??
                //
                if( (ResourceSize + OSRNFS_HOST_NAME_SIZE + strlen(OSRNPName) )
                    > *lpBufferSize)
                    {
                    TRACE("Can't even do ONE!!!\n");
                    *lpBufferSize = ResourceSize + OSRNFS_HOST_NAME_SIZE + strlen(OSRNPName);
                    SetLastError(WN_MORE_DATA);
                    return(WN_MORE_DATA);
                    }

                }


            //
            // Remember where to start next time
            //
            NextToEnum = StartingServer+ServersToDo;

            TRACE1("Enumerating %d. servers\n", ServersToDo);
            TRACE1("Starting with server %d.\n", StartingServer);
            TRACE1("NextToEnum is now %d.\n", NextToEnum);

            //
            // Indicate how many we're enumerating
            //
            *lpcCount = ServersToDo;

            //
            // We put the RESOURCE structures right at the beginning
            // of the returned buffer
            //
            Resource = (NETRESOURCEW *)lpBuffer;

            //
            // The data area of the return buffer will be used to
            // store names -- such as the NP name and the names of
            // resources.  This data area comes AFTER all the resources.
            //
            Buffer = (PUCHAR)lpBuffer + ResourceSize;

            //
            // First in the data area of the buffer is the NP name.
            //
            NPName = Buffer;
            Buffer = CharToUnicode(OSRNPName, Buffer);        // Convert OSRNPName, put it in Buffer,
                                                              //    and update Buffer

            for(i=StartingServer; i < ServersToDo; i++, Resource++)
                {
                Resource->dwScope = RESOURCE_GLOBALNET;
                Resource->dwType = RESOURCETYPE_DISK;
                Resource->dwDisplayType = RESOURCEDISPLAYTYPE_SERVER;
                Resource->dwUsage = RESOURCEUSAGE_CONTAINER;
                Resource->lpLocalName = NULL;
                Resource->lpRemoteName = (PUSHORT)Buffer;
                Buffer = CharToUnicode((PUCHAR)&ServerTable[i].Name[0], Buffer);
                Resource->lpComment = NULL;
                Resource->lpProvider = (PUSHORT)NPName;
                }

            return(WN_SUCCESS);
            }

    //
    // It must be a request to enumerate the resources on a particular server
    //
    default:
            {
            int Return;
            WCHAR Data[1000];
            ULONG BytesWritten;
            ULONG SharesReturned;
            PWCHAR Buffer, ShareNames, NPName;
            ULONG ResourceSize, i;
            WCHAR ServerName[OSRNFS_HOST_NAME_SIZE];

            //
            // We return all the shares we have in one go.  SO, if we've
            // been here, tell 'em no more left.
            //
            if(NextToEnum == EnumRequested)
                {
                *lpcCount = 0;
                SetLastError(WN_NO_MORE_ENTRIES);
                return(WN_NO_MORE_ENTRIES);
                }


            TRACE1("Request for data on server %d.\n", EnumRequested);

            (void)CharToUnicode((PUCHAR)&ServerTable[EnumRequested].Name[0], (PUCHAR)ServerName);

            Return = GetSharesFromServer(ServerName, Data, 1000*sizeof(WCHAR), &BytesWritten);

            if(Return < 0)
                {
                TRACE("Network error on GetShares\n");
                SetLastError(WN_NO_NETWORK);
                return(WN_NO_NETWORK);
                }

            //
            // Retrieve the count of shares returned at the top of the buffer
            //
            SharesReturned = *(ULONG *)Data;

            TRACE1("  There are %d. shares to enum on this server\n", SharesReturned);

            if(!SharesReturned)
                {
                NextToEnum = EnumRequested;
                *lpcCount = 0;
                SetLastError(WN_NO_MORE_ENTRIES);
                return(WN_NO_MORE_ENTRIES);
                }

            //
            // Point to where the returned share names start
            ShareNames = Data+2;

            //
            // How much room will the share resource blocks take?
            //
            ResourceSize = SharesReturned * sizeof(NETRESOURCEW);

            //
            // Make a pointer into the caller's buffer to where we'll be copying
            // the returned share names
            //
            Buffer = (PWCHAR)((PUCHAR)lpBuffer + ResourceSize);

            //
            // Stick the NP name into the returned buffer first
            //
            NPName = Buffer;
            Buffer = (PWCHAR)CharToUnicode(OSRNPName, (PUCHAR)Buffer);

            //
            // If all our shares won't fit, tell 'em how much space we need
            //
            if( (*lpcCount != -1 && *lpcCount > SharesReturned) ||
                (ResourceSize+BytesWritten+(SharesReturned*OSRNFS_HOST_NAME_SIZE) > *lpBufferSize) )
                {
                *lpBufferSize = ResourceSize+BytesWritten;
                SetLastError(WN_MORE_DATA);
                return(WN_MORE_DATA);
                }

            Resource = (NETRESOURCEW *)lpBuffer;

            for(i=0; i < SharesReturned; i++, Resource++)
                {
                Resource->dwScope = RESOURCE_GLOBALNET;
                Resource->dwType = RESOURCETYPE_DISK;
                Resource->dwDisplayType = RESOURCEDISPLAYTYPE_SHARE;
                Resource->dwUsage = RESOURCEUSAGE_CONNECTABLE;
                Resource->lpLocalName = NULL;
                Resource->lpRemoteName = (PUSHORT)Buffer;

                //
                // We build our resource names to look like UNC names, but prefixed
                // with an underscore character. This is because when we're called at our
                // NPAddConnection function, all we're given is the NAME of the resource
                // (without any other clue as to the server on which it resides).  We "fix up"
                // the name for attractive display in our NPFormatNetworkName function (q.v.)
                //
                //                                *Buffer++=L'_';                                        // Start with an "_" flag to ourselves

                wcscpy(Buffer, L"\\\\");            // Add the "\\" UNC thing
                Buffer +=wcslen(Buffer);
                wcscpy(Buffer, ServerName);                // Add the server name
                Buffer +=wcslen(ServerName);
                wcscpy(Buffer, L"\\");                        // Add an NT path separator
                Buffer++;

                wcscpy(Buffer, ShareNames+1);                // Finally, add the share name,
                //    skipping the UNIX path separator
                Buffer +=wcslen(Buffer);
                *Buffer++ = 0x00;
                ShareNames +=wcslen(ShareNames)+1;

                Resource->lpComment = NULL;
                Resource->lpProvider = NPName;
                }

            *lpcCount = SharesReturned;

            NextToEnum = EnumRequested;

            return(WN_SUCCESS);
            }

    }


    return(WN_SUCCESS);
    }

//
// NPFormatNetworkName
//
// This function formats a network name for attractive and useful display.  It gives
// the NP a chance to stuff an ellipsis (...) into the appropriate place if the name
// gets too long to print. This function gets called for all network names,
// including the NP name.  WE take advantage of this function to remove the UNC-style
// path for particular remote network names, so that when ennumerating the resources on NFS
// server GRUMPY instead of seeing:
//
//                                \\GRUMPY\etc
//
// The user will instead see:
//
//                                /etc
//
// Which is much more attractive and is really the expected output, don't you think?  Of
// course we don't ALWAYS want the UNC name removed, such as for displays in the Disconnect
// Network Driver dialog box.  We indicate a name that needs "UNC pruning" by prefixing it
// with an underscore.  Is this a hack?  Well, sure.
//
//        lpRemoteName        Network name to be formatted
//
//        lpDisplayName        Buffer for the short name
//
//        lpnLength                Length of above buffer
//
//        dwFlags                        Formatting being requested:
//                                        WNFMT_MULTILINE                Give full name, put "\n" for line breaks
//                                        WNFMT_ABBREVIATED        Shorten the name for the space available
//                                        WNFMT_INENUM                Name is immediately preceeded by the container
//                                                                                        in which it exists.
//
//        dwAveCharPerLine        As it says...
//
DWORD APIENTRY  NPFormatNetworkName(LPWSTR lpRemoteName,
                                    LPWSTR lpDisplayName,
                                    LPDWORD lpnLength,
                                    DWORD  dwFlags,
                                    DWORD  dwAveCharPerLine )
    {
    DWORD BytesNeeded = dwAveCharPerLine * sizeof( WCHAR); 
    PWCHAR Ptr;

    TRACE("Entering: NPFormat...\n");
    TRACE1("        RemoteName = %ls\n", lpRemoteName);
    if(dwFlags & WNFMT_MULTILINE)
        TRACE("     Flags = WNFMT_MULTILINE\n");
    if(dwFlags & WNFMT_ABBREVIATED)
        TRACE("     Flags = WNFMT_ABBREVIATED\n");
    if(dwFlags & WNFMT_INENUM)
        TRACE("     Flags = WNFMT_INENUM\n");

//    if (*lpnLength < BytesNeeded) 
//        {
//        *lpnLength = BytesNeeded;
//        return(WN_MORE_DATA);
//        }

    //
    // If this is a name being used in an ENUM, and if
    // it's a UNC name, then make it into an NFS-style
    // name
    //
    if( (dwFlags & WNFMT_INENUM) &&
        (lpRemoteName[0] == L'\\') &&
        (lpRemoteName[1] == L'\\') )
        {

        Ptr = wcsstr((PWCHAR)&lpRemoteName[2], L"\\");

        //
        // If we find "\\" at the start of the string, we
        // skip to the first (nt style) path separator, and copy
        // starting from the character following it to the end of
        // the string.  We prefix the string with a UNIX style path
        // separator for the authentic UNIX look and fell. Yeah.
        //
        if(Ptr)
            {
            lpDisplayName[0] = L'/';
            wcscpy(&lpDisplayName[1], Ptr+1);
            }

        }
    else
        {
        //
        // We do NOT have a UNC-style name... just return it
        //
        wcscpy(lpDisplayName, lpRemoteName ); 
        }

    return(WN_SUCCESS);
    }


#if NT_40

//
// NPGetResourceInformation
//
// Called by Explorer to do dynamic "drill down".  An NP
// *MUST* support this function or Explorer won't work properly.
//
// Note that it is not enough for this function to just do
// some parsing.  It must actually CHECK to see if the indicated
// server and share can be reached via this network.
//
DWORD APIENTRY NPGetResourceInformation(LPNETRESOURCEW lpNetResource,
                                        LPVOID lpBuffer,
                                        LPDWORD lpcbBuffer,
                                        LPWSTR * lplpSystem)

    {
    NETRESOURCEW * Resource;
    WCHAR ServerName[OSRNFS_HOST_NAME_SIZE];
    WCHAR ResourceBeingSought[OSRNFS_HOST_NAME_SIZE];
    ULONG i;
    LPWSTR Ptr;
    PWCHAR CurrentChar, Buffer;
    ULONG  SharesReturned;
    int    Return;
    WCHAR Data[1000];
    ULONG BytesWritten;
    PWCHAR ShareNames;


    TRACE("Entering: NPGetResourceInfo...\n");
    TRACE1("        RemoteName = %ls\n", lpNetResource->lpRemoteName);

    if(!NetworkInitialized)
        {
        return(WN_NO_NETWORK);
        }

    CurrentChar = (PWCHAR)lpNetResource->lpRemoteName;

    //
    // Was a UNC name passed in???  Man, I HOPE so!
    //
    if( (*CurrentChar++ != L'\\') || (*CurrentChar++ != L'\\') )
        {
        //
        // We can ID it if we don't know what it is
        //

        TRACE("NPGetResourceInfo: Name is not UNC...\n");
        return(WN_BAD_NETNAME);
        }

    //
    // Find the name of the server.  It runs from
    // the third character, to the first NT path separator.
    //
    Ptr = wcsstr((LPWSTR)CurrentChar, L"\\");
    
    //
    // Create the name of the server from the UNC name
    //
    for(i=0; CurrentChar != (PWCHAR)Ptr; i++, CurrentChar++)
        {
        ResourceBeingSought[i] = *CurrentChar;
        }

    ResourceBeingSought[i] = 0;

    TRACE1("     Server being sought = %ls\n", ResourceBeingSought);

    //
    // Look through the server table by name for the server
    // being sought
    //
    for(i=0; i < ServerCount; i++)
        {
        // 
        // Convert the name from the ServerTable to UNICODE, and
        // store it in ServerName
        //
        (void)CharToUnicode(&ServerTable[i].Name[0], (PUCHAR)ServerName);

        if( !wcsicmp(ServerName, ResourceBeingSought) )
            break;
        }

    if(i >= ServerCount)
        {
        TRACE("NPGetResourceInfo: Name is not one of our servers...\n");
        return(WN_BAD_NETNAME);
        }

    //
    // Copy the name of the server EXACTLY AS PASSED to us
    // (case and all) as this is the REQUIREMENT for this function.
    // Carefully ensure it's zero terminated.
    //
    wcscpy(ServerName, ResourceBeingSought);
    ServerName[wcslen(ResourceBeingSought)] = 0;

    //
    // If we get here, the SERVER NAME RESOURCE being sought
    // in the passed-in UNC NAME is at least a server we know
    // about.
    //

    Return = GetSharesFromServer(ResourceBeingSought, Data, 1000*sizeof(WCHAR), &BytesWritten);

    if(Return < 0)
        {
        TRACE("Network error on GetShares\n");
        SetLastError(WN_NO_NETWORK);
        return(WN_NO_NETWORK);
        }

    //
    // Retrieve the count of shares returned at the top of the buffer
    //
    SharesReturned = *(ULONG *)Data;

    TRACE1("  There are %d. shares on this server\n", SharesReturned);

    if(!SharesReturned)
        {
        SetLastError(WN_NO_MORE_ENTRIES);
        return(WN_NO_MORE_ENTRIES);
        }

    //
    // Point to where the returned share names start
    ShareNames = Data+2;

    //
    // Now get the name of the resource being sought that's on
    // the server. This starts at the character AFTER the last
    // path separator to the character before the next path separator
    // or the end of the string, whichever comes first.
    //
    CurrentChar++;
    Ptr = wcsstr((PWSTR)CurrentChar, L"\\");
    
    if(!Ptr)
        {
        Ptr = CurrentChar + wcslen(CurrentChar);
        }

    //
    // Create the name of the server from the UNC name
    //
    for(i=0; CurrentChar != Ptr; i++, CurrentChar++)
        {
        ResourceBeingSought[i] = *CurrentChar;
        }

    ResourceBeingSought[i] = 0;

    TRACE1("    Searching for resource '%ls'\n", ResourceBeingSought);
    
    //
    // Now search the list of shares on the indicated
    // server.  Is the requested share on this server??
    //
    for(i=0; i < SharesReturned; i++)
        {

        TRACE1("    Server has a share named %ls\n", ShareNames+1);

        //
        // Do case INsensitive compare, skipping UNIX-style path thingy
        //
        if( !wcsicmp(ShareNames+1, ResourceBeingSought) )
            {
            break;
            }
                
        ShareNames += wcslen(ShareNames)+1;

        }
    
    //
    // If we didn't find it, just return
    //
    if (i>=SharesReturned)
        {

        TRACE(" Share name NOT FOUND\n");

        return(WN_BAD_NETNAME);
        }

    //
    // WE FOUND IT!
    //
    // Return the data in the prescribed format
    //
    memset(lpBuffer, 0x00, *lpcbBuffer);
    Resource = (LPNETRESOURCEW)lpBuffer;
    Buffer = (PWCHAR) ((PUCHAR)lpBuffer + sizeof(NETRESOURCEW));

    Resource->lpRemoteName = Buffer;

    wcscpy(Buffer, L"\\\\");            // Add the "\\" UNC thing
    Buffer +=wcslen(Buffer);
    wcscpy(Buffer, ServerName);         // Add the server name
    Buffer +=wcslen(ServerName);
    wcscpy(Buffer, L"\\");                        // Add an NT path separator
    Buffer++;
    wcscpy(Buffer, ResourceBeingSought);                // Finally, add the share name,
    Buffer +=wcslen(Buffer);
    *Buffer++ = 0x00;

    TRACE1("    Returning Found Name = %ls\n", Resource->lpRemoteName);

    //
    // Return the provider name
    Resource->lpProvider = Buffer;
    Buffer = (PWCHAR)CharToUnicode(OSRNPName, (PUCHAR)Buffer);
    TRACE1("    Returning Provider = %ls\n", Resource->lpProvider);

    //
    // Since we only support disk resources, we know what type
    // this thing is, a priori
    //
    Resource->dwType = RESOURCETYPE_DISK;
    Resource->dwUsage = RESOURCEUSAGE_CONNECTABLE;
    Resource->dwDisplayType = RESOURCEDISPLAYTYPE_SHARE;

    //
    // Finally, return the non-identified part of the passed-in name
    //
    *lplpSystem = Buffer;
    wcscpy(Buffer, CurrentChar);
    Buffer +=wcslen(Buffer);
    *Buffer++ = 0x00;
    TRACE1("    Returning System = %ls\n", *lplpSystem);

    return(WN_SUCCESS);
    }

//
// NPGetResourceParent
//
// This is another one of those functions that are supposedly
// required for Explorer to work properly, although we have not
// observed this to be the case.
//
// Unlike NPGetResourceInformation, this function CAN just do
// some parsing to devise the resource's parent.
//
// We must admit to some guessing about what this function is
// supposed to do, but explorer seems to work with this NP...
//
DWORD APIENTRY NPGetResourceParent(LPNETRESOURCEW lpNetResource,
                                    LPVOID lpBuffer,
                                    LPDWORD lpcbBuffer)

    {
    NETRESOURCEW * Resource;
    WCHAR ServerName[OSRNFS_HOST_NAME_SIZE];
    WCHAR ResourceBeingSought[OSRNFS_HOST_NAME_SIZE];
    ULONG i, ElementsFound;
    LPWSTR Ptr;
    PWCHAR CurrentChar, Buffer, StartOfCurrentElement;

    TRACE("Entering: NPGetResourceParent...\n");
    TRACE1("        RemoteName = %ls\n", lpNetResource->lpRemoteName);
    TRACE1("        Provider = %ls\n", lpNetResource->lpProvider);
    if(lpNetResource->dwType == 0x00)
        TRACE("        Type = ALL\n");
    if(lpNetResource->dwType & RESOURCETYPE_DISK)
        TRACE("        Type = RESOURCETYPE_DISK\n");
    if(lpNetResource->dwType & RESOURCETYPE_PRINT)
        TRACE("        Type = RESOURCETYPE_PRINT\n");
    if(lpNetResource->dwUsage == 0x00)
        TRACE("        Usage = ALL\n");
    if(lpNetResource->dwUsage & RESOURCEUSAGE_CONNECTABLE)
        TRACE("        Usage = RESOURCEUSAGE_CONNECTABLE\n");
    if(lpNetResource->dwUsage & RESOURCEUSAGE_CONTAINER)
        TRACE("        Usage = RESOURCEUSAGE_CONTAINER\n");

    CurrentChar = (PWCHAR)lpNetResource->lpRemoteName;

    //
    // We're going to be a bit paranoid here, and check to
    // at least see if we know about the named server.
    //

    //
    // Check for UNC name
    //
    if( (*CurrentChar++ != L'\\') || (*CurrentChar++ != L'\\') )
        {
        //
        // We can ID it if we don't know what it is
        //

        TRACE("NPGetResourceParent: Name is not UNC...\n");
        return(WN_BAD_NETNAME);
        }

    //
    // Find the name of the server.  It runs from
    // the third character, to the first NT path separator.
    //
    Ptr = wcsstr((LPWSTR)CurrentChar, L"\\");
    
    if(!Ptr)
        {
        //
        // Ah, we have a server, but no path separator.
        // We DO see the NP called here with just a server name
        // (i.e. \\server), but have no idea what to return.
        //

        TRACE("NPGetResourceParent: Server but no path...\n");
        return(WN_BAD_VALUE);
        }

    //
    // Create the name of the server from the UNC name
    //
    for(i=0; CurrentChar != (PWCHAR)Ptr; i++, CurrentChar++)
        {
        ResourceBeingSought[i] = *CurrentChar;
        }

    ResourceBeingSought[i] = 0;

    TRACE1("     Server being sought = %ls\n", ResourceBeingSought);

    //
    // Look through the server table by name for the server
    // being sought
    //
    for(i=0; i < ServerCount; i++)
        {
        // 
        // Convert the name from the ServerTable to UNICODE, and
        // store it in ServerName
        //
        (void)CharToUnicode(&ServerTable[i].Name[0], (PUCHAR)ServerName);

        if( !wcsicmp(ServerName, ResourceBeingSought) )
            break;
        }

    if(i >= ServerCount)
        {
        TRACE("NPGetResourceParent: Name is not one of our servers...\n");
        return(WN_BAD_NETNAME);
        }

    //
    // Start to build the RESOURCE structure to return
    //
    memset(lpBuffer, 0x00, *lpcbBuffer);
    Resource = (LPNETRESOURCEW)lpBuffer;
    Buffer = (PWCHAR) ((PUCHAR)lpBuffer + sizeof(NETRESOURCEW));

    //
    // Return the provider name
    Resource->lpProvider = Buffer;
    Buffer = (PWCHAR)CharToUnicode(OSRNPName, (PUCHAR)Buffer);
    TRACE1("    Returning Provider = %ls\n", Resource->lpProvider);
    Buffer +=wcslen(Buffer);
    Buffer++;                           // Null terminate

    //
    // Now progressively build the name of the parent in
    // the buffer
    //
    Resource->lpRemoteName = Buffer;

    ElementsFound = 1;

    wcscpy(Buffer, L"\\\\");              // Start with the "\\" UNC thing
    Buffer +=wcslen(Buffer);
    wcscpy(Buffer, ResourceBeingSought);  // Add the server name
    Buffer +=wcslen(ServerName);

    //
    // Is there ANOTHER path separator?
    //
    while(Ptr = wcsstr((LPWSTR)++CurrentChar, L"\\") )
        {

        wcscpy(Buffer, L"\\");            // Path separator
        Buffer++;

        //
        // Count another syntactic element found
        //
        ElementsFound++;

        //
        // Add the next element 
        //
        while(CurrentChar !=(PWCHAR)Ptr)
            {
            *Buffer++ = *CurrentChar++;
            }

        }


    TRACE1("    Returning Parent Name = %ls\n", Resource->lpRemoteName);
    TRACE1("    Elements Found = %d.\n", ElementsFound);

    //
    // Set the right type, usage, and display type based
    // on the number of elements found.
    //
    switch (ElementsFound)
        {

        case 1:
            //
            // The parent is the SERVER itself
            //
            Resource->dwType = RESOURCETYPE_ANY;
            Resource->dwUsage = RESOURCEUSAGE_CONTAINER;
            Resource->dwDisplayType = RESOURCEDISPLAYTYPE_SERVER;
            break;

        case 2:
            //
            // The parent is the Server and SHARE -- Since we don't
            // support printers via NFS, the type is always DISK.
            //
            Resource->dwType = RESOURCETYPE_DISK;
            Resource->dwUsage = RESOURCEUSAGE_CONNECTABLE | RESOURCEUSAGE_CONTAINER;
            Resource->dwDisplayType = RESOURCEDISPLAYTYPE_SHARE;
            break;

        default:
            //
            // The parent is a directory
            //
            Resource->dwType = RESOURCETYPE_ANY;
            Resource->dwUsage = 0;
            Resource->dwDisplayType = RESOURCEDISPLAYTYPE_DIRECTORY;
            break;

        }


    return(WN_SUCCESS);
    }


//
// NPGetConnectionPerformance
//
// This funciton is used to return performance information about the connection.
// For this example, we "cheat" and always return the same thing.
//
DWORD APIENTRY NPGetConnectionPerformance (LPCWSTR  lpRemoteName, LPNETCONNECTINFOSTRUCT lpNetConnectInfo)
    {
    
    lpNetConnectInfo->dwFlags = WNCON_FORNETCARD;           // Netcard speed returned
    lpNetConnectInfo->dwSpeed = (DWORD)(10000000/100);      // Speed = 10Mbps in 100bps units
    lpNetConnectInfo->dwDelay = 0;                          // Additional delay
    lpNetConnectInfo->dwOptDataSize = 1024;                 // Optimal data size

    return(WN_SUCCESS);
    }

#endif


//
// NPGetUniversalName
//
// This function is used to convert a "drive based resource" to a universal (UNC) name.
//
//        lpLocalPath                Drive-based path of an object on a network resource
//
//        lwInfoLevel                The info requested, either: UNIVERSAL_NAME_INFO_LEVEL (0x01) to indicate that
//                                                a UNC name should be returned in lpBuffer or REMOTE_NAME_INFO which indicates
//                                                that a REMOTE_NAME_INFO structure should be returned in lpBuffer.
// 
DWORD APIENTRY NPGetUniversalName(LPCWSTR lpLocalPath, DWORD dwInfoLevel, LPVOID lpBuffer, LPDWORD lpBufferSize)
    {
    TRACE("Entering: ****************NPGetUniversal...\n");
    TRACE1("                Info Level = %d.\n", dwInfoLevel);
    TRACE1("                LocalPath = %ls\n", lpLocalPath);

    return(WN_NO_NETWORK);
    }


//
// NPGetPropertyText
//
// This function is called to extend the File Manager's properties dialog box. Whenever
// File Manager pops its properties dialog box up, this function is called so that
// it can indicate any buttons that need to be added to the box.  If the user actually
// clicks one of the buttons, File Man calls NPPropertyDialog.
//
//        iButtonDlg                Button index number, 0-5 (or is it 1-6);  Index values with 0x10 set indicate
//                                                multiple items were selected.
//
//        nPropSel                Which entities the box is focused on (WNPS_FILE, WNPS_DIR, WNPS_MULT (which
//                                                means more than one of SOMEthing, could be files or directories).
//
//        lpFileName                Fully qualified name of entitiy;  Multiple entities separated with spaces.
//                                                Note that some file names may be quoted with quotes or up-arrow.
//
//        lpButtonName        Poitner to where to copy name of button.  If this is empty, then no more
//                                                calls are made to this function (until the next time the properties dialog
//                                                is popped up).
//
//        cbButtonName        Size of buffer
//
//        nType                        Entity type.  Always WNTYPE_FILE?
//
DWORD APIENTRY NPGetPropertyText(DWORD iButtonDlg, DWORD nPropSel, LPWSTR lpFileName, LPWSTR lpButtonName, DWORD cbButtonName, DWORD nType)
    {
    TRACE("Entering: NPGetPropertyText\n");

    TRACE1("        FileName = %ls\n", lpFileName);

    if(iButtonDlg == 0x00) 
        {
        wcscpy(lpButtonName, L"Dumb");
        return(WN_SUCCESS);
        } 
    else 
        {
        return(WN_SUCCESS);
        }
    }


//
// NPPropertyDialog
//
// Lets the NP actually take an action on one of the buttons it added above.
//
//        hwndParent        Window handle, so NP can have some user interaction
//        
//        iButtonDlg        Index of pressed button (same index as NPGetPropertyText)
//
//        nPropSel        (same as above)
//
//        lpFileName        (same as above)
//
//        nType                (same as above)
//
DWORD APIENTRY NPPropertyDialog(HWND hwndParent, DWORD iButtonDlg, DWORD nPropSel, LPWSTR lpFileName, DWORD nType)
    {
    TRACE("Entering: NPPropertyDialog\n");
    TRACE1("     Button Pressed = 0x%0x\n", iButtonDlg);

    return(WN_SUCCESS);
    }


//
// NPSearchDialog
//
// If an NP doesn't support enumeration, then this function is called whenever the
// provider's network entry is dobule clicked.  If enumeration is supported, it is
// done using the enumeration functions, and this function is called when the File
// Manager's connection dialog appears.  In the connection dialog, File Man will
// enable it's "Search" button.  If clicked, this function will be called.
//
// Supporting this function allows the NP to provide any sort of view of its
// network resources that makes sense, thus freeing it from the constraints imposed
// by File Man's tree structured view (domains, servers, shares, etc).
//
//        hParent                        The window handle for the NP to use
//
//        lpNetResource        The currently selected item in the Network Connections dialog.
//                                                This could of course be ignored by the NP if it wants.
//
//        lpBuffer                Buffer to receive the result of the search
//
//        cbBuffer                Length
//
//        lpnFlags                On exit, if NP sets this to WNSRCH_REFRESH_FIRST_LEVEL then MPR
//                                                will collapse and then expand the first level below the NP's
//                                                network on return.
//
DWORD APIENTRY NPSearchDialog(HWND hwndParent, LPNETRESOURCEW lpNetResource, LPVOID lpBuffer, DWORD cbBuffer, LPDWORD lpnFlags)
    {
    TRACE("Entering: NPSearchDialog\n");

    return(WN_SUCCESS);
    }


//
// Permissions Items
//
// These functions allow NPs to have the File Man eXtensions call then so that they
// can implement their own "permission editor" attributes and dialogs.
//
//
DWORD APIENTRY NPFMXGetPermCaps(LPWSTR lpDriveName)
    {
    return(WN_SUCCESS);
    }

DWORD APIENTRY NPFMXEditPerm(LPWSTR lpDriveName, HWND hwndFMX, DWORD nType)
    {
    return(WN_SUCCESS);
    }

DWORD APIENTRY NPFMXGetPermHelp(LPWSTR lpDriveName, DWORD nType, BOOL fDirectory, LPVOID lpBuffer, LPDWORD lpBufferSize, LPDWORD lpnHelpCntext)
    {
    return(WN_SUCCESS);
    }

//
// NPGetDirectoryType
//
// This function allows the NP to have File Man display a particular directory as a
// "special" directory.
//
//        lpName        The fully qualified path name of the directory
//
//        lpType        If NP sets this value to something other than zero on exit, File Man displays a
//                                "special" folder icon.
//
//        FlushCache        Set to TRUE on first refresh of a window (allows NP to optimize performance, just
//                                        reading data once and caching it)
//
DWORD APIENTRY NPGetDirectoryType(LPWSTR lpName, LPINT lpType, BOOL FlushCache)
    {

    return(WN_SUCCESS);
    }

//
// NPDirectoryNotify
//
// This function allows the NP to manage certain directory operations on its own.
//
//        hwnd        The window handle
//
//        lpDir        Fully qualified path name of directory
//
//        dwOper        Operation being performs: WNDIR_MKDIR, WNDIR_RMDIR, or WNDIR_MVDIR (make, remove, and move
//                                directory)
//
// On exit, the status is WN_SUCCESS if the operation is fine and FILE MAN should perform the
// indicated operation.  If WN_CANCELLED is returned, FILE MAN should not perform the operation
// since the user cancelled it.  If WN_CONTINUE is returned, this indicates that the NP has
// performed the operation (successfully), and FILE MAN should continue normally, but NOT perform
// the operation itself.
//
DWORD APIENTRY NPDirectoryNotify(HWND hwnd, LPWSTR lpDir, DWORD dwOper)
    {

    return(WN_SUCCESS);
    }


//**********************************************************************************************
// Utility Routines Below
//**********************************************************************************************


//
// ConnectDrive
//
// Performs the connection of a local name to a remote name
//
int ConnectDrive(PWCHAR LocalName, PWCHAR RemoteName)
    {
    ULONG Status, Size;
    UCHAR Data[200];
    POW_FSCTL_CONNECT_INFO ConnectInfo = (POW_FSCTL_CONNECT_INFO)&Data[0];
    ULONG BytesWritten;

    //
    // Copy the local drive letter if there is one.  Otherwise,
    // mark this is a non-drive-letter-associated connection.
    // In the case of a UNC-type connection, the connection will
    // already have been established by the queryname call.
    //
    if(LocalName)
        {
        ConnectInfo->ConnectionIdentifier = LocalName[0];
        }
    else
        {
        ConnectInfo->ConnectionIdentifier = 0x0;
        }


    //
    // Note that this is a disk connection
    // (We don't do printers in this example)
    //
    ConnectInfo->ConnectionType = OW_CONNECTION_DISK;

    //
    // And follow it immediately with the remote name
    //
    if(RemoteName[0] == L'_')
        {
        RemoteName++;
        }

    wcscpy((PWCHAR)&ConnectInfo->ConnectionName[0], RemoteName);

    Size = wcslen(RemoteName)+1;

    //
    // Open the FSD so we can send him the request
    //
    FSDFileHandle = CreateFile("\\\\.\\OSRNFS",              // Device name = OSRNFS
                        GENERIC_READ | GENERIC_WRITE,        // Access
                        0,                                   // share mode
                        NULL,                                // lpsa
                        OPEN_EXISTING,                       // create
                        0,        
                        NULL);

    //
    // Check for an error on create
    //
    if (FSDFileHandle == INVALID_HANDLE_VALUE)
        {
        TRACE("OSRNP: Open driver failed\n");
        return(-1);
        }

    Status = DeviceIoControl(FSDFileHandle,
                            (ULONG)OW_FSCTL_CONNECT,
                            Data,
                            Size,
                            NULL,
                            0,
                            &BytesWritten,
                            NULL);

    TRACE1("Back from IOCTL;  Status = 0x%0x\n", Status);

    CloseHandle(FSDFileHandle);

    if(!Status)
        return(-1);

    return(0);

    }

//
// GetConnections
//
// Lists all known connections
//
int GetConnections(PVOID ConnectionList, ULONG BufferSize, PULONG BytesWritten)
    {
    ULONG Status;

    //
    // Open the FSD so we can send him the request
    //
    FSDFileHandle = CreateFile("\\\\.\\OSRNFS",              // Device name = OSRNFS
                        GENERIC_READ | GENERIC_WRITE,        // Access
                        0,                                   // share mode
                        NULL,                                // lpsa
                        OPEN_EXISTING,                       // create
                        0,        
                        NULL);

    //
    // Check for an error on create
    //
    if (FSDFileHandle == INVALID_HANDLE_VALUE)
        {
        TRACE("OSRNP: Open driver failed\n");
        return(-1);
        }

    Status = DeviceIoControl(FSDFileHandle,
                            (ULONG)OW_FSCTL_GETCONNECTIONS,
                            NULL,
                            0,
                            ConnectionList,
                            BufferSize,
                            BytesWritten,
                            NULL);

    TRACE1("Back from IOCTL;  Status = 0x%0x\n", Status);

    CloseHandle(FSDFileHandle);
    return(0);
    }



//
// GetSharesFromServer
//
int GetSharesFromServer(PWCHAR Server, PWCHAR ShareList, ULONG ShareListSize, PULONG BytesWritten)
    {
    ULONG Status;

    //
    // Open the FSD so we can send him some initialization info
    //
    FSDFileHandle = CreateFile("\\\\.\\OSRNFS",           // Device name = OSRNFS
                            GENERIC_READ | GENERIC_WRITE, // Access
                            0,                            // share mode
                            NULL,                         // lpsa
                            OPEN_EXISTING,                // create
                            0,        
                            NULL);

    //
    // Check for an error on create
    //
    if (FSDFileHandle == INVALID_HANDLE_VALUE)
        {
        TRACE("OSRNP: Open driver failed\n");
        return(-1);
        }

    Status = DeviceIoControl(FSDFileHandle,
                            (ULONG)OW_FSCTL_ENUM,
                            Server,
                            wcslen(Server) * sizeof(WCHAR),
                            ShareList,
                            ShareListSize,
                            BytesWritten,
                            NULL);

    TRACE1("Back from IOCTL;  Status = 0x%0x\n", Status);

    CloseHandle(FSDFileHandle);

    if(!Status)
        return(-1);

    return(0);
    }

//
// InitNetworkInfo
//
// This function is called to establish an initial connection with the Network
// File System implementation.
//
int InitNetworkInfo()
    {
#if 0
    char Buffer[OSRNFS_HOST_NAME_SIZE];
    int * Addr;
    struct hostent * Host;
    int err; 
    WORD wVersionRequested; 
    WSADATA wsaData; 
    LONG Status;
    LONG Type;
    LONG DataSize = 100;
    HKEY MyKey;
    DWORD Size = OSRNFS_SERVERS_SUPPORTED*OSRNFS_HOST_NAME_SIZE;
    char * Ptr;
    DWORD i;
    POSRNFS_INIT InitializeBuffer = 0;
    ULONG InitializeSize = 0;
    ULONG BytesWritten = 0;

    //
    // Open the FSD so we can send him some initialization info
    //
    FSDFileHandle = CreateFile("\\\\.\\OSRNFS",          // Device name = OSRNFS
                        GENERIC_READ | GENERIC_WRITE,    // Access
                        0,                               // share mode
                        NULL,                            // lpsa
                        OPEN_EXISTING,                   // create
                        0,        
                        NULL);

    //
    // Check for an error on create
    //
    if (FSDFileHandle == INVALID_HANDLE_VALUE)
        {
        TRACE("OSRNP: Open driver failed\n");
        return(-1);
        }

    //
    // Initialize WinSock
    //
    wVersionRequested = MAKEWORD(1, 1); 
    err = WSAStartup(wVersionRequested, &wsaData); 

    if (err != 0) 
        {
        TRACE("WSAStartup failed\n");
        return(-1);
        }
    //
    // Get the name of the system on which we're running
    //
    RtlZeroMemory(Buffer, OSRNFS_HOST_NAME_SIZE);

    (void)gethostname((char *)&Buffer[0], OSRNFS_HOST_NAME_SIZE);

    TRACE1("Running on host %s\n", Buffer);

    //
    // Determine which NFS Servers we can talk to by querying the registry.
    // Most NP's would ask their network FSD via an IOCTL, but since
    // NFS servers aren't "browsed" this is a cheap way to do it...
    //
    Status = RegOpenKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\NFSDEMO\\Parameters", &MyKey);

    if(Status != ERROR_SUCCESS)
        {        
        TRACE1("RegOpenKey failed 0x%0x\n", Status);
        return(-1);
        }

    Status = RegQueryValueEx(MyKey, "Servers",NULL, &Type, (PCHAR)&ServerNames, &Size);

    if(Status != ERROR_SUCCESS)
        {
        TRACE1("RegQueryValue failed 0x%0x\n", Status);
        return(-1);
        }

    //
    // No need to keep THIS hanging aroung
    //
    (void)RegCloseKey(MyKey);

    i = 0;
    Ptr = ServerNames;

    //
    // Walk through the REG_MULTI_SZ returned, and build the ServerTable
    // for our later use.  As we move each string from the registry, upcase
    // it.
    //
    do        
        {
        strcpy((char *)&ServerTable[i].Name, Ptr);
        strupr((char *)&ServerTable[i].Name);
        Ptr += strlen(Ptr);
        i++;
        } 
    while(*(++Ptr));

    ServerCount = i;

    //
    // Allocate (more than enough space for) a temp buffer for the init IOCTL
    //
    InitializeSize = (sizeof(OSRNFS_SERVER) * ServerCount) + sizeof(OSRNFS_INIT);

    InitializeBuffer = (POSRNFS_INIT)malloc(InitializeSize);

    RtlZeroMemory((PUCHAR)InitializeBuffer, InitializeSize);

    //
    // Fill in our host name
    //
    strcpy((PUCHAR)&InitializeBuffer->HostName[0], (PUCHAR)&Buffer[0]);

    //
    // How many servers do we know about?
    //
    InitializeBuffer->KnownServers = ServerCount;

    for (i=0; i < ServerCount; i++)
        {

        //
        // Lookup each host's address, and store it in the server list
        //
        Host = gethostbyname(ServerTable[i].Name);

        if(!Host)
            {
            TRACE("gethostbyname failed.\n");

            //
            // Return the initialize Buffer
            //
            free((PUCHAR)InitializeBuffer);

            return(-1);

            }
        else
            {
            Addr = (int *) *Host->h_addr_list;
            ServerTable[i].Address = ntohl(*Addr);
            TRACE1("Server = %s  ", &ServerTable[i].Name);
            TRACE1("Address = 0x%0x\n", ServerTable[i].Address);
            }

        //
        // Fill the initialize buffer while building our own internal table
        //
        InitializeBuffer->Servers[i] = ServerTable[i];
        }


    Status = DeviceIoControl(FSDFileHandle,
                            (ULONG)OW_FSCTL_INIT,
                            InitializeBuffer,
                            InitializeSize,
                            NULL,
                            0,
                            &BytesWritten,
                            NULL);

    TRACE1("Back from IOCTL ow_fsctl_init;  Status = 0x%0x\n", Status);

    if(!Status)
        {
        //
        // Return the initialize Buffer
        //
        free((PUCHAR)InitializeBuffer);
        TRACE("IOCTL_OSRNFS_INIT failed.\n");
        return(-1);
        }

    //
    // Return the initialize Buffer
    //
    free((PUCHAR)InitializeBuffer);

    CloseHandle(FSDFileHandle);

#endif

    TRACE("\n***Network successfully Initialized!\n\n");

    NetworkInitialized = 1;

    return(0);
    }

//
// Character to Unicode conversion
//
// Takes an null terminated ASCII string, coverts it
// to UNICODE, and sticks it where indicated.  Returns
// a pointer to the next free location following the
// string (including the trailing null).
//
PUCHAR CharToUnicode(PUCHAR WhereItComesFrom, PUCHAR WhereToPutIt)
    {
    ULONG SourceLength;
    ULONG i, j;

    SourceLength = strlen(WhereItComesFrom);

    memset(WhereToPutIt, 0x00, (SourceLength*2)+2);

    for(i=0, j=0; i < SourceLength; i++)
        {

        WhereToPutIt[j++] = WhereItComesFrom[i];
        WhereToPutIt[j++] = 0x00;

        }

    WhereToPutIt[j++] = 0x00;
    WhereToPutIt[j++] = 0x00;

    return(WhereToPutIt+j);
    }
