// -*- Mode: c++ -*-
//
// (C) Copyright 1996 OSR Open Systems Resources, Inc.
// All Rights Reserved.
//

//
// This header file contains IOCTL function codes specific to the
// OSR FSDK Wrapper.
//

#ifndef __OSR_WRAPPER_UTIL_H__
#define __OSR_WRAPPER_UTIL_H__ (1)

//
// FSCTL codes declared externally.  Derived from public sources.
//

#define FSCTL_REQUEST_OPLOCK_LEVEL_1    CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 0, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_REQUEST_OPLOCK_LEVEL_2    CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 1, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_REQUEST_BATCH_OPLOCK      CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 2, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_OPLOCK_BREAK_ACKNOWLEDGE  CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_OPBATCH_ACK_CLOSE_PENDING CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 4, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_OPLOCK_BREAK_NOTIFY       CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 5, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_OPLOCK_BREAK_ACK_NO_2     CTL_CODE(FILE_DEVICE_FILE_SYSTEM,20, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_SHUTDOWN_VOLUME           CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 9, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_IS_VOLUME_MOUNTED         CTL_CODE(FILE_DEVICE_FILE_SYSTEM,10, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_IS_PATHNAME_VALID         CTL_CODE(FILE_DEVICE_FILE_SYSTEM,11, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_MARK_VOLUME_DIRTY         CTL_CODE(FILE_DEVICE_FILE_SYSTEM,12, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_QUERY_RETRIEVAL_POINTERS  CTL_CODE(FILE_DEVICE_FILE_SYSTEM,14, METHOD_NEITHER, FILE_ANY_ACCESS)

#define FSCTL_MARK_AS_SYSTEM_HIVE       CTL_CODE(FILE_DEVICE_FILE_SYSTEM,19, METHOD_NEITHER, FILE_ANY_ACCESS)

#define FSCTL_LOCK_VOLUME               CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 6, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_UNLOCK_VOLUME             CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 7, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define FSCTL_DISMOUNT_VOLUME           CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 8, METHOD_BUFFERED, FILE_ANY_ACCESS)

//
// OSR defined File System Control Codes.  Specific to the OSR file system(s).
//
#define OW_GET_TRACE_BUFFER             CTL_CODE(FILE_DEVICE_UNKNOWN, 2049, METHOD_OUT_DIRECT, FILE_READ_ACCESS)
#define OW_TRACE_PRINT_ON               CTL_CODE(FILE_DEVICE_UNKNOWN, 2051, METHOD_NEITHER, FILE_WRITE_ACCESS)
#define OW_TRACE_PRINT_OFF              CTL_CODE(FILE_DEVICE_UNKNOWN, 2053, METHOD_NEITHER, FILE_WRITE_ACCESS)
#define OW_FSCTL_MOUNT_PSEUDO           CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3031, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#define OW_FSCTL_FORCE_DISMOUNT_VOLUME  CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3032, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#define OW_FSCTL_DISMOUNT_PSEUDO        CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3033, METHOD_NEITHER, FILE_WRITE_ACCESS)

//
// Testing codes for testing IOCTLs within the wrapper.  All four of these simply COPY the data from
// input to output buffers.
//

#define OW_IOCTL_TEST_1 CTL_CODE(FILE_DEVICE_UNKNOWN, 2055, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define OW_IOCTL_TEST_2 CTL_CODE(FILE_DEVICE_UNKNOWN, 2057, METHOD_IN_DIRECT, FILE_ANY_ACCESS)
#define OW_IOCTL_TEST_3 CTL_CODE(FILE_DEVICE_UNKNOWN, 2059, METHOD_OUT_DIRECT, FILE_ANY_ACCESS)
#define OW_IOCTL_TEST_4 CTL_CODE(FILE_DEVICE_UNKNOWN, 2055, METHOD_NEITHER, FILE_ANY_ACCESS)

#define OW_FSCTL_TEST_1 CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3033, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define OW_FSCTL_TEST_2 CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3034, METHOD_IN_DIRECT, FILE_ANY_ACCESS)
#define OW_FSCTL_TEST_3 CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3035, METHOD_OUT_DIRECT, FILE_ANY_ACCESS)
#define OW_FSCTL_TEST_4 CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3036, METHOD_NEITHER, FILE_ANY_ACCESS)


//
// OSR Defined File System Control Codes for Network File Systems
//
#define OW_FSCTL_INIT                   CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3100, METHOD_NEITHER, FILE_ANY_ACCESS)
#define OW_FSCTL_ENUM                   CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3101, METHOD_NEITHER, FILE_READ_ACCESS)
#define OW_FSCTL_CONNECT                CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3102, METHOD_NEITHER, FILE_WRITE_ACCESS)
#define OW_FSCTL_DISCONNECT             CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3103, METHOD_NEITHER, FILE_WRITE_ACCESS)
#define OW_FSCTL_GETCONNECTIONS         CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3104, METHOD_NEITHER, FILE_READ_ACCESS)
#define OW_FSCTL_GET_FULL_CONNECTIONS   CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 3105, METHOD_NEITHER, FILE_READ_ACCESS)

//
// Pseudo mount/dismount data structures
//

typedef struct {
    HANDLE PseudoVolumeHandle;
    USHORT PseudoDeviceNameLength;
    PWCHAR PseudoDeviceName;
    USHORT PseudoLinkNameLength;
    PWCHAR PseudoLinkName;
} OW_PSEUDO_MOUNT_INFO, *POW_PSEUDO_MOUNT_INFO;

typedef struct {
    HANDLE  PseudoVolumeHandle;
    BOOLEAN Force;
} OW_PSEUDO_DISMOUNT_INFO, *POW_PSEUDO_DISMOUNT_INFO;


//
// Connect structure(s)
// Note that for backwards compatibility, we refer to the
// "connect" name alone.
//

typedef struct {
    WCHAR   ConnectionIdentifier; // connection identifier requested
    UCHAR   ConnectionType;       // disk, printer
    WCHAR   ConnectionName[1];       // connection name to use
} OW_FSCTL_CONNECT_INFO, *POW_FSCTL_CONNECT_INFO;

typedef struct {
    WCHAR   ConnectionIdentifier; // drive letter in use (if any)
    UCHAR   Force;                // disconnect should be forced
    WCHAR   ConnectionName[1];       // connection name to use
} OW_FSCTL_DISCONNECT_INFO, *POW_FSCTL_DISCONNECT_INFO;

typedef struct {
    ULONG NumberOfEntries;
    OW_FSCTL_CONNECT_INFO Connections[1];
} OW_FSCTL_GETCONNECTION_INFO, *POW_FSCTL_GETCONNECTION_INFO;

//
// connection types
//

#define OW_CONNECTION_DISK       (0x00)
#define OW_CONNECTION_PRINTER    (0x01)


//
// These IOCTL codes are from winioctl.h - they aren't IN the DDK, but they
// should be...
//

#ifndef IOCTL_DISK_BASE
#define IOCTL_DISK_BASE FILE_DEVICE_DISK
#endif

#ifndef IOCTL_DISK_CHECK_VERIFY
#define IOCTL_DISK_CHECK_VERIFY     CTL_CODE(IOCTL_DISK_BASE, 0x0200, METHOD_BUFFERED, FILE_READ_ACCESS)
#endif

#ifndef IOCTL_DISK_MEDIA_REMOVAL
#define IOCTL_DISK_MEDIA_REMOVAL    CTL_CODE(IOCTL_DISK_BASE, 0x0201, METHOD_BUFFERED, FILE_READ_ACCESS)
#endif

#ifndef IOCTL_DISK_EJECT_MEDIA
#define IOCTL_DISK_EJECT_MEDIA      CTL_CODE(IOCTL_DISK_BASE, 0x0202, METHOD_BUFFERED, FILE_READ_ACCESS)
#endif

#ifndef IOCTL_DISK_LOAD_MEDIA
#define IOCTL_DISK_LOAD_MEDIA       CTL_CODE(IOCTL_DISK_BASE, 0x0203, METHOD_BUFFERED, FILE_READ_ACCESS)
#endif 

#ifndef IOCTL_DISK_RESERVE
#define IOCTL_DISK_RESERVE          CTL_CODE(IOCTL_DISK_BASE, 0x0204, METHOD_BUFFERED, FILE_READ_ACCESS)
#endif

#ifndef IOCTL_DISK_RELEASE
#define IOCTL_DISK_RELEASE          CTL_CODE(IOCTL_DISK_BASE, 0x0205, METHOD_BUFFERED, FILE_READ_ACCESS)
#endif

#ifndef IOCTL_DISK_FIND_NEW_DEVICES
#define IOCTL_DISK_FIND_NEW_DEVICES CTL_CODE(IOCTL_DISK_BASE, 0x0206, METHOD_BUFFERED, FILE_READ_ACCESS)
#endif

#ifndef IOCTL_DISK_GET_MEDIA_TYPES
#define IOCTL_DISK_GET_MEDIA_TYPES CTL_CODE(IOCTL_DISK_BASE, 0x0300, METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif

#endif // __OSR_WRAPPER_UTIL_H__



